/* wv1522.cpp V 1.5  1997/03/20 00:00:00 heirich */
/*********************************************************************
 *
 * Module: WV1522.CPP
 *
 * send bugreports to <winvn@heirich.in-berlin.de>
 *
 * Purpose: Support for RFC 1522
 * Version 1.7
 *********************************************************************/
/*********************************************************************
 * Notes:
 *
 * there is a flag strictRFC in class WVRFCDecoder
 * if TRUE decoding works like defined in RFC1522
 * if FALSE a modification is done for postings created
 * by Netscape ( Modzilla ) 2.01
 * Netscape doesn't fit RFC1522 because it creates
 * f=?ISO-8859-1?Q?=FC?=r instead =?ISO-8859-1?Q?f=FCr?= which is correct for "fr"
 *
 * the classes don't make full syntax-analyses of headers by RFC822
 * BE WARNED: if you use non-ASCII-characters in a part which should not
 * contains these characters, it will work wrong
 *********************************************************************/
/*********************************************************************
 * Changelog:   E = encoder / D = decoder
 *
 * V 1.7	fixed following bugs
 *	      E -during encoding the string expands in length, this causes some
 *		overwrite of data if string isn't long enougth
 *		using WVString class this caused GPF because memory pointers will be
 *		overwritten.
 *
 * V 1.6        fixed following bugs
 * not fixed  D -by RFC is said, that if only ASCII chars in encoded text for ISO-8859-X
 *              we should show this as normal
 * not fixed  D -quoted pairs are not handled properly
 * not fixed  E -a word which begins with =? and ends with ?= should always encoded
 *              using charset US-ASCII if no other need to encode. see chapter 7 of RFC
 * not fixed  E -maybe in text fields like subject "(" and ")" doesn't enclose
 *              a comment. So there is *no* word boundary.
 * not fixed  E -enshure that long words ( > 75 characters after encoding )
 *              handled properly
 * not fixed  E -enshure that the 76 character max line length defined in RFC
 *              is handled properly
 * not fixed  D -enshure that folding ( encoded words delimited by CRLF SPACE )
 *              handled properly
 * not fixed  E -encoding of (was: A) works not properly
 *              maybe we should insert folding with CRLF SPACE before (
 *
 * V 1.5
 *
 * 1997/05/19 D -Error on aborting when unknown charset, status must changed
 *              as last statement when all abort conditions are checked.
 * 
 * 1997/05/03 D -don't decode if it contains 0x0A or 0x0D
 *
 * V 1.4        -added support for the new WVString-class
 *              introduced with 0.99.9 of WinVN 
 *
 * V 1.3        new layout of classes to make it usefull for other purposes
 * 1996/09/05   -counting of nonASCII-characters is done in the new RFCFIFO class
 *              instead of FIFO
 * 1996/09/05   -make the most work of quoted-printable decoding in a internal class
 *              WVQIDecoder, but not the RFC1522 special replacement of _ with 0x20
 *              this is done in WVQDecoder. so we are able to put RFC1521 quoted-printable
 *              into this class ( or a inherited from this ) and take it from.
 * 1996/09/05   -modify it to be able to use in a UNICODE environment
 *              using TCHAR.H
 * 1996/09/04   -make able to encode and decode charsets with 2 byte width ( UNICODE )
 *              by my own additional implementation. THIS DOESN'T FIT IN ANY RFC !
 *
 *              fixed following bugs
 * 1996/09/04   -multiword- encoding in B doesn't work properly. There was a markup of stream
 *              after every word, but we should this only done after all words.
 *
 * V 1.2        fixed following bugs
 * 1996/08/24 D -fixed that a sequence =?text with text > 75 chars was wrong
 *              handled as RFC encoding. Now we abort because no RFC-encoding
 * 1996/08/24 D -fixed that a sequence =?text1?text2 with text2 > 75 chars was
 *              wrong handled as RFC encoding. Now we abort because no RFC-encoding
 * 1996/08/22 D -aborting maps charset an coder to uppercase. this is wrong
 *              fixed, that it works case insensitive
 * 1996/08/22 E -calculate length was improperly done over all text
 *              by RFC B encoding should used if more non-ASCII characters
 *              then ASCII-characters must encode.
 * 1996/08/19 E -in Q-Encoding length calculating works not properly
 *              because ?,= and _ must encoded and (, ) should encoded and needs 3 characters
 * 1996/08/22 D -if there is a characterset which we don't understand
 *              we should abort decoding
 * 1996/08/19 D -if charset, coder or encoded text are empty
 *              we should abort decoding, because all of them has at least
 *              1 character in ABNF grammar in RFC1522
 *
 * V 1.1        initial release
 *              fundamental rewrite from 1.0, allowes multiword encoding
 *              and aborting decoding if we found it doesn't fit RFC1522
 *              syntax
 *
 * V 1.0        earlier release for test and first internal usage
 *
 *********************************************************************/


#ifdef STANDALONE
#define RFC1522

#include <windows.h>
#include <windowsx.h>
#include <tchar.h>
#include <stdio.h>
#include <string.h>


#include "wv1522.h"

BOOL UseRFC1522 = TRUE;
BOOL EncodeRFC1522 = TRUE;


#else

#include <windows.h>
#include <windowsx.h>
#include "wvglob.h"
#include "winvn.h"
#pragma hdrstop
#include "WVClass.h"

#endif

#include <ctype.h>

#ifdef RFC1522


/*******************************************************************************************/

#ifdef UNICODE
size_t strtowcs(wchar_t *pwcs, char *s)
{
    size_t count = 0;
    
    while (*s) {
        *pwcs++ = *s++;
        count++;
    }
    *pwcs = L'\0';
    return count;
}


size_t strntowcs(wchar_t *pwcs, char *s, size_t n)
{
    size_t count = 0;
    
    while ((--n > 0) && (*s)) {
        *pwcs++ = *s++;
        count++;
    }
    *pwcs = L'\0';
    return count;
}


/* function wcstostr and wcsntostr are only usefull
 * when the wide character string only contains ISO-8859-1 characters
 */
size_t wcstostr(char *s, wchar_t *pwcs)
{
    size_t count = 0;
    
    while (*pwcs) {
        *s++ = (char)*pwcs++;
        count++;
    }
    *s = '\0';
    return count;
}


size_t wcsntostr(char *s, wchar_t *pwcs, size_t n)
{
    size_t count = 0;
    
    while ((--n > 0) && (*pwcs)) {
        *s++ = (char)*pwcs++;
        count++;
    }
    *s = '\0';
    return count;
}
#endif
/*******************************************************************************************/

FIFO::FIFO()
{
    last = anchor = NULL;
    PendingCharacters = 0;
}

FIFO::FIFO(const char * s)
{
    last = anchor = NULL;
    PendingCharacters = 0;
    while (*s) Putchar(*s++);
}

FIFO::FIFO(unsigned int n,const char *s)
{
    last = anchor = NULL;
    PendingCharacters = 0;
    while ( n > 0) {
      Putchar(*s++);
      n--;
    };
}

#ifdef UNICODE
FIFO::FIFO(const wchar_t * s)
{
    last = anchor = NULL;
    PendingCharacters = 0;
    while (*s) Putchar(*s++);
}

FIFO::FIFO(unsigned int n,const wchar_t *s)
{
    last = anchor = NULL;
    PendingCharacters = 0;
    while ( n > 0) {
      Putchar(*s++);
      n--;
    };
}

#endif

FIFO::~FIFO()
{
    DeleteAll();
}


void FIFO::Chain(const TCHAR ch)
{
    WVC_Char *p;

    /* create a new element and setup properly */
    p = new WVC_Char;
    p->next = NULL;
    p->c = ch;

    /* put it in the chain */
    if (anchor == NULL) {
        anchor = p; /* first in the chain */
    } else {
        last->next = p;  /* append after the last of the chain */
    }
    last = p; /* setup last properly */

    PendingCharacters++;
}


void FIFO::Ungetchar(const TCHAR ch)
{
   WVC_Char *p;

   /* create a new element and put it on the begin of the chain */
   p = new WVC_Char;
   p->next = anchor;
   p->c = ch;

   anchor = p;
   PendingCharacters++;
}


void FIFO::DeleteAll()
{
    int junk;
    
    while ((junk = Getchar()) != EOF);
}

int FIFO::Testchar()
{
    if (anchor == NULL )
       return EOF;
    return anchor->c;
}

int FIFO::Getchar()
{
    TCHAR wch;
    WVC_Char * p;

    if (anchor == NULL)
       return EOF;
    p = anchor->next;
    wch = anchor->c;
    delete anchor;
    anchor = p;
    PendingCharacters--;
    return wch;
}

BOOL FIFO::Getallowed()
{
    return (anchor!=NULL);
}


void FIFO::Putchar(const TCHAR ch)
{
    Chain(ch);
}


BOOL FIFO::Putallowed()
{
    return TRUE; /* currently always true, but maybe later usefull */
}

unsigned int FIFO::Pendingchars()
{
    return PendingCharacters;
}

void FIFO::EndOfStream()
{
}

/******************************************************************************************/
RFCFIFO::RFCFIFO()
{
    nonASCIICharacters = 0;
}



RFCFIFO::RFCFIFO(const char * s)
{
    nonASCIICharacters = 0;
    while (*s) Putchar(*s++);
}


RFCFIFO::RFCFIFO(unsigned int n,const char *s)
{
    nonASCIICharacters = 0;
    while ( n > 0) {
      Putchar(*s++);
      n--;
    };
}

#ifdef UNICODE
RFCFIFO::RFCFIFO(const wchar_t * s)
{
    nonASCIICharacters = 0;
    while (*s) Putchar(*s++);
}

RFCFIFO::RFCFIFO(unsigned int n,const wchar_t *s)
{
    nonASCIICharacters = 0;
    while ( n > 0) {
      Putchar(*s++);
      n--;
    };
}
#endif

RFCFIFO::~RFCFIFO()
{
}

void RFCFIFO::Chain(const TCHAR ch)
{
    if (!_istascii(ch))
        nonASCIICharacters++;
    FIFO::Chain(ch);
}

void RFCFIFO::Ungetchar(const TCHAR ch)
{
   if (!_istascii(ch))
      nonASCIICharacters++;
   FIFO::Ungetchar(ch);
}

int RFCFIFO::Getchar()
{
    int wch;
    
    wch = FIFO::Getchar();
    if (!_istascii((TCHAR)wch))
       nonASCIICharacters--;
    return wch;
}

unsigned int RFCFIFO::NonASCIIchars()
{
    return nonASCIICharacters;
}

/******************************************************************************************/

WVBDecoder::~WVBDecoder()
{
}

WVBDecoder::WVBDecoder()
{
#ifdef UNICODE    
    octetwidth = 1;
#endif    
    Base64Index = Base64Bits = 0;
#ifdef RFC1113
    rfc1113 = directCharacters = FALSE;
#endif
}

#ifdef RFC1113
WVBDecoder::WVBDecoder(BOOL useRFC1113)
{
#ifdef UNICODE    
    octetwidth = 1;
#endif    
    Base64Index = Base64Bits = 0;
    directCharacters = FALSE;
    rfc1113 = useRFC1113;
}
#endif

WVBDecoder::WVBDecoder(const char * s)
{
#ifdef UNICODE    
    octetwidth = 1;
#endif    
    Base64Index = Base64Bits = 0;
#ifdef RFC1113
    rfc1113 = directCharacters = FALSE;
#endif
    while (*s) Putchar(*s++);
}


#ifdef RFC1113
WVBDecoder::WVBDecoder(const char * s, BOOL useRFC1113)
{
#ifdef UNICODE    
    octetwith = 1;
#endif    
    Base64Index = Base64Bits = 0;
    directCharacters = FALSE;
    rfc1113 = useRFC1113;
    while (*s) Putchar(*s++);
}
#endif

WVBDecoder::WVBDecoder(unsigned int n, const char * s)
{
#ifdef UNICODE    
    octetwidth = 1;
#endif    
    Base64Index = Base64Bits = 0;
#ifdef RFC1113
    rfc1113 = directCharacters = FALSE;
#endif
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

#ifdef RFC1113
WVBDecoder::WVBDecoder(unsigned int n, const char * s, BOOL useRFC1113)
{
#ifdef UNICODE    
    octetwidth = 1;
#endif    
    Base64Index = Base64Bits = 0;
    directCharacters = FALSE;
    rfc1113 = useRFC1113;
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}
#endif

#ifdef UNICODE
WVBDecoder::WVBDecoder(unsigned int octets)
{
    octetwidth = octets;
    Base64Index = Base64Bits = 0;
#ifdef RFC1113
    rfc1113 = directCharacters = FALSE;
#endif
}

#ifdef RFC1113
WVBDecoder::WVBDecoder(BOOL useRFC1113, unsigned int octets)
{
    octetwidth = octets;
    Base64Index = Base64Bits = 0;
    directCharacters = FALSE;
    rfc1113 = useRFC1113;
}
#endif

WVBDecoder::WVBDecoder(const char * s, unsigned int octets)
{
    octetwidth = octets;
    Base64Index = Base64Bits = 0;
#ifdef RFC1113
    rfc1113 = directCharacters = FALSE;
#endif
    while (*s) Putchar(*s++);
}


#ifdef RFC1113
WVBDecoder::WVBDecoder(const char * s, BOOL useRFC1113, unsigned int octets)
{
    octetwith = octets;
    Base64Index = Base64Bits = 0;
    directCharacters = FALSE;
    rfc1113 = useRFC1113;
    while (*s) Putchar(*s++);
}
#endif

WVBDecoder::WVBDecoder(unsigned int n, const char * s, unsigned int octets)
{
    octetwidth = octets;
    Base64Index = Base64Bits = 0;
#ifdef RFC1113
    rfc1113 = directCharacters = FALSE;
#endif
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

#ifdef RFC1113
WVBDecoder::WVBDecoder(unsigned int n, const char * s, BOOL useRFC1113, unsigned int octets)
{
    octetwidth = octets;
    Base64Index = Base64Bits = 0;
    directCharacters = FALSE;
    rfc1113 = useRFC1113;
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}
#endif

WVBDecoder::WVBDecoder(const wchar_t * s)
{
    octetwidth = 1;
    Base64Index = Base64Bits = 0;
#ifdef RFC1113
    rfc1113 = directCharacters = FALSE;
#endif
    while (*s) Putchar(*s++);
}

WVBDecoder::WVBDecoder(const wchar_t * s, unsigned int octets)
{
    octetwidth = octets;
    Base64Index = Base64Bits = 0;
#ifdef RFC1113
    rfc1113 = directCharacters = FALSE;
#endif
    while (*s) Putchar(*s++);
}

#ifdef RFC1113
WVBDecoder::WVBDecoder(const wchar_t * s, BOOL useRFC1113)
{
    octetwidth = 1;
    Base64Index = Base64Bits = 0;
    directCharacters = FALSE;
    rfc1113 = useRFC1113;
    while (*s) Putchar(*s++);
}

WVBDecoder::WVBDecoder(const wchar_t * s, BOOL useRFC1113, unsigned int octets)
{
    octetwidth = octets;
    Base64Index = Base64Bits = 0;
    directCharacters = FALSE;
    rfc1113 = useRFC1113;
    while (*s) Putchar(*s++);
}
#endif

WVBDecoder::WVBDecoder(unsigned int n, const wchar_t * s)
{
    octetwidth = 1;
    Base64Index = Base64Bits = 0;
#ifdef RFC1113
    rfc1113 = directCharacters = FALSE;
#endif
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

WVBDecoder::WVBDecoder(unsigned int n, const wchar_t * s, unsigned int octets)
{
    octetwidth = octets;
    Base64Index = Base64Bits = 0;
#ifdef RFC1113
    rfc1113 = directCharacters = FALSE;
#endif
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

#ifdef RFC1113
WVBDecoder::WVBDecoder(unsigned int n, const wchar_t * s, BOOL useRFC1113)
{
    octetwidth = 1;
    Base64Index = Base64Bits = 0;
    directCharacters = FALSE;
    rfc1113 = useRFC1113;
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

WVBDecoder::WVBDecoder(unsigned int n, const wchar_t * s, BOOL useRFC1113,
                       unsigned int octets)
{
    octetwidth = octets;
    Base64Index = Base64Bits = 0;
    directCharacters = FALSE;
    rfc1113 = useRFC1113;
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}
#endif
#endif // UNICODE

#ifdef JAPANESE
void WVBDecoder::DeleteAll() // S.Yoshida
{
    int junk;

    while ((junk = Getchar()) != EOF);
    Base64Index = Base64Bits = 0;
}
#endif


BYTE WVBDecoder::base64num(const char ch)
{
    if ((ch >= 'A') && (ch <= 'Z'))
       return (BYTE)(ch-'A');
    if ((ch >= 'a') && (ch <= 'z'))
      return (BYTE)(ch-'a'+26);
    if ((ch >= '0') && (ch <= '9'))
      return (BYTE)(ch-'0'+52);
    if (ch == '+')
      return (BYTE)62;
   if (ch == '/')
      return (BYTE)63;
   return (BYTE)0;
}


void WVBDecoder::Putchar(const TCHAR ch)
{
   unsigned int c;
#ifdef RFC1113
    if ((rfc1113) && (directCharacters)) {
      if (ch == '*')
        directCharacters = !directCharacters;  
      else
        Chain(ch);  
    } else {
#endif
      if ((ch >= 'A' && ch <= 'Z') || (ch >= 'a' && ch <= 'z') || (ch >= '0' && ch <= '9') ||
          (ch == '+') || (ch == '/')) {
        Base64Bits = (Base64Bits << 6) + base64num((char)ch);
#ifdef UNICODE
        if (octetwidth == 1) {
#endif
           switch ( ++Base64Index) {
                case 1:
                   break; 
                case 2:
#ifdef JAPANESE
                   if ((c = (Base64Bits >> 4)) != 0)
#else
                   if ((c = (Base64Bits >> 4))>=0x20)
#endif
                        Chain((char)c);
                   Base64Bits = Base64Bits & 0x0F;
                   break;
                case 3:
#ifdef JAPANESE
                   if ((c = (Base64Bits >> 2)) != 0)
#else
                   if ((c = (Base64Bits >> 2)) >= 0x20)
#endif
                        Chain((char)c);
                   Base64Bits = Base64Bits & 0x03;
                   break;
                case 4:
#ifdef JAPANESE
                   if (Base64Bits != 0)
#else
                   if (Base64Bits >= 0x20)
#endif
                       Chain((char)Base64Bits);
                   Base64Bits = 0;
                   Base64Index = 0;
           }
#ifdef UNICODE     
        } else {
           switch ( ++Base64Index) {
                case 1:
                case 2:
                   break; 
                case 3:
#ifdef JAPANESE
                   if ((c = (Base64Bits >> 2)) != 0)
#else
                   if ((c = (Base64Bits >> 2)) >= 0x0020)
#endif
                        Chain((wchar_t)c);
                   Base64Bits = Base64Bits & 0x0003;
                   break;
                case 4:
                case 5:
                   break;
                case 6:
#ifdef JAPANESE
                   if ((c = (Base64Bits >> 4)) != 0)
#else
                   if ((c = (Base64Bits >> 4)) >= 0x0020)
#endif
                        Chain((wchar_t)c);
                   Base64Bits = Base64Bits & 0x000F;
                case 7:
                   break;
                case 8:
#ifdef JAPANESE
                   if (Base64Bits != 0)
#else
                   if (Base64Bits >= 0x0020)
#endif
                       Chain((wchar_t)Base64Bits);
                   Base64Bits = 0;
                   Base64Index = 0;
           } 
        }
#endif  
      }
      if (ch == '=')
         Base64Index = Base64Bits = 0;
#ifdef RFC1113
      if ((rfc1113) && (ch == '*'))
           directCharacters = !directCharacters;
    }
#endif
}


/******************************************************************************************/

WVQIDecoder::~WVQIDecoder()
{
}

WVQIDecoder::WVQIDecoder()
{
#ifdef UNICODE    
    octetwidth = 1;
#endif    
    outstanding = 0;
}

WVQIDecoder::WVQIDecoder(const char * s)
{
#ifdef UNICODE    
    octetwidth = 1;
#endif    
    outstanding = 0;
    while (*s) Putchar(*s++);
}

WVQIDecoder::WVQIDecoder(unsigned int n, const char * s)
{
#ifdef UNICODE    
    octetwidth = 1;
#endif    
    outstanding = 0;
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

#ifdef UNICODE
WVQIDecoder::WVQIDecoder(unsigned int octets)
{
    octetwidth = octets;
    outstanding = 0;
}

WVQIDecoder::WVQIDecoder(const char * s, unsigned int octets)
{
    octetwidth = octets;
    outstanding = 0;
    while (*s) Putchar(*s++);
}

WVQIDecoder::WVQIDecoder(unsigned int n, const char * s, unsigned int octets)
{
    octetwidth = octets;
    outstanding = 0;
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

WVQIDecoder::WVQIDecoder(const wchar_t * s)
{
    octetwidth = 1;
    outstanding = 0;
    while (*s) Putchar(*s++);
}

WVQIDecoder::WVQIDecoder(const wchar_t * s, unsigned int octets)
{
    octetwidth = octets;
    outstanding = 0;
    while (*s) Putchar(*s++);
}

WVQIDecoder::WVQIDecoder(unsigned int n, const wchar_t * s)
{
    octetwidth = 1;
    outstanding = 0;
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

WVQIDecoder::WVQIDecoder(unsigned int n, const wchar_t * s, unsigned int octets)
{
    octetwidth = octets;
    outstanding = 0;
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}
#endif

BYTE WVQIDecoder::hexnum(const char ch)
{
    if ((ch >= '0') && (ch <= '9'))
       return (BYTE)(ch-'0');
    if ((ch >= 'A') && (ch <= 'F'))
      return (BYTE)(ch-'A'+0x0A);
    if ((ch >= 'a') && (ch <='f'))
      return (BYTE)(ch-'a'+0x0A);
   return (BYTE)0;  
}

void WVQIDecoder::Putchar(const TCHAR ch)
{
    unsigned int i;
    unsigned int j;

    if (outstanding) {
       --outstanding;
       *DecoderPtr++ = (char)ch; /* hexchars in sequence =xy */
       if (!outstanding) {
         /* encoded sequence is complete */
         DecoderPtr = DecoderBlock;
         j = 0;
#ifdef UNICODE
         for (i = 0;i < decodelength ; i++) {
           j = (j << 4) + hexnum(*DecoderPtr++);
         } /* for */
#else
         for (i = 0;i < 2 ; i++) {
           j = (j << 4) + hexnum(*DecoderPtr++);
         } /* for */
#endif
#ifdef JAPANESE
         if (j != 0) /* should be a space or above, no control-chars ! */
#else
         if (j >= 0x20) /* should be a space or above, no control-chars ! */
#endif
           Chain((TCHAR)j);
       } /* if no more outstanding characters */
    } else {
       if (ch == '=') {
#ifdef UNICODE   
         decodelength = outstanding = (octetwidth == 1)?2:4;
#else
         outstanding = 2;
#endif   
         DecoderPtr = DecoderBlock;
       } else {
         Chain(ch);
       }
    }
}

/******************************************************************************************/

WVBCoder::~WVBCoder()
{
}

WVBCoder::WVBCoder()
{
#ifdef UNICODE    
    octetwidth = 1;
#endif    
    Base64Index = Base64Bits = 0;
}


WVBCoder::WVBCoder(const char * s)
{
#ifdef UNICODE    
    octetwidth = 1;
#endif    
    Base64Index = Base64Bits = 0;
    while (*s) Putchar(*s++);
}

WVBCoder::WVBCoder(unsigned int n, const char * s)
{
#ifdef UNICODE   
    octetwidth = 1;
#endif
    Base64Index = Base64Bits = 0;
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

#ifdef UNICODE
WVBCoder::WVBCoder(unsigned int octets)
{
    octetwidth = octets;
    Base64Index = Base64Bits = 0;
}


WVBCoder::WVBCoder(const char * s, unsigned int octets)
{
    octetwidth = octets;
    Base64Index = Base64Bits = 0;
    while (*s) Putchar(*s++);
}

WVBCoder::WVBCoder(unsigned int n, const char * s, unsigned int octets)
{
    octetwidth = octets;
    Base64Index = Base64Bits = 0;
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

WVBCoder::WVBCoder(const wchar_t * s, unsigned int octets)
{
    octetwidth = octets;
    Base64Index = Base64Bits = 0;
    while (*s) Putchar(*s++);
}


WVBCoder::WVBCoder(unsigned int n, const wchar_t * s, unsigned int octets)
{
    octetwidth = octets;
    Base64Index = Base64Bits = 0;
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}
#endif

char WVBCoder::base64char(const BYTE base64)
{
    const char Base64Table[]
    ="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
    
    if (base64 <= 63)
       return Base64Table[base64];
    return ' ';
}


void WVBCoder::EndOfStream()
{
    Base64Bits = Base64Bits <<8;
    switch ( ++Base64Index ) {
        case 1:
            /* encoding input was integral multiple of 24 Bits */
            break;
        case 2:
            /* encoding input has a final quantum of 8 bit */
            Chain(base64char((BYTE)(Base64Bits >> 4)));
            Chain('=');
            Chain('=');
            break;
        case 3:
            /* encoding input has a final quantum of 16 bit*/
            Chain(base64char((BYTE)(Base64Bits >> 6)));
            Chain('=');
    }
    
    Base64Bits = 0;
    Base64Index = 0;

    FIFO::EndOfStream(); /* say FIFO stack that stream has ended */
}


void WVBCoder::Putchar(const char ch)
{
    Base64Bits = (Base64Bits << 8) + ch;
    switch ( ++Base64Index ) {
        case 1:
            Chain(base64char((BYTE)(Base64Bits >> 2)));
            Base64Bits = Base64Bits & 0x03;
            break;
        case 2:
            Chain(base64char((BYTE)(Base64Bits >> 4)));
            Base64Bits = Base64Bits & 0x0F;
            break;
        case 3:
            Chain(base64char((BYTE)(Base64Bits >> 6)));
            Base64Bits = Base64Bits & 0x3F;
            Chain(base64char((BYTE)Base64Bits));
            Base64Bits = 0;
            Base64Index = 0;
    }
}

#ifdef UNICODE
void WVBCoder::Putchar(const wchar_t ch)
{
    if (octetwidth != 1)
        Putchar((char)(ch >> 8));
    Putchar((char)ch);
}
#endif

/*******************************************************************************************/

WVQICoder::~WVQICoder()
{
}

WVQICoder::WVQICoder()
{
#ifdef UNICODE    
    octetwidth = 1;
#endif    
}


WVQICoder::WVQICoder(const char * s)
{
#ifdef UNICODE    
    octetwidth = 1;
#endif    
    while (*s) Putchar(*s++);
}

WVQICoder::WVQICoder(unsigned int n, const char * s)
{
#ifdef UNICODE    
    octetwidth = 1;
#endif    
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

#ifdef UNICODE
WVQICoder::WVQICoder(unsigned int octets)
{
    octetwidth = octets;
}


WVQICoder::WVQICoder(const char * s, unsigned int octets)
{
    octetwidth = octets;
    while (*s) Putchar(*s++);
}

WVQICoder::WVQICoder(unsigned int n, const char * s, unsigned int octets)
{
    octetwidth = octets;
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

WVQICoder::WVQICoder(const wchar_t * s)
{
    octetwidth = 1;
    while (*s) Putchar(*s++);
}

WVQICoder::WVQICoder(const wchar_t * s, unsigned int octets)
{
    octetwidth = octets;
    while (*s) Putchar(*s++);
}

WVQICoder::WVQICoder(unsigned int n, const wchar_t * s)
{
    octetwidth = 1;
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

WVQICoder::WVQICoder(unsigned int n, const wchar_t * s, unsigned int octets)
{
    octetwidth = octets;
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}
#endif

char WVQICoder::hexchar(const BYTE hex)
{
    const char HexTable[]
    ="0123456789ABCDEF";

    if (hex <= 0x0F)
       return HexTable[hex];
    return ' ';
}

void WVQICoder::Putchar(const TCHAR ch)
{
    if (_istascii(ch) && _istprint(ch) &&
                (ch != '_') && (ch != '=') && (ch != '?') &&
                (ch != '(') && (ch != ')') ) {
          Chain(ch);
    } else {
#ifdef UNICODE  
          if (octetwidth != 1) {
              Chain('=');
              Chain(hexchar( (BYTE)((ch >> 12) & 0x0F)));
              Chain(hexchar( (BYTE)((ch >> 8) & 0x0F)));
              Chain(hexchar( (BYTE)((ch >> 4) & 0x0F)));
              Chain(hexchar( (BYTE)((ch) & 0x0F)));           
          } else {
#endif        
              Chain('=');
              Chain(hexchar( (BYTE)((ch >> 4) & 0x0F)));
              Chain(hexchar( (BYTE)((ch) & 0x0F)));
#ifdef UNICODE        
          }
#endif    
    }
}


/******************************************************************************************/

WVQCoder::~WVQCoder()
{
}

WVQCoder::WVQCoder()
{
}

WVQCoder::WVQCoder(const char * s)
{
    while (*s) Putchar(*s++);
}

WVQCoder::WVQCoder(unsigned int n, const char * s)
{
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

#ifdef UNICODE
WVQCoder::WVQCoder(unsigned int octets) : WVQICoder(octets)
{
}


WVQCoder::WVQCoder(const char * s, unsigned int octets) : WVQICoder(octets)
{
    while (*s) Putchar(*s++);
}

WVQCoder::WVQCoder(unsigned int n, const char * s, unsigned int octets) : WVQICoder(octets)
{
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

WVQCoder::WVQCoder(const wchar_t * s, unsigned int octets) : WVQICoder(octets)
{
    while (*s) Putchar(*s++);
}


WVQCoder::WVQCoder(unsigned int n, const wchar_t * s, unsigned int octets)
                  : WVQICoder(octets)
{
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}
#endif

void WVQCoder::Putchar(const TCHAR ch)
{
    if (ch == 0x20 ) {
        WVQICoder::Putchar('_');
    } else {
        WVQICoder::Putchar(ch);
    }
}

/******************************************************************************************/

WVQDecoder::~WVQDecoder()
{
}

WVQDecoder::WVQDecoder()
{
}

WVQDecoder::WVQDecoder(const char * s)
{
    while (*s) Putchar(*s++);
}

WVQDecoder::WVQDecoder(unsigned int n, const char * s)
{
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

#ifdef UNICODE
WVQDecoder::WVQDecoder(unsigned int octets) : WVQIDecoder(octets)
{
}

WVQDecoder::WVQDecoder(const char * s, unsigned int octets) : WVQIDecoder(octets)
{
    while (*s) Putchar(*s++);
}

WVQDecoder::WVQDecoder(unsigned int n, const char * s, unsigned int octets)
                      : WVQIDecoder(octets)
{
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

WVQDecoder::WVQDecoder(const wchar_t * s)
{
    while (*s) Putchar(*s++);
}


WVQDecoder::WVQDecoder(const wchar_t * s, unsigned int octets) : WVQIDecoder(octets)
{
    while (*s) Putchar(*s++);
}

WVQDecoder::WVQDecoder(unsigned int n, const wchar_t * s)
{
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}

WVQDecoder::WVQDecoder(unsigned int n, const wchar_t * s, unsigned int octets)
                      : WVQIDecoder(octets)
{
    while (n > 0) {
       Putchar(*s++);
       n--;
    }
}
#endif


void WVQDecoder::Putchar(const TCHAR ch)
{
    
    if (ch == '_')
      WVQIDecoder::Putchar(0x20); /* substitution defined in Q encoding 4.2.(2) */
    else
      WVQIDecoder::Putchar(ch);
}

/******************************************************************************************/

WVDecoder::~WVDecoder()
{
    switch (whichDecoder) {
        case DecoderQ:
              delete (WVQDecoder *)Decoder;
              break;
        case DecoderB:
        default:
              delete (WVBDecoder *)Decoder;
    }
}

#ifdef UNICODE
void WVDecoder::Init(const char * s, unsigned int octets)
{
    if (_stricmp(s,"Q") == 0) {
        whichDecoder = DecoderQ;
    };
    if (_stricmp(s,"B") == 0) {
        whichDecoder = DecoderB;
    };
    switch (whichDecoder) {
        case DecoderQ:
              Decoder = (void *) new WVQDecoder(octets);
              break;
        case DecoderB:
        default:
#ifdef RFC1113
              Decoder = (void *) new WVBDecoder(TRUE,octets);
#else
              Decoder = (void *) new WVBDecoder(octets);
#endif
    }
}
#else //UNICODE
void WVDecoder::Init(const char * s)
{
    if (_stricmp(s,"Q") == 0) {
        whichDecoder = DecoderQ;
    };
    if (_stricmp(s,"B") == 0) {
        whichDecoder = DecoderB;
    };
    switch (whichDecoder) {
        case DecoderQ:
              Decoder = (void *) new WVQDecoder;
              break;
        case DecoderB:
        default:
#ifdef RFC1113
              Decoder = (void *) new WVBDecoder(TRUE);
#else
              Decoder = (void *) new WVBDecoder;
#endif
    }
}
#endif //UNICODE

WVDecoder::WVDecoder(const char * s)
{
#ifdef UNICODE
    Init(s,1);
#else
    Init(s);
#endif    
}

#ifdef UNICODE
WVDecoder::WVDecoder(const char * s, unsigned int octets)
{
    Init(s, octets);
}

WVDecoder::WVDecoder(wchar_t * s)
{
    char mybuf[WVC_CODERNAMELENGTH];

    wcsntostr(mybuf, s, WVC_CODERNAMELENGTH);  
    Init(mybuf,1);
}

WVDecoder::WVDecoder(wchar_t * s, unsigned int octets)
{
    char mybuf[WVC_CODERNAMELENGTH];

    wcsntostr(mybuf, s, WVC_CODERNAMELENGTH);  
    Init(mybuf,octets);
}
#endif

void WVDecoder::EndOfStream()
{
    switch (whichDecoder) {
        case DecoderQ:
              ((WVQDecoder *)Decoder)->EndOfStream();
              break;
        case DecoderB:
        default:
              ((WVBDecoder *)Decoder)->EndOfStream();
    }
}

void WVDecoder::Putchar(const TCHAR ch)
{
    switch (whichDecoder) {
        case DecoderQ:
              ((WVQDecoder *)Decoder)->Putchar(ch);
              break;
        case DecoderB:
        default:
              ((WVBDecoder *)Decoder)->Putchar(ch);
    }
}

BOOL WVDecoder::Putallowed()
{
    switch (whichDecoder) {
        case DecoderQ:
              return ((WVQDecoder *)Decoder)->Putallowed();
        case DecoderB:
        default:
              return ((WVBDecoder *)Decoder)->Putallowed();
    }
}

int WVDecoder::Testchar()
{
    switch (whichDecoder) {
        case DecoderQ:
              return ((WVQDecoder *)Decoder)->Testchar();
        case DecoderB:
        default:
              return ((WVBDecoder *)Decoder)->Testchar();
    }
}


int WVDecoder::Getchar()
{
    switch (whichDecoder) {
        case DecoderQ:
              return ((WVQDecoder *)Decoder)->Getchar();
        case DecoderB:
        default:
              return ((WVBDecoder *)Decoder)->Getchar();
    }
}

void WVDecoder::Ungetchar(const TCHAR ch)
{
    switch (whichDecoder) {
        case DecoderQ:
              ((WVQDecoder *)Decoder)->Ungetchar(ch);
              break;
        case DecoderB:
        default:
              ((WVBDecoder *)Decoder)->Ungetchar(ch);
    }
}

unsigned int WVDecoder::Pendingchars()
{
    switch (whichDecoder) {
        case DecoderQ:
              return ((WVQDecoder *)Decoder)->Pendingchars();
        case DecoderB:
        default:
              return ((WVBDecoder *)Decoder)->Pendingchars();
    }
}

unsigned int WVDecoder::NonASCIIchars()
{
    switch (whichDecoder) {
        case DecoderQ:
              return ((WVQDecoder *)Decoder)->NonASCIIchars();
        case DecoderB:
        default:
              return ((WVBDecoder *)Decoder)->NonASCIIchars();
    }
}

BOOL WVDecoder::Getallowed()
{
    switch (whichDecoder) {
        case DecoderQ:
              return ((WVQDecoder *)Decoder)->Getallowed();
        case DecoderB:
        default:
              return ((WVBDecoder *)Decoder)->Getallowed();
    }
}


/*******************************************************************************************/
#ifdef JAPANESE

#ifdef UNICODE
WVCharsetDecoder::WVCharsetDecoder(int charset, int coding, unsigned int octets)
{
   WVCharsetDecoder::octets = octets;
   WVCharsetDecoder::charset = charset;
   WVCharsetDecoder::coding = coding;
   state = prevstate = 0;
   BDecoder = NULL;
   uniindex = mbsindex = 0;
}
#endif

WVCharsetDecoder::WVCharsetDecoder(int charset, int coding)
{
#ifdef UNICODE
   WVCharsetDecoder::octets = octets;
#endif
   WVCharsetDecoder::charset = charset;
   WVCharsetDecoder::coding = coding;
   state = prevstate = 0;
   BDecoder = NULL;
   uniindex = mbsindex = 0;
}

WVCharsetDecoder::~WVCharsetDecoder()
{
   delete BDecoder;
}

void WVCharsetDecoder::SwitchStream(int charset, int coding)
{
   if( charset != WVCharsetDecoder::charset
    || coding  != WVCharsetDecoder::coding ) {
     EndOfStream();
     WVCharsetDecoder::charset = charset;
     WVCharsetDecoder::coding  = coding;
     delete BDecoder;
     BDecoder = NULL;
     state = prevstate = 0;
     uniindex = mbsindex = 0;
   }
}

enum s_iso2022 {
    s_roman, s_kanji, s_kanji_2, s_kana,
    s_esc, s_esc_dol, s_esc_dolpar, s_esc_par, s_so
};

void WVCharsetDecoder::Putchar(const TCHAR ch)
{
  unsigned char highbyte, lowbyte;

  switch( coding ) {
  case WVCDS_ISO2022:
    if (ch == 0x0e) {
      prevstate = state;
      state = s_so;
      return;
    }

    switch (state) {
    case s_roman:
      if( ch == 0x1b ) {
        prevstate = state;
        state = s_esc;
        break;
      }
      Thruchar(ch);
      break;

    case s_so:
      if( ch == 0x0f ) {
        state = prevstate;
        break;
      }
      Thruchar(ch | 0x80);
      break;

    case s_kana:
      if( ch == 0x1b ) {
        prevstate = state;
        state = s_esc;
        break;
      }
      Thruchar(ch | 0x80);
      break;

    case s_kanji:
      if( ch == 0x1b ) {
        prevstate = state;
        state = s_esc;
        break;
      }
      if( ch < '!' ) {
        state = s_roman;
        break;
      }
      inbuf[0] = ch;
      state = s_kanji_2;
      break;

    case s_kanji_2:
      if( ch < '!' ) {
        state = s_roman;
        Thruchar( inbuf[0] );
        break;
      }

      state = s_kanji;
      highbyte = ((inbuf[0] - 0x21) >> 1) + 0x81;
      if (highbyte >= 0xa0) {
        highbyte += 0x40;
      }

      if (inbuf[0] & 1) {
        lowbyte = ch + 0x1f;
        if (lowbyte >= 0x7f) lowbyte ++;
        else
        if (lowbyte <= 0x3f) lowbyte |= 0x80; //mozilla
      } else {
        lowbyte = ch + 0x7e;
      }

      Thruchar( highbyte );
      Thruchar( lowbyte );
      break;

    case s_esc:
      if( ch == '$' ) {
        state = s_esc_dol;
      } else if( ch == '(' ) {
        state = s_esc_par;
      } else {
        state = prevstate;
        Thruchar( 0x1b );
        Thruchar( ch );
        break;
      }
      break;

    case s_esc_dol:
      if( ch == '@' || ch == 'B' ) {
        state = s_kanji;
      } else {
        state = prevstate;
        Thruchar( 0x1b );
        Thruchar( '$' );
        Thruchar( ch );
        break;
      }
      return;

    case s_esc_par:
      if( ch == 'B' || ch == 'J' ) {
        state = s_roman;
      } else if( ch == 'I' ) {
        state = s_kana;
      } else {
        state = prevstate;
        Thruchar( 0x1b );
        Thruchar( '(' );
        Thruchar( ch );
        break;
      }
      return;
    }
    return;

  case WVCDS_UTF7:
    if( !state ) {
      if( ch != '+' ) {
        break;
      }
      state = 1;
      if (!BDecoder) BDecoder = new WVBDecoder;
            else     BDecoder->DeleteAll();
    } else if( state == 1 && ch == '-' ) {
      Thruchar('+');
    } else {
      state = 2;
      BDecoder->Putchar( ch );
      if( BDecoder->Getallowed() ) {
        PutHalfwchar( BDecoder->Getchar() );
      }
    }
    return;

  case WVCDS_UTF8:
    if( !state ) {
      if( (ch & 0x80) && (ch & 0x40) ) {
        if( ch & 0x20 ) {
          if( ch & 0x10 ) {
            if( ch & 0x08 ) {
              if( ch & 0x04 ) {
                inbuf[state = 5] = ch & 0x01;
              } else {
                inbuf[state = 4] = ch & 0x03;
              }
            } else {
              inbuf[state = 3] = ch & 0x07;
            }
          } else {
            inbuf[state = 2] = ch & 0x0f;
          }
        } else {
          inbuf[state = 1] = ch & 0x1f;
        }
      } else {
        break;
      }
    } else {
      inbuf[--state] = ch & 0x3f;
      if( state == 0 ) {
        /* UCS4.0 (inbuf[5] << 6) | (inbuf[4])
           UCS4.1 (inbuf[3] << 2) | (inbuf[2]>>4)
           UCS4.2 (inbuf[2] << 4) | (inbuf[1]>>2)
           UCS4.3 (inbuf[1] << 6) | (inbuf[0])       */
        PutHalfwchar( (TCHAR)( (inbuf[2] << 4)|(inbuf[1] >> 2) ) );
        PutHalfwchar( (TCHAR)( (inbuf[1] << 6)|(inbuf[0])      ) );
      }
    }
    return;

  case WVCDS_IGNORECTRL:
    if (ch < 0x20)
      return;
    break;

  case WVCDS_THRU:
    break;
  }
  Thruchar(ch);
}

void WVCharsetDecoder::Flushwcache()
{
#ifdef UNICODE
#else
  if( 0 < uniindex ) {
    int x = WVUniToMulti(mbscache, sizeof(mbscache), (wchar_t*)unicache, uniindex / 2, CP_ACP );
    for( int i = 0; i < x; ++ i ) {
      Thruchar( mbscache[i] );
    }
    uniindex = 0;
  }
#endif
}

void WVCharsetDecoder::PutHalfwchar(TCHAR wch)
{
#ifdef UNICODE
  Thruchar( wch );
#else
  if( uniindex & 1 ) {
    unicache[ uniindex ] = 0;
    unicache[ uniindex - 1 ] = wch;
  } else {
    unicache[ uniindex + 1 ] = wch;
  }
  ++ uniindex; 
  if( uniindex >= sizeof(unicache) ) {
    Flushwcache();
  }
#endif
}

void WVCharsetDecoder::EndOfStream()
{
  switch( coding ) {
  case WVCDS_UTF7:
    if( state ) {
      BDecoder->EndOfStream();
      while( BDecoder->Getallowed() ) {
        PutHalfwchar( BDecoder->Getchar() );
      }
      BDecoder->DeleteAll();
      Flushwcache();
    }
    break;
  case WVCDS_UTF8:
    Flushwcache();
    state = 0;
    break;
  }
}

/*******************************************************************************************/

#endif // JAPANESE


void WVRFCDecoder::Init()
{
    strictRFC = TRUE;
    status = WVC_RFC_ENCODING_ALLOWED;
    CharsetPtr = Charsetname;
    CoderPtr = Codername;
    *CharsetPtr = *CoderPtr = '\0';

    original = NULL;
    charsets = NULL;
    RegisterCharsetName("US-ASCII"); /* this characterset is always valid uppercase !! */
#ifdef JAPANESE
    foldingStatus = 0;
    RegisterCharsetName("ISO-2022-JP");
    charsets->coding = WVCDS_ISO2022;
    charsets->codepage = OSCP_JAP_SJIS;
#endif
}

void WVRFCDecoder::RegisterCharsetName(const char * str)
{
    WVC_CharsetList * p;
    unsigned int len;

    p = new WVC_CharsetList;
    len = strlen(str);
#ifdef UNICODE    
    p->octetwidth = 1;
#endif    
    p->name = new char[len+1];
    strcpy(p->name, str);
#ifdef JAPANESE
    p->coding = WVCDS_IGNORECTRL;
    p->codepage = OSCP_THRU;
#endif
    p->next = charsets;
    charsets = p;
}

#ifdef UNICODE
void WVRFCDecoder::RegisterCharsetName(const char * str, unsigned int octets)
{
    WVC_CharsetList * p;
    unsigned int len;

    if ((octets < 1) || (octets > 2))
        octets = 1;

    p = new WVC_CharsetList;
    len = strlen(str);
    p->octetwidth = octets;
    p->name = new char[len+1];
    strcpy(p->name, str);
    p->next = charsets;
#ifdef JAPANESE
    p->coding = WVCDS_IGNORECTRL;
    p->charset = OSCP_THRU;
#endif
    charsets = p;
}

void WVRFCDecoder::RegisterCharsetName(wchar_t * str)
{
    char mybuf[WVC_CHARSETNAMELENGTH];

    wcsntostr(mybuf,str,WVC_CHARSETNAMELENGTH);
    RegisterCharsetName(mybuf);
}

void WVRFCDecoder::RegisterCharsetName(wchar_t * str, unsigned int octets)
{
    char mybuf[WVC_CHARSETNAMELENGTH];

    wcsntostr(mybuf,str,WVC_CHARSETNAMELENGTH);
    RegisterCharsetName(mybuf,octets);
}
#endif

WVRFCDecoder::WVRFCDecoder()
{
    Init();
}


WVRFCDecoder::WVRFCDecoder(const BOOL RFCstrictly)
{
    Init();
    strictRFC = RFCstrictly;
}

WVRFCDecoder::WVRFCDecoder(const char * s)
{
    Init();
    RegisterCharsetName(s);
}

WVRFCDecoder::WVRFCDecoder(const BOOL RFCstrictly, const char * s)
{
    Init();
    strictRFC = RFCstrictly;
    RegisterCharsetName(s);
}

#ifdef UNICODE
WVRFCDecoder::WVRFCDecoder(const char * s, unsigned int octets)
{
    Init();
    RegisterCharsetName(s,octets);
}

WVRFCDecoder::WVRFCDecoder(const BOOL RFCstrictly, const char * s, unsigned int octets)
{
    Init();
    strictRFC = RFCstrictly;
    RegisterCharsetName(s,octets);
}

WVRFCDecoder::WVRFCDecoder(wchar_t * s)
{
    Init();
    RegisterCharsetName(s);
}

WVRFCDecoder::WVRFCDecoder(wchar_t * s, unsigned int octets)
{
    Init();
    RegisterCharsetName(s,octets);
}

WVRFCDecoder::WVRFCDecoder(const BOOL RFCstrictly, wchar_t * s)
{
    Init();
    strictRFC = RFCstrictly;
    RegisterCharsetName(s);
}

WVRFCDecoder::WVRFCDecoder(const BOOL RFCstrictly, wchar_t * s, unsigned int octets)
{
    Init();
    strictRFC = RFCstrictly;
    RegisterCharsetName(s,octets);
}
#endif //UNICODE

WVRFCDecoder::~WVRFCDecoder()
{
    WVC_CharsetList * p;
    WVC_CharsetList * nxt;

    p = charsets;
    while ( p != NULL ) {
        nxt = p->next;
        delete p->name;
        delete p;
        p = nxt;
    }
    if (original != NULL)
        delete original;
}

void WVRFCDecoder::ChainSetStatus(const TCHAR ch)
{
#ifdef JAPANESE
    CharsetDecoder.Putchar(ch);
    if (CharsetDecoder.Getallowed()) {
      Chain( (TCHAR)CharsetDecoder.Getchar() );
    }
#else
    Chain(ch); /* append character which causes abort past regenerated sequence */
#endif
    status = WVC_US_ASCII_TEXT;
    if ((ch == ' ') || (ch == '(') || (ch == TAB))
       status = WVC_RFC_ENCODING_ALLOWED;
}

void WVRFCDecoder::abort()
{
    /* regenerate input if syntax errors in encoding */
    if ((status == WVC_US_ASCII_TEXT) || (status == WVC_RFC_ENCODING_ALLOWED)) {
        return;  
    };
    Chain('='); /*  after = */
    if (status == WVC_POSSIBLE_RFC_ENCODING) {
        return;
    };
    Chain('?');
    for (CharsetPtr = Charsetname; *CharsetPtr; CharsetPtr++)
        Chain(*CharsetPtr);
    if (status == WVC_CHARSET_PART) {
        return;
    };
    Chain('?');
    for (CoderPtr = Codername; *CoderPtr; CoderPtr++)
        Chain(*CoderPtr);
    if (status == WVC_CODING_PART) {
        return;
    };    
    Chain('?');
    if (original != NULL) {
        original->EndOfStream();
        while (original->Getallowed())
           Chain((TCHAR)original->Getchar());
        delete original;
        original = NULL;
    };
    if (status == WVC_TEXT_PART) {
        return;
    }
    Chain('?');
    if (status == WVC_WAS_RFC_ENCODING) {
        return;
    }
    Chain('=');
}

void WVRFCDecoder::Putchar(const TCHAR ch)
{
       
       switch (ch) {
             case '=' :
                          switch (status) {
                              case WVC_POSSIBLE_RFC_ENCODING:
                                       /* sequence = = no RFC encoding */
                                       Chain('='); /* put the first of them in chain */
                              case WVC_US_ASCII_TEXT:
                                       if (strictRFC) {
                                           /* by RFC a encoded word begins at a start
                                           of a field or after a whitespace or ( */
                                           ChainSetStatus('=');
                                           break;
                                       };
                              case WVC_RFC_ENCODING_ALLOWED:    
                                       status = WVC_POSSIBLE_RFC_ENCODING;
                                       break;
                                       
                              case WVC_WAS_RFC_ENCODING:
                                       status = WVC_SUCCESSFULL_RFC_ENCODING;
                                       if (!strictRFC) {
                                           /* OK it was a RFC encoding, get all out of
                                           the Decoder, which it contains */
                                #ifdef JAPANESE
                                           while (Decoder->Getallowed()) {
                                                  TCHAR ch = Decoder->Getchar();
                                                  CharsetDecoder.Putchar(ch);
                                                  if (CharsetDecoder.Getallowed())
                                                    Chain( (TCHAR)CharsetDecoder.Getchar() );
                                           }
                                           CharsetDecoder.EndOfStream();
                                           while( CharsetDecoder.Getallowed() ) {
                                             Chain( (TCHAR)CharsetDecoder.Getchar() );
                                           }
                                #else
                                           while (Decoder->Getallowed()) {
                                                  Chain((TCHAR)Decoder->Getchar());
                                           }
                                #endif
                                           delete Decoder;
                                           status = WVC_US_ASCII_TEXT;
                                       }
                                       break;
                              case WVC_TEXT_PART:
                                       original->Putchar(ch);
                                       Decoder->Putchar(ch);
                                       break;
                              case WVC_CHARSET_PART:
                              case WVC_CODING_PART:
                              case WVC_SUCCESSFULL_RFC_ENCODING:
                                       abort();
                                       ChainSetStatus(ch);
                          };
                          break;
             case '?' :
                          switch (status) {
                              case WVC_US_ASCII_TEXT:
                              case WVC_RFC_ENCODING_ALLOWED:
                                       ChainSetStatus(ch);
                                       break;
                              case WVC_POSSIBLE_RFC_ENCODING:
                                       status = WVC_CHARSET_PART;
                                       CharsetPtr = Charsetname;
                                       *CharsetPtr = '\0';
                                       CharsetnameLen = 0;
                                       break;
                              case WVC_CHARSET_PART:
                                       if (Charsetname[0] == '\0') {
                                           abort();
                                           ChainSetStatus(ch);
                                           break;
                                       };
                                       status = WVC_CODING_PART;
                                       CoderPtr = Codername;
                                       *CoderPtr = '\0';
                                       CodernameLen = 0;
                                       break;
                              case WVC_CODING_PART:
                                       if (Codername[0] == '\0') {
                                           abort();
                                           ChainSetStatus(ch);
                                           break;
                                       };

                                       if (original != NULL ) /* delete old FIFO if exist */
                                          delete original;
                                       original = new FIFO;
                                       
                                       /* now Test what charset we should use */
                                       actualcharset = charsets;
                                       while (actualcharset != NULL) {
                                           if (!_stricmp(actualcharset->name, Charsetname))
                                               break;
                                           actualcharset = actualcharset->next;
                                       };
                                       if (actualcharset == NULL) {
                                           /* no registered charset fit, so  we abort */
                                           abort();
                                           ChainSetStatus(ch);
                                           break;
                                       };
#ifdef JAPANESE
                                       CharsetDecoder.SwitchStream( actualcharset->codepage, actualcharset->coding );
                                       while( CharsetDecoder.Getallowed() ) {
                                         Chain( (TCHAR)CharsetDecoder.Getchar() );
                                       }
#endif

#ifdef UNICODE                                 
                                       Decoder = new WVDecoder(Codername,
                                                      actualcharset->octetwidth);
#else
                                       Decoder = new WVDecoder(Codername);
#endif
                                       status = WVC_TEXT_PART;               
                                       break;
                              case WVC_TEXT_PART:
                                       if (original->Pendingchars() == 0) {
                                           abort();
                                           ChainSetStatus(ch);
                                           break;
                                       };
                                       Decoder->EndOfStream();
                                       
                                       status = WVC_WAS_RFC_ENCODING;
                                       break;
                              case WVC_WAS_RFC_ENCODING:
                              case WVC_SUCCESSFULL_RFC_ENCODING:
                                       abort();
                                       ChainSetStatus(ch);
                          };
                          break;
                          
             case '(':
             case ')':
             case '<':
             case '>':
             case '@':
             case ',':
             case ';':
             case ':':
             case '"':
             case '/':
             case '[':
             case ']':
             case '.':    
                          /* these characters are defined in RFC 1522 as especials */
                          /* and not allowed in Characterset-Name or Name of Encoder */
                          /* in Encoded Text OK by syntax, but maybe not usefull */
                          if (status == WVC_TEXT_PART) {
                              if (original)
                                  original->Putchar(ch);
                              Decoder->Putchar(ch);
                              break;
                          };
                          if ((status == WVC_SUCCESSFULL_RFC_ENCODING) && (ch != ')')) {
                              abort();
                              ChainSetStatus(ch);
                              break;
                          };
             case 0x00:
             case CR:
             case LF:
             case TAB:             
             case ' ':
                          switch (status) {
                              case WVC_SUCCESSFULL_RFC_ENCODING:
                                      while (Decoder->Getallowed()) {
                                             Chain((TCHAR)Decoder->Getchar());
                                      }
                                      delete Decoder;
                              case WVC_RFC_ENCODING_ALLOWED:
                              case WVC_US_ASCII_TEXT:            
                                      ChainSetStatus(ch);
                                      break;
                                      
                              case WVC_WAS_RFC_ENCODING:
                              case WVC_POSSIBLE_RFC_ENCODING:
                              case WVC_CHARSET_PART:
                              case WVC_CODING_PART:
                              case WVC_TEXT_PART:
                                      abort();
                                      ChainSetStatus(ch);
                          };
                          break;   
             default:
                          switch (status) {
                              case WVC_CHARSET_PART:
                                       if (_istascii(ch) &&
                                           (++CharsetnameLen < WVC_CHARSETNAMELENGTH)) {
                                            *CharsetPtr++ = (char)ch;
                                            *CharsetPtr = '\0';
                                       } else {
                                             abort();
                                             ChainSetStatus(ch);
                                       }
                                       break;
                              case WVC_CODING_PART:
                                       if (_istascii(ch) &&
                                           (++CodernameLen < WVC_CODERNAMELENGTH)) {
                                            *CoderPtr++ = (char)ch;
                                            *CoderPtr = '\0';
                                       } else {
                                            abort();
                                            ChainSetStatus(ch);
                                       }
                                       break;

                              case WVC_TEXT_PART:
                                       if (original)
                                           original->Putchar(ch);
                                       Decoder->Putchar(ch);
                                       break;
                              case WVC_POSSIBLE_RFC_ENCODING:
                              case WVC_RFC_ENCODING_ALLOWED:
                              case WVC_US_ASCII_TEXT:
                              case WVC_WAS_RFC_ENCODING:
                              case WVC_SUCCESSFULL_RFC_ENCODING:
                                       abort();
                                       ChainSetStatus(ch);
                          };
       } /* switch (ch) */
}


void WVRFCDecoder::EndOfStream()
{
    
    switch (status) {
        case WVC_SUCCESSFULL_RFC_ENCODING:
                Decoder->EndOfStream();
                while (Decoder->Getallowed()) {
                   Chain((TCHAR)Decoder->Getchar());
                }
                delete Decoder;                          
        case WVC_RFC_ENCODING_ALLOWED:
        case WVC_US_ASCII_TEXT:            
                break;
                                      
        case WVC_WAS_RFC_ENCODING:
        case WVC_POSSIBLE_RFC_ENCODING:
        case WVC_CHARSET_PART:
        case WVC_CODING_PART:
        case WVC_TEXT_PART:
                abort();
    };
    RFCFIFO::EndOfStream();
}

/************************************************************************************************************************/


WVC_prequeue::WVC_prequeue()
{
    next = NULL;
    Specialcharacters = 0;
}

WVC_prequeue::~WVC_prequeue()
{
}

void WVC_prequeue::setnextword(WVC_prequeue * nextword)
{
    next = nextword;
}

WVC_prequeue * WVC_prequeue::getnextword()
{
    return next;
}


unsigned int WVC_prequeue::GetSpecialchars()
{
    return Specialcharacters;
}


void WVC_prequeue::Putchar(const TCHAR ch)
{
    if ((ch == '?') || (ch == '=') || (ch == '_') || (ch == '(') || (ch == ')'))
       Specialcharacters++; 
    Chain(ch);
}

/************************************************************************************************************************/

WVRFCCoder::WVRFCCoder()
{
    strcpy(Charsetname,"US-ASCII");
#ifdef UNICODE    
    octetwidth = 1;
#endif    
    lastchar = ' ';
    prequeue = root = new WVC_prequeue;
}

WVRFCCoder::WVRFCCoder(const char * s)
{
    Charsetname[0] = '\0';
    if (strlen(s) < WVC_CHARSETNAMELENGTH)
       strcpy(Charsetname,s);
#ifdef UNICODE       
    octetwidth = 1;
#endif    
    lastchar = ' ';
    prequeue = root = new WVC_prequeue;
}

#ifdef UNICODE
WVRFCCoder::WVRFCCoder(const char * s, unsigned int octets)
{
    Charsetname[0] = '\0';
    if (strlen(s) < WVC_CHARSETNAMELENGTH)
       strcpy(Charsetname,s);
    octetwidth = octets;
    lastchar = ' ';
    prequeue = root = new WVC_prequeue;
}

WVRFCCoder::WVRFCCoder(wchar_t * s)
{
    Charsetname[0] = '\0';
    wcsntostr(Charsetname, s, WVC_CHARSETNAMELENGTH);
    octetwidth = 1;
    lastchar = ' ';
    prequeue = root = new WVC_prequeue;
}

WVRFCCoder::WVRFCCoder(wchar_t * s, unsigned int octets)
{
    Charsetname[0] = '\0';
    wcsntostr(Charsetname, s, WVC_CHARSETNAMELENGTH);
    octetwidth = octets;
    lastchar = ' ';
    prequeue = root = new WVC_prequeue;
}
#endif

WVRFCCoder::~WVRFCCoder()
{
}


BOOL WVRFCCoder::whitespace(const TCHAR ch)
{
     return ((ch == ' ') || (ch == TAB) || (ch == CR) || (ch == LF) || (ch== 0x00) );
}

void WVRFCCoder::processwords(BOOL mustEncode)
{
    char * s;
    WVC_prequeue * p;
    WVC_prequeue * Qstop;        /* this points to the last word of Q-multiwordencoding */
    WVC_prequeue * Bstop;        /* this points to the last word of B-multiwordencoding */

#ifdef UNICODE
    WVQCoder QCoder(octetwidth);
    WVBCoder BCoder(octetwidth);
#else
    WVQCoder QCoder;
    WVBCoder BCoder;
#endif    

    unsigned int Qlen;           /* used to calculate current length which Q will use */
    unsigned int Blen;           /* used to calculate current length which B will use */
    unsigned int MaxQBlen;       /* max length for Q or B encoding output */
    unsigned int Total;          /* count for all characters in multiwordencoding */
    unsigned int maybeTotal;     /* same if we encounter a word to encode later */
    unsigned int ToEncode;       /* count for characters to encode in multiwordencoding */
    unsigned int maybeToEncode;  /* same if we encounter a word to encode later */
    
    
    p = root;
    do {
        if (p->NonASCIIchars() == 0 ) {
            /* all characters are us-ascii */
            while (p->Getallowed())
              Chain((TCHAR)p->Getchar()); /* put it in the inherited fifo */
            root = p->getnextword();
            delete p;
            p = root;
        }   
    } while ((p != NULL) && (p->NonASCIIchars() == 0));
            /* while a word and and all US-ASCII */
    
    if (root) {
        /* there are words which contains non US-ASCII */
        /* determine max length allowed by RFC */
        /* maxlength includes =? charsetname ? Q|B ? and later ?= */
        Qstop = Bstop = p = root;
        MaxQBlen = WVC_MAX_ENCODING_LENGTH - 2 - strlen(Charsetname) -1 - 1 - 1 - 2;
        Total = ToEncode = maybeTotal = maybeToEncode = 0; 
        do {
           maybeTotal += p->Pendingchars();
           maybeToEncode += p->NonASCIIchars() + p->GetSpecialchars();

#ifdef UNICODE
           if (octetwidth == 1) {
#endif
               /* now calculate if we reached maximum encoding length defined in RFC */
               /* in Q-encoding ascii represents itsself nonascii are 3 bytes */           
               Qlen = maybeTotal + 2 * maybeToEncode;

               Blen = 4 * (maybeTotal / 3 );
               /* if not a integral quantum of 24 bits, add padding chars */
               if ((maybeTotal % 3) !=0 ) 
                  Blen += 4;
#ifdef UNICODE
           } else {
               /* this is only usefull for my additional implementation of encoding UNICODE */
               /* THIS DOESN'T FIT IN ANY RFC ! */
               /* WARNING: to be confirm to RFC never use a other octetwidth than 1 */
               /* now calculate if we reached maximum encoding length defined in RFC */
               /* in Q-encoding ascii represents itsself nonascii are 5 bytes */           

               /* Qlen = maybeTotal + 4 * maybeToEncode; */
               Qlen = MaxQBlen + 1; /* we use only B-encoding yet, so set QLen to a maximum */
               
               Blen = 4 * ((2 * maybeTotal) / 3 );
               /* if not a integral quantum of 24 bits, add padding chars */
               if (((2*maybeTotal) % 3) !=0 ) 
                  Blen += 4;
           }
#endif       
           if (p->NonASCIIchars() > 0) {
              /* this is a word which must encoded */
              if (Qlen < MaxQBlen) {
                  /* we have not too much characters, this word is ok */
                  Total = maybeTotal;
                  ToEncode = maybeToEncode;
                  Qstop = p;
              }
              if (Blen < MaxQBlen) {
                  /* we have not too much characters, this word is ok */
                  Total = maybeTotal;
                  ToEncode = maybeToEncode;
                  Bstop = p;
              }
           }                
           p = p->getnextword();   
        } while ((p != NULL) && ((Blen < MaxQBlen ) || (Qlen < MaxQBlen)));
        
        if ((p!=NULL) || (mustEncode)) {
            /* enough words to encode or we must encode because no more data */
            /* use the encoder which give us a encoded word with smallest length */
            /* this is the one which contains still most characters */
            Chain('=');
            Chain('?');
            s = Charsetname;
            while (*s)
               Chain(*s++);
            Chain('?');
            
            /* now we have got the problem to determine which encoding either Q or B we
            should use, by RFC using of Q is recommended, when  most of the characters
            to be encoded are in ASCII. otherwise B is recommended.
            But there is a problem in this recommendation:
            if 16,66 % or more of all characters are nonASCII Q-encoding results a longer
            encoded word then B-encoding.
            The other side is, that Q-Encoding is more useable to read if a newsreader not
            confirms to rfc1522.
            if ISO-8859-* is assumed we can encode:
            
            20 % non ASCII chars means about 41 chars in Q-encoding 42 chars in B-encoding
            30 % non ASCII chars means about 36 chars in Q-encoding 42 chars in B-encoding
            40 % non ASCII chars means about 32 chars in Q-encoding 42 chars in B-encoding
            50 % non ASCII chars means about 29 chars in Q-encoding 42 chars in B-encoding

            Since a text with 40 % non ASCII is NOT readable by people with a old newsreader
            we should probably decrease from 50 % to 40 % in the future.
            
            Nevertheless, if one of Q or B encoding needs to encode in more then these
            characters which is allowed by RFC ( = 75, including RFC1522 envelope )
            we use the encoding which fits.
            */
            if ((Qlen > MaxQBlen ) || /* Q-encoding doesn't fit */
            
              /* ((Blen <= MaxQBlen ) && ( 5 * ToEncode >= 2 * Total )) ) */ /* 40 % */
                 ((Blen <= MaxQBlen ) && ( 2 * ToEncode >= Total )) ) {  /* 50 % */
        
              Chain('B');
              Chain('?');
              do {
                 p = root;
                 while (p->Getallowed()) {
                    BCoder.Putchar((TCHAR)p->Getchar());
                 }
                 root = p->getnextword();
                 delete p;    
              } while ((root != NULL) && (p != Bstop));
              BCoder.EndOfStream();
              while (BCoder.Getallowed())
                  Chain((TCHAR)BCoder.Getchar());
            } else {
              Chain('Q');
              Chain('?');
              do {
                 p = root;
                 while (p->Getallowed()) {
                    QCoder.Putchar((TCHAR)p->Getchar());
                 }
                 root = p->getnextword();
                 delete p;    
              } while ((root != NULL) && (p != Qstop));
              QCoder.EndOfStream();
              while (QCoder.Getallowed())
                  Chain((TCHAR)QCoder.Getchar());
            }
                        
            Chain('?');
            Chain('=');
        } 
    } else {
        prequeue = root; /* there is nothing in chain */
    }    
}

void WVRFCCoder::Putchar(const TCHAR ch)
{
    WVC_prequeue * p;
    BOOL wordboundary;

    wordboundary = FALSE;
    if (whitespace(ch) && !whitespace(lastchar))
           wordboundary = TRUE;
    if (!whitespace(ch) && whitespace(lastchar))
           wordboundary = TRUE;
           
    /* ( and ) stands always alone as a word */
    if ((ch == ')') || (ch == '('))
           wordboundary = TRUE;
    if ((lastchar == ')') || (lastchar == '('))
           wordboundary = TRUE;
           
    if (wordboundary) {
        prequeue->EndOfStream(); /* say FIFO for this word it was all */
        processwords(FALSE); /*check if we can send some characters */
        /* maybe the queue is empty */
        
        if (root != NULL) {
           p = new WVC_prequeue; /* make ready for the next word */
           prequeue->setnextword(p);
           prequeue = p;
        } else {
           prequeue = root = new WVC_prequeue;
        }
    }
    prequeue->Putchar(ch);
    lastchar = ch;
}

void WVRFCCoder::EndOfStream()
{
    while (root != NULL)
       processwords(TRUE);
    FIFO::EndOfStream();
}


/******************************************************************************************/
/* Only internal, don't use it directly, use encodeHeader instead                         */
/******************************************************************************************/

#ifdef UNICODE
void encodeRFC1522W(WVString &str)
{
    WVRFCCoder Coder(L"ISO-8859-1",1);

    const wchar_t *p;
    wchar_t lastchar;

    WVString tmp;
    int idx;

    p = (const wchar_t *)str;

    idx = 0;

    while ( *p ) {
        if (Coder.Putallowed())
            Coder.Putchar(*p++);
        if (Coder.Getallowed()) {
            lastchar = (wchar_t)Coder.Getchar();
	    if (lastchar) {
		tmp += L" "; /* add one blank to get a byte in memory */
		tmp[idx++] = lastchar; /* and replace it with the current char */
	    }
	}
    };
    Coder.EndOfStream();
    while (Coder.Getallowed()) {
        lastchar = (wchar_t)Coder.Getchar();
        if (lastchar) {
		tmp += L" "; /* add one blank */
		tmp[idx++] = lastchar; /* and replace it */
	}
    }
    str = tmp;

}
#endif

void encodeRFC1522A(WVString &str)
{
    WVRFCCoder Coder("ISO-8859-1");

    const char *p;
    char lastchar;

    WVString tmp;
    int idx;

    p = (const char *)str;

    idx = 0;

    while ( *p ) {
        if (Coder.Putallowed())
            Coder.Putchar(*p++);
        if (Coder.Getallowed()) {
            lastchar = (char)Coder.Getchar();
	    if (lastchar) {
		tmp += " "; /* add one blank to get a byte in memory */
		tmp[idx++] = lastchar; /* and replace it with the current char */
	    }
	}
    };
    Coder.EndOfStream();
    while (Coder.Getallowed()) {
        lastchar = (char)Coder.Getchar();
        if (lastchar) {
		tmp += " "; /* add one blank */
		tmp[idx++] = lastchar; /* and replace it */
	}
    }
    str = tmp;
}


/************************************************************************************************************************/

#ifdef UNICODE
void encodeHeaderW(WVString &str)
{

    if (EncodeRFC1522)
        encodeRFC1522W(str);
}
#endif

void encodeHeaderA(WVString &str)
{

    if (EncodeRFC1522)
        encodeRFC1522A(str);
}

#ifdef UNICODE
void decodeRFC1522W(WVString &str)
{
    WVRFCDecoder Decoder(FALSE,"ISO-8859-1");
    
    const wchar_t *p;
    wchar_t *o;
    wchar_t lastchar;

    if (UseRFC1522) {
        Decoder.RegisterCharsetName("UNICODE-1-1",2);
        p = (const wchar_t *)str;
        o = (wchar_t *)p;
        while ( *p ) {
            if (Decoder.Putallowed())
                Decoder.Putchar(*p++);
            if (Decoder.Getallowed() && (o<p))
                /* there is a character and we have a byte where we can store it */
                *o++ = lastchar = (wchar_t)Decoder.Getchar();
        };
        Decoder.EndOfStream();
        while (Decoder.Getallowed())
            *o++ = lastchar = (wchar_t)Decoder.Getchar();
        if (lastchar)    
            *o = '\0'; /* make shure there is a \0 on the end of the string */
    }
}
#endif

void decodeRFC1522A(WVString &str)
{
    WVRFCDecoder Decoder(FALSE,"ISO-8859-1");

    const char *p;
    char *o;
    char lastchar ='\0';

    if (UseRFC1522) {
        p = (const char *)str;
        o = (char * )p;
        while ( *p ) {
            if (Decoder.Putallowed())
                Decoder.Putchar(*p++);
            if (Decoder.Getallowed() && (o<p))
                /* there is a character and we have a byte where we can store it */
                *o++ = lastchar = (char)Decoder.Getchar();
        };
        Decoder.EndOfStream();
        while (Decoder.Getallowed())
            *o++ = lastchar = (char)Decoder.Getchar();
        if (lastchar)    
            *o = '\0'; /* make shure there is a \0 on the end of the string */
    }
}


/************************************************************************************************************************/

#ifdef UNICODE
void encodeStrHeaderW(wchar_t * s)
{
    WVString str;

    if (EncodeRFC1522) {
        str = s;
        encodeRFC1522W(str);
        strcpy(s,(const wchar_t *) str);
    }
}
#endif

void encodeStrHeaderA(char * s)
{
    WVString str;

    if (EncodeRFC1522) {
        str = s;
        encodeRFC1522A(str);
        strcpy(s,(const char *) str);
    }
}

#ifdef UNICODE
void decodeStrRFC1522W(wchar_t * s)
{
    WVRFCDecoder Decoder(FALSE,"ISO-8859-1");
    
    const wchar_t *p;
    wchar_t *o;
    wchar_t lastchar;

    if (UseRFC1522) {
        Decoder.RegisterCharsetName("UNICODE-1-1",2);
        o = p = s;
        while ( *p ) {
            if (Decoder.Putallowed())
                Decoder.Putchar(*p++);
            if (Decoder.Getallowed() && (o<p))
                /* there is a character and we have a byte where we can store it */
                *o++ = lastchar = (wchar_t)Decoder.Getchar();
        };
        Decoder.EndOfStream();
        while (Decoder.Getallowed())
            *o++ = lastchar = (wchar_t)Decoder.Getchar();
        if (lastchar)    
            *o = '\0'; /* make shure there is a \0 on the end of the string */
    }
}
#endif

void decodeStrRFC1522A(char * s)
{
    WVRFCDecoder Decoder(FALSE,"ISO-8859-1");

    char *p;
    char *o;
#ifdef JAPANESE
    char lastchar = !0;
#else
    char lastchar='\0';
#endif


    if (UseRFC1522) {
        o = p = s;
        while ( *p ) {
            if (Decoder.Putallowed())
                Decoder.Putchar(*p++);
            if (Decoder.Getallowed() && (o<p))
                /* there is a character and we have a byte where we can store it */
                *o++ = lastchar = (char)Decoder.Getchar();
        };
        Decoder.EndOfStream();
        while (Decoder.Getallowed())
            *o++ = lastchar = (char)Decoder.Getchar();
        if (lastchar)    
            *o = '\0'; /* make shure there is a \0 on the end of the string */
    }
}


#ifdef JAPANESE

BOOL // FALSE- no more
decodeMLStrRFC1522(char *s) // MultiLine Decode wrapper
{
    static WVRFCDecoder *Decoder = NULL;
    static FIFO *Pending = NULL;
    static int FlushMode = 0; // 0:NO 1:TONEXTLINE 2:ALL -1:ERR
    int Pendingchars;

    if (!UseRFC1522) {
      return *s;
    }

    if (s == MLTLN_INIT) {
      delete Decoder;
      Decoder = NULL;
      delete Pending;
      Pending = NULL;
      FlushMode = 0;
      return 0;
    }

    if (!Decoder && FlushMode != -1) {
      Decoder = new WVRFCDecoder(FALSE, "ISO-8859-1");
      Pending = new FIFO;
      if (!Decoder || !Pending) {
        delete Decoder;
        Decoder = NULL;
        delete Pending;
        Pending = NULL;
        FlushMode = -1;
      }
    }
    if (FlushMode == -1)
      return *s;
    else if (*s == '.' && *(s+1) == '\0')
      FlushMode = 2;
    else if (FlushMode == 0 && (*s != ' ' && *s != '\t'))
      FlushMode = 1;

    Pendingchars = Pending->Pendingchars();

    if (FlushMode != 2) {
      char *p = s;
      while (*p) {
        if (Pending->Putallowed())
          Pending->Putchar(*p);
        ++ p;
      }
    }

    if (FlushMode) {
      char *o = s;
      char *f = s + sizeof(CommLineIn) - 2;
      char lastchar = !0;
      char ch = 0;

      while( Pending->Getallowed() && Pendingchars -- ) {
        ch = Pending->Getchar();
        if (Decoder->Putallowed()) {
          Decoder->Putchar(ch);
        }
        if (Decoder->Getallowed()) {
          ch = Decoder->Getchar();
          if (ch == '\n') {
            Decoder->Ungetchar(ch);
            break;
          }
          if (o < f) {
            *o ++ = lastchar = ch;
          }
        }
      }
      Decoder->EndOfStream();
      while( Decoder->Getallowed() ) {
        ch = Decoder->Getchar();
        if (ch == '\n') {
          //Decoder->Ungetchar(ch);
          break;
        }
        if (o < f)
          *o ++ = lastchar = ch;
      }
      if (lastchar) *o = '\0';
      if (Decoder->Getallowed() == NULL && FlushMode == 1)
        FlushMode = 0;
    } else {
      *s = 0;
    }
    return *s;
}

void decodeNetworkCoding(char *p, int coding)
{
    WVCharsetDecoder Decoder;
    char *o = p;
    char lastchar = !0;
    while ( *p ) {
      if (Decoder.Putallowed()) {
        Decoder.Putchar(*p);
      }
      ++ p;
      if (Decoder.Getallowed() && o < p ) {
        *o ++ = lastchar = Decoder.Getchar();
      }
    }
    Decoder.EndOfStream();
    while (Decoder.Getallowed()) {
      *o ++ = lastchar = Decoder.Getchar();
    }
    if (lastchar) *o = '\0';
}

void encodeNetworkCoding(char *p, int coding)
{
}

int my_wtos(char *out, int out_length, wchar_t *in, int in_length)
{
  while( in_length && out_length ) {
    *out ++ = *in ++;
    in_length --;
    out_length --;
  }
  *out = '\0';
  return in_length;
}

int my_stow(wchar_t *out, int out_length, char *in, int in_length)
{
  while( in_length && out_length ) {
    *out ++ = *in ++;
    in_length --;
    out_length --;
  }
  *out = '\0';
  return in_length;
}

#ifdef _WIN32
 int system_codepage;
#else
 //TODO: support external dll.
 #define system_codepage      (-1)
 #define IsValidCodePage(cp)  (FALSE)
#endif

#ifdef _MSC_VER
  #define AllocA(ptr, size) \
       char *ptr = _alloca(size)
#else
  class stackcompat {
  public:
    stackcompat(size_t size);
    ~stackcompat();
    char *get();
  protected:
    char *ptr;
  } ALIGNED;
  stackcompat::stackcompat(size_t size)
  {
    ptr = new char[size+1];
  }
  stackcompat::~stackcompat()
  {
    delete[] ptr;
  }
  char *stackcompat::get()
  {
    return ptr;
  }
  #define AllocA(ptr, size) \
       stackcompat alloca##ptr(size); char *ptr = alloca##ptr.get();
#endif
  #define FUNCUSE_UNICONV(ptr, len) AllocA(ptr, len)

void WVDecodeCharset(WVString& str, int codepage)
{
#ifdef _WIN32
  if( codepage == system_codepage || !IsValidCodePage(codepage) )
#endif
  {
    return;
  }
#ifdef _WIN32
  const char *p = (const char *)str;
  int ilen = strlen(p);
  int olen;
  AllocA( utmp, ilen * 3 );  /* MultiByte1~2 -> Unicode1(*2) */
  olen = WVMultiToUni( (wchar_t*)utmp, ilen * 3 / 2, (char *)p, ilen, codepage);
  char *mtmp = str.MakeBuffer( ilen * 3 ); /* Mb 1~2 -> Mb 1~2 */
  WVUniToMulti(mtmp, ilen * 3, (wchar_t*)utmp, olen, CP_ACP);
#endif
}

int WVMultiToUni(wchar_t *out, int out_length, char *in, int in_length, int codepage)
{
#ifdef _WIN32
  int i = MultiByteToWideChar(codepage, 0, in, in_length, out, out_length);
  out[i] = '\0';
  return i;
#endif
  return my_stow(out, out_length, in, in_length);
}

int WVUniToMulti(char* out, int out_length, wchar_t *in, int in_length, int codepage)
{
#ifdef _WIN32
  BOOL DefaultUsed;
  int i = WideCharToMultiByte(codepage, 0, in, in_length, out, out_length, NULL, &DefaultUsed);
  out[i] = '\0';
  return i;
#endif
  return my_wtos(out, out_length, in, in_length);
}

#endif // JAPANESE

#endif // RFC1522


#ifdef STANDALONE


main()
{
    char line[256];
    wchar_t wline[256];

    
    strcpy(line,"=?ISO-8859-1?Q?Kristian_K=F6hntop?=");
    strtowcs(wline,line);
    
    decodeRFC1522W(wline, 256 );
    
    WVMessageBox ( NULL,wline, L"Unicode Test", MB_OK | MB_ICONSTOP);
    
    strcpy(line,"=?unicode-1-1?b?IVoA1yFWAD0hUw==?= (was: ass)");
    strtowcs(wline,line);
    
    decodeRFC1522W(wline, 256);
    encodeHeaderW(wline, 256);
        
    WVMessageBox ( NULL,wline, L"Unicode Test", MB_OK | MB_ICONSTOP);
}
#endif
