// -*- C++ -*-
/*
 * $Id: wvutil.cpp 1.106 1997/04/05 00:39:18 dumoulin Exp $
 *
 */

/*-- WVUTIL.C -- File containing utility routines.
 */

#include <windows.h>
#include <windowsx.h>           // for GlobalFreePtr (JSC)
#include "wvglob.h"
#include "winvn.h"
#include "xface.h"
#include "wvimage.h"
#pragma hdrstop

#ifndef _WIN32
#include <dos.h>                // for _dos_getdiskfree  
#include <ver.h>

#else

#include <winver.h>

#endif

#include "wvclass.h"
#include <commdlg.h>            // for GetOpenFileName dialog (JSC)
#include <ctype.h>
#include <stdlib.h>
#include <time.h>
#include <io.h>                 // for _open etc
#include <fcntl.h>
#include <shellapi.h>           // for Registry access
#include <stdarg.h>
#define FIX_df__20000427        /* Dirk Fieldhouse XOVER server lies fixes */

char far *mylstrcpy (char_p ptr1, char far * ptr2);
char *get_xhdr_line (char *line);
/* Ray Dunn:99/10: added datetime arg for date sorting */
time_t parse_usenet_date (struct tm *datetime, char *date);
void finish_header_retrieval ();
void GenerateFileFilters (HWND hParentWnd, char *filters);

BOOL TrimHeader (char *header);
void AddCommLineToDoc (char *line);
void WrapAddCommLineToDoc (char *line);
void MakeXFaceWnd(const char * szFace);
void MakeFaceWnd(const char * szFace);
extern BOOL MakeRectVisible(RECT& rRect);



// please update this if you modify XHDR retrieval
// This will now be either 5 or 7, depending on whether threading
// is enabled via 'threadp'.
unsigned int total_xhdrs = 5;

/*--- function GetNum --------------------------------------------
 *
 *  Cracks off a positive integer number from a string.
 *
 *  Entry    *ptr  is the character position to start scanning
 *                 for an integer
 *
 *  Exit     *ptr  is the character position at which we stopped
 *                 scanning (because of a non-digit).
 *           *num  is the cracked off number.
 *           Returns TRUE iff we got a number.
 */
BOOL GetNum (char **ptr, unsigned long *num)
{
	BOOL gotit = FALSE;

	/* Skip initial spaces                                            */

	while ((**ptr) && **ptr == ' ')
		(*ptr)++;

	*num = 0;
	while (**ptr && isdigit (**ptr)) 
	{
		*num = 10 * (*num) + (**ptr - '0');
		gotit = TRUE;
		(*ptr)++;
	}
	return (gotit);
}

char * get_xhdr_line (char *line)
{
	char *cptr;
	/* skip past the art # and space */
	for (cptr = line; isdigit (*cptr); cptr++);
	for (; *cptr == ' '; cptr++);
	return (cptr);
}

#if 0
//MRB already did this
void make_neat_from (char far * in, char far * out)
{
	char far *left, far * right;

	/* this is controlled from .ini */
	if (FullNameFrom) 
	{
		left = strchr (in, '(');
		right = strrchr (in, ')');

		if ((left && right) && (left < right)) 
		{
			strncpy (out, left + 1, (size_t) (right - left - 1));
			out[(right - left - 1)] = (char) 0;
		}
		else                        /* No name in parens */
			strcpy (out, in);
	}
	else                          /* !FullNameFrom */
		strcpy (out, in);
}
#endif

 
// Simple check to see if Username is legal.  This needs
// to be more complicated to handle all possible illegal
// names, but for now, just handle the main ones.
// Return TRUE if valid.   (JD 3/26/97)
//        
BOOL ValidEmailAddress(char far * name)
{
	char far *atsign;

	if (name) 
	{
		atsign = strchr (name, '@');
		if (atsign &&                          // needs an @ sign
			(atsign > name) &&                 // nonblank to left of @
			(strstr (atsign, "@.") == NULL) && // simple typo check
			(strchr (atsign, '.') != NULL) &&  // at least one '.' rt of @
			(strchr (name, '|') == NULL))      // don't allow pipes in name
		{
			return TRUE;
		}
	}
	return FALSE;
}

/*-- function StrToRGB -------------------------------------------------
 *
 *  Takes an ASCII string of the form "r,g,b" where r, g, and b are
 *  decimal ASCII numbers, and converts it to an RGB color number.
 */
COLORREF StrToRGB (char *cstring)
{
	BYTE red, green, blue;
	unsigned long int lred, lgreen, lblue;

	GetNum (&cstring, &lred);
	cstring++;
	GetNum (&cstring, &lgreen);
	cstring++;
	GetNum (&cstring, &lblue);
	red = (BYTE) lred;
	green = (BYTE) lgreen;
	blue = (BYTE) lblue;

	return (RGB (red, green, blue));
}

/*-- function RGBToStr -------------------------------------------------
 *
 *  Takes an RGB color ref and converts to a string of the form "r,g,b"
 *  result is placed in buf
 *  (JSC)
 */

char * RGBToStr (char *buf, COLORREF rgbVal)
{
	_sprintf(buf, "%u,%u,%u", GetRValue (rgbVal),
		   GetGValue (rgbVal),
		   GetBValue (rgbVal));
	return (buf);
}


/* String to Timezone */
time_t dTimeZone(char *s)
{
#define SIZEOF(array)   ((int)(sizeof array / sizeof array[0]))
#define ENDOF(array)    (&array[SIZEOF(array)])
#define HOUR(x)     (x * 60)

typedef struct _TABLE 
{
	char   *name;
	time_t value;
} ALIGNED TABLE;

/* Timezone table. */
static TABLE    TimezoneTable[] = {
    { "gmt",    HOUR( 0) }, /* Greenwich Mean */
    { "ut",     HOUR( 0) }, /* Universal */
    { "utc",    HOUR( 0) }, /* Universal Coordinated */
    { "cut",    HOUR( 0) }, /* Coordinated Universal */
    { "z",      HOUR( 0) }, /* Greenwich Mean */
    { "wet",    HOUR( 0) }, /* Western European */
    { "bst",    HOUR( 0)-60 },  /* British Summer */
    { "nst",    HOUR( 3)+30 }, /* Newfoundland Standard */
    { "ndt",    HOUR( 3)+30-60 }, /* Newfoundland Daylight */
    { "ast",    HOUR( 4) }, /* Atlantic Standard */
    { "adt",    HOUR( 4)-60 },  /* Atlantic Daylight */
    { "est",    HOUR( 5) }, /* Eastern Standard */
    { "edt",    HOUR( 5)-60 },  /* Eastern Daylight */
    { "cst",    HOUR( 6) }, /* Central Standard */
    { "cdt",    HOUR( 6)-60 },  /* Central Daylight */
    { "mst",    HOUR( 7) }, /* Mountain Standard */
    { "mdt",    HOUR( 7)-60 },  /* Mountain Daylight */
    { "pst",    HOUR( 8) }, /* Pacific Standard */
    { "pdt",    HOUR( 8)-60 },  /* Pacific Daylight */
    { "yst",    HOUR( 9) }, /* Yukon Standard */
    { "ydt",    HOUR( 9)-60 },  /* Yukon Daylight */
    { "akst",   HOUR( 9) }, /* Alaska Standard */
    { "akdt",   HOUR( 9)-60 },  /* Alaska Daylight */
    { "hst",    HOUR(10) }, /* Hawaii Standard */
    { "hast",   HOUR(10) }, /* Hawaii-Aleutian Standard */
    { "hadt",   HOUR(10)-60 },  /* Hawaii-Aleutian Daylight */
    { "ces",    -HOUR(1)-60 },  /* Central European Summer */
    { "cest",   -HOUR(1)-60 },  /* Central European Summer */
    { "mez",    -HOUR(1) }, /* Middle European */
    { "mezt",   -HOUR(1)-60 },  /* Middle European Summer */
    { "cet",    -HOUR(1) }, /* Central European */
    { "met",    -HOUR(1) }, /* Middle European */
    { "eet",    -HOUR(2) }, /* Eastern Europe */
    { "msk",    -HOUR(3) }, /* Moscow Winter */
    { "msd",    -HOUR(3)-60 },  /* Moscow Summer */
    { "wast",   -HOUR(8) }, /* West Australian Standard */
    { "wadt",   -HOUR(8)-60 },  /* West Australian Daylight */
    { "hkt",    -HOUR(8) }, /* Hong Kong */
    { "cct",    -HOUR(8) }, /* China Coast */
    { "jst",    -HOUR(9) }, /* Japan Standard */
    { "kst",    -HOUR(9) }, /* Korean Standard */
    { "kdt",    -HOUR(9)-60 },  /* Korean Daylight */
    { "cast",   -(HOUR(9)+30) }, /* Central Australian Standard */
    { "cadt",   -(HOUR(9)+30)-60 }, /* Central Australian Daylight */
    { "east",   -HOUR(10) },    /* Eastern Australian Standard */
    { "eadt",   -HOUR(10)-60 }, /* Eastern Australian Daylight */
    { "nzst",   -HOUR(12) },    /* New Zealand Standard */
    { "nzdt",   -HOUR(12)-60 }, /* New Zealand Daylight */
    { "", 0 }
};
	char *p;
	int tz;

	for (p = s; isalpha((BYTE)(*p)); p++)
		*p = tolower ((BYTE)*p);

	if (isalpha((BYTE)*s)) 
	{
		char cbak;
		TABLE *tp;

		cbak = *p;
		*p = '\0';
		/* Try for a timezone. */
		for (tp = TimezoneTable; tp < ENDOF(TimezoneTable); tp++)   
		{
			if (s[0] == tp->name[0] && s[1] == tp->name[1] && strcmp(s, tp->name) == 0)
				return (tp->value * (-60));
		}
		*p = cbak;
		s = p;
	}

	//handle +-200 tz format
	p = s;
	if (*p == '+')
		p++;

	tz = atoi(p);
	if (tz >= 12 || tz <= -12) 
	{
		int hour, min;
		hour = tz / 100;
		min = tz % 100;
		return ((size_t)(hour * 3600 + min * 60));
	}
	return ((size_t)tz * 3600);
}

/* This was lifted from ANU news. */

time_t parse_usenet_date (struct tm *datetime, char *s)
{
	char *cp, mon[80], tzone[80];
	int dom = 0, yr = 0, hr = 0, mn = 0, sc = 0, mth = 0;
	static char fmtMonthTable[37] = "janfebmaraprmayjunjulaugsepoctnovdec";
	time_t time;
	struct tm *ptime;

	if (datetime)
		memset (datetime, 0, sizeof (struct tm));

	if (!s || !*s)
		return (0);
	if (cp = strchr (s, ','))
	s = ++cp;
	while (isspace (*s))
		s++;
	*mon = '\0';
	if (isdigit (*s)) 
	{
		sscanf (s, "%d %s %d %d:%d:%d %s", &dom, mon, &yr, &hr, &mn, &sc, tzone);
		if (yr < 100) 
		{
			//
			//    Fix Year 2000 problem waiting to happen (JD 2/7/97)
			//      yr += 1900;
			//
			yr += (yr < 50) ? 2000 : 1900;
		}
	}
	else
		sscanf (s, "%*s %d %d:%d:%d %d", mon, &dom, &hr, &mn, &sc, &yr);

	if (!dom || !yr || !*(cp = mon))
		return (0);
	if ((dom <= 0) || (dom >= 32))
		return (0);
	if ((yr < 1980) || (yr > 2050))
		return (0);
	if (strlen (mon) > 10)
		return (0);
	if ((hr < 0) || (hr > 23))
		return (0);
	if ((mn < 0) || (mn > 59))
		return (0);
	if ((sc < 0) || (sc > 59))
		return (0);

	for (cp = mon; *cp; cp++)
	*cp = tolower (*cp);

	if (cp = strstr (fmtMonthTable, mon))
		mth = (cp - fmtMonthTable) / 3;

	/*  Setup a Posix time structure and calculate time in absolute
	time (seconds since midnight, Jan 1, 1970    JD 06/25/93 */

	datetime->tm_year = yr - 1900;
	datetime->tm_mon = mth;
	datetime->tm_mday = dom;
	datetime->tm_hour = hr;
	datetime->tm_min = mn;
	datetime->tm_sec = sc;
	datetime->tm_isdst = (_daylight != 0 ? 1 : 0);  //prevent +1hr conversion

	time = mktime (datetime); 
	time -=dTimeZone(tzone); //to gmt
	time -= (_timezone - ((_daylight != 0 ? 1 : 0) * 3600));   //to local

	ptime = localtime(&time);
	if (ptime)
		memcpy(datetime, ptime, sizeof (struct tm ) );   //store datetime

	return time;
}

/*-- function StringDate ----------------*/
char * StringDate (char *s, time_t time)
{
	struct tm *datetime;
	if (time != 0) 
	{
		datetime = localtime (&time);

		if (fmtDaysB4Mth) 
		{
			_sprintf (s, "%02d%s%02d%s%02d %02d:%02d",
			datetime->tm_mday, fmtDateDelim, datetime->tm_mon + 1,
			fmtDateDelim,datetime->tm_year % 100,
			datetime->tm_hour, datetime->tm_min);
		}
		else 
		{
			_sprintf (s, "%02d%s%02d%s%02d %02d:%02d",
			datetime->tm_mon + 1, fmtDateDelim, datetime->tm_mday,
			fmtDateDelim,datetime->tm_year % 100,
			datetime->tm_hour, datetime->tm_min);
		}
		return (s);
	}
	else 
	{
		strcpy(s, "-----");
		return (s);
	}
}

#define MSGF_SLEEPMSG 0x5300

BOOL SleepMsg(DWORD dwTimeout)
{
	DWORD dwStart = GetTickCount();
	DWORD dwElapsed;

#ifndef MWFMO_WAITANY
#define MWFMO_WAITANY 0
#endif

	while ((dwElapsed = GetTickCount() - dwStart) < dwTimeout) 
	{
		DWORD dwStatus = MsgWaitForMultipleObjectsEx(0, NULL,
							dwTimeout - dwElapsed, QS_ALLINPUT,
							MWFMO_WAITANY | MWMO_INPUTAVAILABLE);
		if (dwStatus == WAIT_OBJECT_0) 
		{
			MSG msg;
			while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) 
			{
				if (msg.message == WM_QUIT) 
				{
					PostQuitMessage((int)msg.wParam);
					return FALSE; // abandoned due to WM_QUIT
				}
				if (!CallMsgFilter(&msg, MSGF_SLEEPMSG)) 
				{
					TranslateMessage(&msg);
					DispatchMessage(&msg);
				}
			}
		}
	}
	return TRUE; // timed out
}

void LoadGroup()
 {
	BOOL dolist;

	SetStatbarText (NetDoc.hWndFrame, "Load Group List?", &NetDoc, TRUE, TRUE);

	dolist = DoList;
	if (dolist == ID_DOLIST_ASK - ID_DOLIST_BASE)
	{
		if (WVMessageBox (NetDoc.hDocWnd, IDS_WVUTIL_REQUEST_LATEST, IDS_WVUTIL_REQUEST_LIST,
					MB_ICONQUESTION | MB_YESNO | MB_DEFBUTTON2) == IDNO)
			dolist = 0;
	}

	/* may have lost connection while dialog box up... */
	//We don't care now if autoreconnecting
	//if (Initializing == INIT_NOT_CONNECTED) 
	//{
	//
	//break;
	// }
	if (dolist) 
	{
		StartList ();
		did_list = 1;
	}
	else 
	{
		did_list = 0;
		CommState = ST_NONE;
		CommBusy = FALSE;
		Initializing = INIT_READY;
		if (CommDoc)
			SetStatbarText (CommDoc->hWndFrame, "", CommDoc, TRUE, TRUE);
		InvalidateRect (NetDoc.hDocWnd, NULL, FALSE);
	}
}

/*-- function DoCommInput ---------------------------------------
 *
 *
 */
void DoCommInput ()
{
    int ch;

    if ((CommBusy == TRUE) && (StallTimeout)) 
    {
        
		if (CommDecoding || (CommDoc && CommDoc->ParentDoc && CommDoc->ParentDoc->savingArtIndex != (UINT)-1) )		
		{
			if ((--InternalTimeoutCounter) == 0) 
			{

				TRACE("Download Stalled!\r\n");
				InternalTimeoutCounter = StallTimeout * 1000 / COMM_TIMER_INTERVAL;	 //seconds/250
				
				MRRCloseComm(); //CommState = ST_CLOSED_COMM;

				PostMessage (NetDoc.hDocWnd, (UINT) WM_COMMAND,
								(WPARAM) IDM_DISCONNECTED, (LPARAM) 0);

				return;


			}
		}
	}

	

    
	while ((CommState != ST_CLOSED_COMM) && ((ch = MRRReadComm ()) >= 0)) 
    {
        
		if (ch == IgnoreCommCh) 
        {
        }
        else if (ch == EOLCommCh) 
        {
            *CommLinePtr = '\0';
            DoCommState ();
            CommLinePtr = CommLineIn;
			InternalTimeoutCounter = StallTimeout * 1000 / COMM_TIMER_INTERVAL;

        }
        else 
        {
            *(CommLinePtr++) = (char) ch;
            if (CommLinePtr == CommLineLWAp1)
				CommLinePtr--;
			InternalTimeoutCounter = StallTimeout * 1000 / COMM_TIMER_INTERVAL;
        }
    }
}

void update_window_title (HWND hwnd,
                     char *group_name,
                     unsigned long line_number,
                     unsigned long total_lines)
{
	char tPercent[20];
	static int prevPercent, newPercent;
	// to avoid flicker, update percent only if it has changed more than 1%

	line_number *= 100;
	if (newPercent < prevPercent)
		prevPercent = 0;

	if ((line_number % UPDATE_TITLE_FREQ) == 0) 
	{
		newPercent = (total_lines > 0) ? (int) (line_number / total_lines) : 0;
		if (newPercent != prevPercent && newPercent - prevPercent > 1) 
		{
			_sprintf (tPercent, "%d", newPercent);
			WVArg3SetWindowText (hwnd, IDS_GROUP_TITLE, CurrentGroup, tPercent, CurrentDescription);
			SetStatbarPercent(hwnd, newPercent, CommDoc, TRUE);
			prevPercent = newPercent;
		}
	}
}


int check_server_code (int retcode)
{
	HWND hAlertWnd;
	int iClass = retcode / 100;

	if (ComposeWnd) 
	{
		hAlertWnd = ComposeWnd->hWnd;
	}
	else if (CommDoc) 
	{
		if ( (CommDecoding || ( CommDoc->ParentDoc && CommDoc->ParentDoc->savingArtIndex != (UINT)-1 )) && 
			(retcode == 423 || retcode == 430) ) //bad message #, no such article
		{
			//If we're saving multiple messages, don't bother showing bad article msg MDS 
			return (-1); 
		}
		hAlertWnd = CommDoc->hWndFrame;

	}
	else 
	{
		hAlertWnd = NetDoc.hWndFrame;
	}

	switch (iClass) 
	{
		case 5:
		CommBusy = FALSE;
		CommState = ST_NONE;
#ifdef _WIN32
		WVMessageBox (hAlertWnd, IDS_WVUTIL_ERR_SERVER, IDS_WVUTIL_WINVN1, MB_OK | MB_ICONHAND | MB_SETFOREGROUND);
#else
		WVMessageBox (hAlertWnd, IDS_WVUTIL_ERR_SERVER, IDS_WVUTIL_WINVN1, MB_OK | MB_ICONHAND );

#endif
		return (1);
		break;
	case 4:
		CommBusy = FALSE;
		CommState = ST_NONE;
#ifdef _WIN32
		WVArgMessageBox (hAlertWnd, IDS_WVUTIL_STRING1, IDS_WVUTIL_MESSAGE_FROM,
						MB_OK | MB_ICONHAND | MB_SETFOREGROUND, CommLineIn );
#else
		WVArgMessageBox (hAlertWnd, IDS_WVUTIL_STRING1, IDS_WVUTIL_MESSAGE_FROM, MB_OK | MB_ICONHAND, CommLineIn  );
#endif
		if (retcode == 400)
		{
			Disconnect();
			break;
		}
		return (1);
		break;
	}
	return (0);
}

/*  Function sync_artnum

   Normally XREF returns lists of the same length for each header type
   but some servers have errors that could cause these lists to get
   out of sync. This function tries to find the proper location in the
   headers array and returns that location.  If the article number isn't
   found, it returns -1.  JD 6/19/93 */

long sync_artnum (unsigned long artnum,
             unsigned long activenum,
             header_p headers, TypGroup far * GroupDoc)
{
	long headerloc = CommDoc->ActiveLines;
	if (artnum == activenum)
		return (headerloc);
	else if (artnum < activenum) 
	{
		while ((artnum != activenum) && (headerloc > 0)) 
		{
			headerloc--;
			if ((header_elt (headers, headerloc))->number == artnum)
				return (headerloc);
		}
		return (-1);
	}
	else 
	{
		while ((artnum != activenum) && (headerloc < GroupDoc->total_headers - 1)) 
		{
			headerloc++;
			if ((header_elt (headers, headerloc))->number == artnum)
				return (headerloc);
		}
		return (-1);
	}
}

// Some Usenet voting software is generating obscenely long message-id's
// with non-unique front-parts 30 chars or greater... this was throwing
// off winvn in an interesting way.
// We solve this by replacing the last 4 characters of
// mid's longer than HEADER_MESSAGE_ID_LENGTH with a 4-char hex hash value
// string of the entire mid.

// This must be done to all references _and_ mid's in order for threading
// to be able to match 'em.

// slightly modified version of 'HashGroup' from wvlist.c

WVString message_id_with_frob(const char * source)
{
	int pos;
	unsigned long sum;
	const char * p;
	static char sRetval[HEADER_MESSAGE_ID_LENGTH + 1];

	sRetval[0] = '\0';
	if(source)
	{
		//  if('<' == *source) // skip any leading '<'
		//      source++;
		// fix a bug where a bad references line causes a GPF
		// JD 2/10/97
		//
		while (source && (('<' == *source) || 
						  (' ' == *source) ||
						  ('>' == *source)))
			source++;
		if(source)
		{
			strncpy(sRetval, source, HEADER_MESSAGE_ID_LENGTH);
			sRetval[HEADER_MESSAGE_ID_LENGTH] = '\0';
			
			if(pos = strcspn(sRetval, "> \t")) // chop trailing whitespace and '>'
				sRetval[pos] = 0;

			// only frob if we need to
			if(strlen(sRetval) >= HEADER_MESSAGE_ID_LENGTH - 1)
			{
				sum = 0;
				for (p = source; *p; p++)
				{
					sum = (sum << 1) + *p;
				}
				_sprintf(sRetval+HEADER_MESSAGE_ID_LENGTH-8, "%08x", sum);
			}
		}
	}

	// if there is no valid reference, somehow string objects
	// release hugh chunks of memory when you allocate an object
	// of zero length.  This is debug code so we can set a breakpoint.
	// (JD 2/14/97)
	//
	// if(source == 0)
	//    return 0;
	// else
	//   return WVString(sRetval);

	return WVString(sRetval);
}


//
// Side effect:
//  The last '>' replaced with '\0'
//
char * get_best_reference (char * refer,WVString& message_id)
{
	char * start, * end;
	start = strrchr (refer, '<');
	if (!start) 
	{
		// no valid ref here without a '<'
		return NULL;
	} 
	else 
	{
		start++;
		if (!start) 
			return NULL;          // JD 2/27/97
		end = strrchr (start, '>');
		if (!end) 
		{
			// the last ref is not complete, recurse
			*(start-1) = '\0';
			return (get_best_reference (refer, message_id));
		} 
		else 
		{
			if (strcmp(message_id_with_frob(start),message_id)==0) 
			{
				/* some broken software append the own message_id on the references, recurse */
				*(start-1) = '\0';
				return (get_best_reference(refer, message_id));
			} 
			else 
			{
				*end = 0; // was: *(end+1) = (char)0;  HAB 19951229
				return start;
			}
		}
	}
}



// clear out the latest header info
void ClearLatestInfo()
{
	Latest.lines = 0;
	Latest.iGroups = 0;
	Latest.sFrom       = "";
	Latest.sEmail      = "";
	Latest.sSubject    = "";
	Latest.sMessage_id = "";
	Latest.sNewsgroups = "";
	Latest.sNNTPPostHost= "";
	Latest.sXFace = "";
	Latest.sFace = "";
	Latest.cAddWrap = 0;
}

int CountTabs(char *pStr, int nMax)
{
	int ntab = 0;
	char *p = pStr;
	while (*p && p - pStr < nMax)
	if (*p++ == 0x09)
		ntab++;
	return ntab;
}

BOOL StripNewline( char* pStr, int nMax)
{
	char *pSrc = pStr;
	char *pDest = pStr;

	while(*pSrc && pSrc - pStr < nMax) 
	{
		if (*pSrc == 0x0D || *pSrc == 0x0A) 
		{
			pSrc++;
			continue;
		}

		*pDest++ = *pSrc++;
	}
	if (pDest - pStr < nMax)
		*pDest = '\0';

	return (pDest != pSrc);
}

// return pointer to text of specified field.
// fields are separated by tabs
char * GetField(char * pStr, int nMax, int field)
{
	int ntab = 0;
	char *p = pStr;
	while (*p && p - pStr < nMax && ntab != field)
		if (*(p++) == 0x09)
			ntab++;

	if (p - pStr >= nMax)
		return NULL;
	return p;
}

//determine if the field contains only numbers
BOOL IsNumberField(char * p)
{
	BOOL bNum = FALSE;
	while (*p && *p > 0x1F) 
	{
		if (!isdigit(*p) && !isspace(*p)) 
		{
			bNum = FALSE;
			break;
		}
		bNum = TRUE;
		p++; 
	}
	return bNum;
}

	//determine if the field text is bracketed with '<' '>'
BOOL IsBracketField(char *p)
{
	BOOL bBracket = TRUE;
	BOOL bLookForLeft = TRUE;
	BOOL bOutsideText = FALSE;
	int nSets = 0;

	while (*p && *p != 0x09) 
	{ //ignore <CR><LF>
		if (isspace(*p)) 
		{
			p++;
			continue;
		}
		if ( *p == '<') 
		{
			if (bLookForLeft) 
				bLookForLeft = FALSE;
			//looking for right & got another left
			p++;
			continue;
		}

		if (*p == '>') 
		{
			if (!bLookForLeft) 
			{
				bLookForLeft = TRUE;
				nSets++;
			}
			//looking for left & got another right
			p++;
			continue;
		}
		if (bLookForLeft)
			bOutsideText = TRUE;
		p++;
	}

	return (nSets != 0);
}

// XOVER should contain the following fields
// articlenumber, subject, author, date, message-id, references, bytes, lines, Xref:
// Some servers break that sequence w/ multiple references before the last two fields
// Look for:  number<tab>*<tab>*<tab>*<tab>*<tab>*<tab>number<tab>number*
// OR number<tab>(multiple fields)<tab>number<tab>number*
// OR number<tab>*<tab>*<tab>*<tab>*<tab>*<tab><tab>   (empty lines, and/or bytes)
//
// return: 0 - Invalid
//         1 - Standard  (7 standard fields)
//         2 - Extra fields, line & byte fields at end
int IsValidXOVER( char * pStr, int nMax)
{
	char * p;
	BOOL bMoreTabsOK = TRUE;
	int nTabCount = CountTabs(pStr, nMax);

	if ( nTabCount < 7)
		return 0;

	if (!IsNumberField(pStr)) //article #
		return 0;

	if (nTabCount > 7) 
	{
		p = GetField(pStr, nMax, nTabCount -1);
		if ( p == NULL || ! IsNumberField(p))
			bMoreTabsOK = FALSE;

		if (bMoreTabsOK) 
		{
			p = GetField(pStr, nMax, nTabCount );
			if ( p == NULL || ! IsNumberField(p))
				bMoreTabsOK = FALSE;      
		}

		if (bMoreTabsOK) //there are extra fields, but it ends as expected
			return 2;
	}

	p = GetField(pStr, nMax, 6);  //allow for empty fields
	if ( p == NULL || (!IsNumberField(p) && *p != 0x09))
		return 0;
	
	p = GetField(pStr, nMax, 7);  //allow for empty fields
	if ( p == NULL || (!IsNumberField(p) && *p != 0x09 && *p != 0x00))
		return 0;

	return 1;
}
/*--- function ParseAddress --------------------------
 *   String version of the function defined in WVHeader
 */
void
ParseAddress(const char *headerline, WVString& sAddr, WVString& sName)
{
	char addr[MAXHEADERLINE];
	char name[MAXHEADERLINE];
	ParseAddress(headerline, addr, sizeof(addr), name, sizeof(name));
	sAddr = addr;
	sName = name;
}


// store detailed info on the latest header read
void StoreHeaderInfo(const char* line)
{
	const char* sTmp = line;

	if (Latest.cAddWrap)
	{
		//wrapped header lines start with space
		if (!sTmp || !isspace(*sTmp))
			Latest.cAddWrap = 0;
	}   

	if(!sTmp) 
		return;
	
	while(*sTmp && !isspace(*sTmp)) // find first whitespace
		sTmp++;
	while(*sTmp && isspace(*sTmp)) // find first char following whitespace
		sTmp++;

	if(!_strnicmp ("subject:", line, 8))
	{
		Latest.sSubject = sTmp;
	}
	else if(!_strnicmp ("from:", line, 5))
	{
		ParseAddress(sTmp, Latest.sEmail, Latest.sFrom);
	}
	else if(!_strnicmp ("lines:", line, 6))
	{
		Latest.lines = atoi(sTmp);
	}

	else if(!_strnicmp ("newsgroups:", line, 11))
	{
		Latest.sNewsgroups = sTmp;
		Latest.iGroups = 1;
		while(sTmp = strchr(sTmp, ','))
		{
			Latest.iGroups++;
			sTmp++;
		}
	}

	else if(!_strnicmp ("message-id:", line, 11))
	{
		if(sTmp)
			Latest.sMessage_id = message_id_with_frob(sTmp);
		else
			Latest.sMessage_id = "";
	}
	else if (!_strnicmp ("NNTP-Posting-Host:", line, 18))
	{
		if (sTmp)
			Latest.sNNTPPostHost = sTmp;
	} 
	else if ( !_strnicmp ("x-face: ", line, 8))
	{
		Latest.cAddWrap = 1;
		if (sTmp)
		{
			Latest.sXFace = sTmp;
			MakeXFaceWnd( Latest.sXFace );

		}


	}
	else if ( !_strnicmp ("Face: ", line, 6))
	{
		Latest.cAddWrap = 2;
		if (sTmp)
		{
			Latest.sFace = sTmp;
		}

	}
	else if (Latest.cAddWrap)
	{
		if (sTmp)
		{
			
			if (Latest.cAddWrap == 1)
			{
				Latest.sXFace += sTmp;
				MakeXFaceWnd( Latest.sXFace );
			}
			else if (Latest.cAddWrap == 2)
			{
				Latest.sFace += sTmp;
			}

		}

	}
}


//   Build a list of frobbed references
//
//  Side effects:
//    Source string gets clobbered with nulls
//
void build_ref_list(WVString& sDest, const char*src,WVString& message_id)
{
	char * sTmp;
	WVString mid;
	size_t midlen;
#if 1 // S.Yoshida
	const char * start;
#else
	char * start;
#endif
	char * chk;
	BOOL found;

	if(src[0])
	{
		sTmp = new char[2 + strlen(src)];
		if(sTmp)
		{
			sTmp[0] = 0;
			start = strchr(src, '<');
			while(start)
			{
				/* some broken software inserts < > or <> which causes to add a reference twice, so jump over */
				while (start && (('<' == *start) || (' ' == *start) || ('>' == *start))) start++;
				mid = message_id_with_frob(start); /* some broken software puts itself into the references */
				if (strcmp(mid,message_id)!=0) 
				{
					/* i'm shure that it isn't the mid of the current article */
					mid += " "; /* add a blank to separate entries */
					midlen = strlen(mid);
					chk = sTmp;
					found = FALSE;
					while (chk) 
					{
						while ((chk) && (*chk == ' ')) 
							chk++;
						if (strncmp(chk,mid,midlen) == 0) 
						{
							/* this message-id is already part of the reflist */
							/* don't refer it twice */
							found = TRUE;
						}
						chk = strchr(chk,' ');

					}
					if (!found) 
					{
						/* not twice, add it to the ref_list */
						strcat(sTmp, mid);
					}
				}

				if (start)
					start = strchr(start,'>');
				if (start)
					start = strchr(start,'<');
			}
			sDest = sTmp;
			delete [] sTmp;
		}
	}
	else
		sDest = "";
}


/*-- function DoCommState ----------------------------------------------
 *
 *  Function to implement an FSA to process incoming lines from
 *  the server.
 *  This function is called once for each line from the server.
 *
 *    Entry    CommLineIn  is a zero-terminated line received from
 *                         the server.
 *             CommState   is the current state of the FSA.
 */
void DoCommState ()
{
	TypLine far *LinePtr;
	TypBlock far *BlockPtr;
	HANDLE hBlock;
	unsigned int Offset;
	TypLineID MyLineID;
	int retcode;
	int found;
	static uint32 uLastArtNum = 0;
	long syncnum;
	int mylen;
	BOOL done = FALSE;
	BOOL DlgStatus = FALSE;
	BOOL do_authinfo;
	static char group[MAXINTERNALLINE];
	static char LastCommLine[MAXCOMMLINE];
	char mybuf[MAXHEADERLINE];
	char mybuf2[MAXHEADERLINE];
	char * ptr;
	char * ptr1;
	const char far *lpsz;
	HANDLE header_handle;
	HANDLE thread_handle;
	header_p headers;
	header_p header;
	TypGroup far *GroupDoc;
	WVSeenDB*pSDB;
	static int PrevState = ST_NONE;

	TypDoc *statDoc;


	/* CommDoc is !NULL if retrieving group list, article headers or articles */
	/* CommDecoding is true if retrieving article in decode mode (not to a doc) */
	/* PostEdit !NULL if we are posting (this is from an edit, no doc involved) */
	if (CommDoc || CommDecoding || SendingPost) 
	{
		if (CommState != PrevState) 
		{
			PrevState = CommState;
			statDoc = CommDoc ? CommDoc : &NetDoc;
			WVSetStatbarText (statDoc->hWndFrame, CommState, statDoc, TRUE, TRUE);
		}

	switch (CommState) 
	{
	case ST_NONE:
#ifdef _DEBUG
		MessageBeep( (UINT) -1);
		OutputDebugString(CommLineIn);
		OutputDebugString("\r\n");	
#endif
	break;

	case ST_ESTABLISH_COMM:
		if (!sscanf (CommLineIn, "%u", &retcode))
			break;
		/* check for innd, send 'mode reader' command */
		/* this is only necessary in unusual cases... */
		/* ... at least until someone ports INN to NT 8^) */
		if (strstr (CommLineIn, "InterNetNews") &&
			!strstr (CommLineIn, "NNRP")) 
		{
			PutCommLine ("mode reader");
			break;
		}
		sprintf(mybuf, "N: %s", NNTPHost);
		g_bUseXHDR = GetPrivateProfileInt (mybuf, "ActionUseXHDR", 1, szAppProFile);
  
		if (retcode == 200 || retcode == 201)    /* was 500 from induced error */
		{
			CommBusy = TRUE;
			do_authinfo = FALSE;
			Authenticated = FALSE;
			if (strlen (NNTPUserName)) 
			{
				/* We have the AUTHINFO username.  Do we have the password? */
				if (!strlen (NNTPPasswordEncrypted)) 
				{
					/* Get the news server user password from the user */
					if (DialogBoxParam (hInst, MAKEINTRESOURCE(IDD_WINVNCOMM), NetDoc.hDocWnd, (DLGPROC) lpfnWinVnCommDlg, 1L)
							&& strlen (NNTPPasswordEncrypted)) 
					{
						do_authinfo = TRUE;
					}
					//has connection been closed while dlg up?
					if (CommState == ST_CLOSED_COMM)
					{
						Authenticated = FALSE;
						NNTPPasswordEncrypted[0] = '\0';
						CommBusy = FALSE;
						WVSetStatbarText (NetDoc.hWndFrame, CommState, &NetDoc, TRUE, TRUE);
						break;
					}
				}
				else 
				{
					do_authinfo = TRUE;
				}
			}
			if (do_authinfo) 
			{
				_sprintf (mybuf, "AUTHINFO user %s", NNTPUserName);
				CommState = ST_CHECK_AUTHINFO_USERNAME;
				WVSetStatbarText (NetDoc.hWndFrame, CommState, &NetDoc, TRUE, TRUE);
				PutCommLine (mybuf);
			}
			else 
			{
				goto End_Authinfo;
			}

		}
		else 
		{
			Disconnect ();
			InvalidateRect (NetDoc.hDocWnd, NULL, TRUE);
			WVArgMessageBox (NetDoc.hDocWnd, IDS_WVUTIL_STRING2, IDS_WVUTIL_ACCESS_PROB, MB_OK, CommLineIn);
		}
		break;

	case ST_CHECK_AUTHINFO_USERNAME:
		retcode = 0;
		sscanf (CommLineIn, "%u", &retcode);
		if (!retcode)
			break;
		if (retcode >= 500) 
		{
			Disconnect ();
			InvalidateRect (NetDoc.hDocWnd, NULL, TRUE);
			WVMessageBox (NetDoc.hDocWnd, IDS_WVUTIL_AUTHORITY, IDS_WVUTIL_WINVN2, MB_OK | MB_ICONHAND);
			break;
		}
		MRRDecrypt (NNTPPasswordEncrypted, (unsigned char *) mybuf2, MAXINTERNALLINE);
		_sprintf (mybuf, "AUTHINFO pass %s", mybuf2);
		CommState = ST_CHECK_AUTHINFO_PASSWORD;
		WVSetStatbarText (NetDoc.hWndFrame, CommState, &NetDoc, TRUE, TRUE);
		PutCommLine (mybuf);
		break;

		case ST_CHECK_AUTHINFO_PASSWORD:
		retcode = 0;
		if (sscanf (CommLineIn, "%u", &retcode) <= 0)
			break;
		if (retcode < 200 || retcode > 299) 
		{
			Disconnect ();
			InvalidateRect (NetDoc.hDocWnd, NULL, TRUE);
			WVArgMessageBox (NetDoc.hDocWnd, IDS_WVUTIL_ARG_AUTHORITY, IDS_WVUTIL_WINVN3, MB_OK | MB_ICONHAND, CommLineIn );
			break;
		}
		else 
		{
			/* Authentication was successful.  Store this fact, and the name under
			* which the user was authenticated.
			*/
			Authenticated = TRUE;
			strntcpy (AuthenticatedName, NNTPUserName, MAXNNTPSIZE - 1);
		}
		goto End_Authinfo;


    case ST_END_AUTHINFO:
End_Authinfo:;
		/* allow exit now... */
		SendMessage (NetDoc.hWndFrame, WM_MYINITMENU, (WPARAM) 0, (LPARAM) 0);

		if (Reconnecting)
		{
			Reconnecting = 0;

			CommBusy = FALSE;

			WVSetStatbarText (NetDoc.hWndFrame, CommState, &NetDoc, TRUE, TRUE);
			HideShowWindows(SW_SHOW);
			Initializing = INIT_READY;
			SetStatbarText (CommDoc->hWndFrame, "", CommDoc, TRUE, TRUE);
			InvalidateRect (NetDoc.hDocWnd, NULL, FALSE);

			if (*CurrentGroup)
			{
				CommState = ST_GROUP_RETURN;
				CommBusy = TRUE;
				_sprintf(mybuf,"GROUP %s\r\n",CurrentGroup);
				PutCommData(mybuf, strlen(mybuf));

			}
			else
			{
				CommState = ST_RECONNECTED;
			}
			//in case main wnd is hidden, don't rely on WM_PAINT
			SetUserMenus (&NetDoc, ENABLE);
			Initializing = INIT_DONE;
			break;
		}

		/* now check for the XOVER command */
		CommState = ST_XOVER_CHECK;
		PutCommLine ("XOVER");
		break;

	case ST_XOVER_CHECK:
		retcode = 0;
		sscanf (CommLineIn, "%u", &retcode);

		//Charter is returning 224 these days... prevent multiple list request dlgs
		if (strcmp (CommLineIn, ".") == 0 || (retcode == 224 && strchr(CommLineIn, '.') != 0)) 
			break;

		/* 412 == 'not in a newsgroup', others return 411 no such newsgroup*/
		if ( ((retcode/100) == 4 || retcode == 224) && (force_xhdr == 0 || force_xhdr == HDR_XOVER))    
			xoverp = 1;
		else if ((retcode/100) == 5)      /* 500 == 'command not understood' */
			xoverp = 0;
			


		if (retcode == 480)  //auth required
		{
			Disconnect ();
			InvalidateRect (NetDoc.hDocWnd, NULL, TRUE);
			WVArgMessageBox (NetDoc.hDocWnd, IDS_WVUTIL_UNEXPECTED, IDS_WVUTIL_WINVN3, MB_OK | MB_ICONHAND, CommLineIn );
			break;
		}
		InvalidateRect (NetDoc.hDocWnd, NULL, FALSE);
		/* now check for the OVERVIEW.FMT command */
		CommState = ST_OVERVIEW_FMT_CHECK;
		PutCommLine ("LIST OVERVIEW.FMT");

		break;
        
	case ST_OVERVIEW_FMT_CHECK:
		//look for xref: full so we don't have to load crossref & posting groups
		retcode = 0;
		sscanf (CommLineIn, "%u", &retcode);

		if (retcode/100 == 2)
		{
			CommState = ST_OVERVIEW_FMT;
		}
		else
		{
			LoadGroup();	
		}
		break;
	case ST_OVERVIEW_FMT:
		//server is telling us the order of the XOVER fields, we just look for the one that makes
		//crossreferences & ngroup filters easy
		if (strcmp (CommLineIn, ".") == 0) 
		{
			LoadGroup();

		}
		else
		{

			if (!stricmp(CommLineIn, "Xref:full") )
			{
				TRACE("Hey Xref:full is supported\r\n");
				g_bHaveGroupsHeader = TRUE;
			}
		}

		break;



    case ST_LIST_RESP:
		retcode = 0;
		sscanf (CommLineIn, "%u", &retcode);
		if (retcode != 215) 
		{
			check_server_code (retcode);
			CommBusy = FALSE;
			CommState = ST_NONE;
			//MDS Reset back to initial state to kill hourglass
			Initializing = INIT_READY;
			SetStatbarText (CommDoc->hWndFrame, "", CommDoc, TRUE, TRUE);
			InvalidateRect (NetDoc.hDocWnd, NULL, FALSE);

			break;
		}

		CommState = ST_LIST_GROUPLINE;
		RcvLineCount = 0;
      break;

    case ST_LIST_GROUPLINE:
      if (strcmp (CommLineIn, ".") == 0) 
	  {
        CommState = ST_NONE;
        CommBusy = FALSE;
        Initializing = INIT_READY;
        SetStatbarText (CommDoc->hWndFrame, "", CommDoc, TRUE, TRUE);
        InvalidateRect (NetDoc.hDocWnd, NULL, FALSE);

        ProcEndList ();
      }
      else if (!EnableGroupFilter || MatchFilter (CommLineIn, GroupFilter)) 
	  {
        ProcListLine ((unsigned char *) CommLineIn);
      }
      break;

    case ST_GROUP_DESCRIPTION_RESP:
      retcode = 0;
      sscanf (CommLineIn, "%u", &retcode);
      if (retcode != 215) {
        CommState = ST_GROUP_RESP;
        mylen = _sprintf(mybuf,"GROUP %s",CurrentGroup);
        PutCommLine(mybuf);
        break;
      }
      CommState = ST_GROUP_DESCRIPTION_LINE;
      break;

    case ST_GROUP_DESCRIPTION_LINE:
		if (strcmp(CommLineIn,".") == 0) 
		{
			CommState = ST_GROUP_RESP;
			mylen = _sprintf(mybuf,"GROUP %s",CurrentGroup);
			PutCommLine(mybuf);
			break;
		}
		ptr = CommLineIn;
		while ((*ptr) && (*ptr != ' ') && (*ptr != '\t'))
			ptr++;
		//MDS make sure it is the group we want.
		if ( strnicmp(CommLineIn, CurrentGroup, max((int)strlen(CurrentGroup), (int)(ptr-CommLineIn))) != 0) 
		{
			//maybe we should automatically turn off ShowDescription here? since it has caused an unexpected group list
			//ShowDescription = 0;
			break;
		}
		while ((*ptr == ' ') || (*ptr == '\t'))
			ptr++;
		mylstrncpy(CurrentDescription, ptr, MAXDESCRIPTION);
      break;

    case ST_INLINEGROUP_RESP:
      retcode = 0;
      sscanf (CommLineIn, "%u", &retcode);
	  //one way or another, right now we are done receiving messages so set CommState to NONE
	  //so we don't keep popping up threshold dialogs
	  CommState = ST_NONE;
      switch (retcode) 
	  {
		  case 411:	//no group
		  case 502: //restricted
			break;

		  default:
			if (check_server_code (retcode)) 
			{
			  CommBusy = FALSE;
			  FlagJumpGroup = 0;
			  return;
			}
			break;
      }
	  {

			TypGroup far *TempGroupDoc = InlineGroupPtr;
			if (!InlineGroupPtr)
				break;

			sscanf (CommLineIn, "%u %lu %lu %lu %s", &retcode, &TempGroupDoc->ServerEstNum,
				&TempGroupDoc->ServerFirst, &TempGroupDoc->ServerLast, mybuf2);
			TempGroupDoc->Determined = TRUE;
			TempGroupDoc->NumUnread = CalcNumUnread(TempGroupDoc);

			/*

			if (TempGroupDoc->HighestPrevSeen > TempGroupDoc->ServerLast)
			{

				// best we can do is mark all as unseen and to reset HighestPrevSeen
				if (TempGroupDoc->pSeendb)
				{ 
					
					if (TempGroupDoc->ServerLast < TempGroupDoc->pSeendb->GetLowRange() &&
					TempGroupDoc->HighestPrevSeen - TempGroupDoc->ServerLast > MaxDBSize)
					{
						//something is seriously wrong, we could end up 
						//reallocating billions of art entries.
						//So start over

						pSDB = TempGroupDoc->pSeendb;
						TempGroupDoc->pSeendb = new WVSeenDB(TempGroupDoc->ServerFirst, max(TempGroupDoc->ServerFirst, TempGroupDoc->ServerLast) );
						if(TempGroupDoc->pSeendb)
						{
							TempGroupDoc->HighestPrevSeen = TempGroupDoc->ServerFirst;
							delete pSDB;
						}
						else
							TempGroupDoc->pSeendb = pSDB; //failed? what now?


					}
					else
					{	 //art ranges incorrect, toss old db

						pSDB = TempGroupDoc->pSeendb;
						TempGroupDoc->pSeendb = new WVSeenDB(TempGroupDoc->ServerFirst, max(TempGroupDoc->ServerFirst, TempGroupDoc->ServerLast) );
						if(TempGroupDoc->pSeendb)
						{
							TempGroupDoc->HighestPrevSeen = TempGroupDoc->ServerFirst;
							delete pSDB;
						}
						else
							TempGroupDoc->pSeendb = pSDB; //failed? what now?


						TempGroupDoc->pSeendb->MarkUnseen(1, TempGroupDoc->HighestPrevSeen);

						
					}
				}

				TempGroupDoc->HighestPrevSeen = TempGroupDoc->ServerFirst-1;
				NewsrcDirty = TRUE;
			}

			TempGroupDoc->uLoadFirst = TempGroupDoc->ServerFirst;
			TempGroupDoc->uLoadLast = max(TempGroupDoc->ServerFirst, TempGroupDoc->ServerLast); // ServerLast can be zero
			TempGroupDoc->Determined = TRUE;

			if ( TempGroupDoc->ServerEstNum && (TempGroupDoc->uLoadLast - TempGroupDoc->uLoadFirst) > (TempGroupDoc->ServerEstNum*2) && 
				(TempGroupDoc->uLoadLast - TempGroupDoc->uLoadFirst) > MaxDBSize)
			{
				//this is an extremely strange range, and it's probably wrong, and we'll use HUGE amounts of memory creating
				//a DB to hold the seen list of say 8,million arts.
				//So punt.

				TempGroupDoc->uLoadFirst = TempGroupDoc->uLoadLast  - min(MaxDBSize, (TempGroupDoc->ServerEstNum*2)); 
			}
			*/

	  }
	  break;
	case ST_GROUP_RESP:
      retcode = 0;
      sscanf (CommLineIn, "%u", &retcode);
	  //one way or another, right now we are done receiving messages so set CommState to NONE
	  //so we don't keep popping up threshold dialogs
	  CommState = ST_NONE;
      switch (retcode) 
	  {
      case 411:
        /* abort the fledgling group window */
        DestroyWindow (CommDoc->hWndFrame);
        CommBusy = FALSE;
        FlagJumpGroup = 0;
        WVMessageBox (NetDoc.hDocWnd, IDS_WVUTIL_NO_SUCH_GROUP, IDS_WVUTIL_ERROR1, MB_OK | MB_ICONHAND);
        return;
        break;
      case 502:
        /* abort the fledgling group window */
        DestroyWindow (CommDoc->hWndFrame);
        CommBusy = FALSE;
        FlagJumpGroup = 0;
        WVMessageBox (NetDoc.hDocWnd, IDS_WVUTIL_RESTRICTED, IDS_WVUTIL_WINVN4, MB_OK | MB_ICONHAND);
        return;
        break;

      default:
        if (check_server_code (retcode)) 
        {
          CommBusy = FALSE;
          FlagJumpGroup = 0;
          return;
        }
        break;
      }

      LockLine (CommDoc->hParentBlock, CommDoc->ParentOffset,
                CommDoc->ParentLineID, &BlockPtr, &LinePtr);
      GroupDoc = GetGroup(LinePtr);

      sscanf (CommLineIn, "%u %lu %lu %lu %s", &retcode, &GroupDoc->ServerEstNum,
        &GroupDoc->ServerFirst, &GroupDoc->ServerLast, group);

/* Ray Dunn:99/12: Lets do some consistency checking of newsrc. */
      if (GroupDoc->HighestPrevSeen > GroupDoc->ServerLast)
      {
        
        // best we can do is mark all as unseen and to reset HighestPrevSeen
        if (GroupDoc->pSeendb)
        { 
			 //art ranges incorrect, toss old db

            pSDB = GroupDoc->pSeendb;
            GroupDoc->pSeendb = new WVSeenDB(GroupDoc->ServerFirst, max(GroupDoc->ServerFirst, GroupDoc->ServerLast) );
            if(GroupDoc->pSeendb)
            {
                GroupDoc->HighestPrevSeen = GroupDoc->ServerFirst;
                delete pSDB;
            }
            else
                GroupDoc->pSeendb = pSDB; //failed? what now?
            
			DEBUG_BREAK;
			GroupDoc->pSeendb->MarkUnseen(1,GroupDoc->HighestPrevSeen);

            
        }

        GroupDoc->HighestPrevSeen = GroupDoc->ServerFirst-1;
        NewsrcDirty = TRUE;
      }
/* Ray Dunn:99/12:End */

      GroupDoc->uLoadFirst = GroupDoc->ServerFirst;
      GroupDoc->uLoadLast = max(GroupDoc->ServerFirst, GroupDoc->ServerLast); // ServerLast can be zero
      GroupDoc->Determined = TRUE;

      if ( GroupDoc->ServerEstNum && (GroupDoc->uLoadLast - GroupDoc->uLoadFirst) > (GroupDoc->ServerEstNum*2) && 
         (GroupDoc->uLoadLast - GroupDoc->uLoadFirst) > MaxDBSize)
      {
        //this is an extremely strange range, and it's probably wrong, and we'll use HUGE amounts of memory creating
        //a DB to hold the seen list of say 8,million arts.
        //So punt.
        
         GroupDoc->uLoadFirst = GroupDoc->uLoadLast  - min(MaxDBSize, (GroupDoc->ServerEstNum*2)); 
      }

      if(!(GroupDoc->pAA))
      {
        GroupDoc->pAA = new WVArticleAction(CurrentGroup);
        GroupDoc->pAA->ReadActions();
      }
      pSDB = GroupDoc->pSeendb;
      GroupDoc->pSeendb = new WVSeenDB(GroupDoc->uLoadFirst, GroupDoc->uLoadLast, pSDB);
      if(GroupDoc->pSeendb)
        delete pSDB;
      else
        GroupDoc->pSeendb = pSDB;

      if(!(GroupDoc->pAddHdr))
      {
          GroupDoc->pAddHdr = new WVAddHdr(CurrentGroup);
          GroupDoc->pAddHdr->ReadAddHdr();
      }

      if(!(GroupDoc->Description))
      {
                GroupDoc->Description = (char *) GlobalAllocPtr (GMEM_MOVEABLE, MAXDESCRIPTION * sizeof (char));
                mylstrncpy(GroupDoc->Description, CurrentDescription, MAXDESCRIPTION);
      }


      arts_to_retrieve = GroupDoc->ServerEstNum;
      server_first = GroupDoc->ServerFirst;
      server_last = GroupDoc->ServerLast;
      
      est_num_unread = CalcNumUnread(GroupDoc);
      est_num_unread = min(est_num_unread, arts_to_retrieve);
   
      /* we don't want to grab *that* many! */
      if(arts_to_retrieve >= article_threshold || IsFlagSet(CommDoc, InstantUpdate))
      {

		TypDoc * SaveDoc = CommDoc;
        if (!IsFlagSet(CommDoc, InstantUpdate) && ShowUnreadOnly) 
        {
          DlgStatus = ID_THRESHOLD_UNREAD; /* don't ask set it for the user */
        } 
        else 
        {
/* Ray Dunn:99/10: calculate the number of articles in the list from the first unseen */
/*                 Note: cancelled articles screws this up somewhat! */
          art_num = GroupDoc->pSeendb->NextUnseen();
          arts_reqd_for_first_unseen = 1 + GroupDoc->uLoadLast - art_num;

/* Ray Dunn:99/11: Correction */
          if (arts_reqd_for_first_unseen > arts_to_retrieve)
            arts_reqd_for_first_unseen = arts_to_retrieve;
          if (arts_reqd_for_first_unseen < est_num_unread)
            arts_reqd_for_first_unseen = est_num_unread;
/* Ray Dunn:99/11:End */
/* Ray Dunn:99/10:End */

		  

          DlgStatus = DialogBox (hInst, MAKEINTRESOURCE(IDD_THRESHOLD), CommDoc->hDocWnd, (DLGPROC) lpfnWinVnThresholdDlg);
          // (arts_to_retrieve and DlgStatus set)
        }

		if ( Initializing == INIT_NOT_CONNECTED && Autoconnect && Reconnecting && DlgStatus != FALSE)
		{
			if (!InlineReconnect(NetDoc.hDocWnd))
				return ;

			CommDoc = SaveDoc;
		}

        if (!CommDoc) //server disconnected while dlg was up?
		{
            DestroyWindow (SaveDoc->hWndFrame);
			return;
		}

          
        if (CommDoc && (DlgStatus == FALSE)) 
        { /* user has aborted, return to main-window */

		  DestroyWindow (CommDoc->hWndFrame);
          CommBusy = FALSE;
          CommState = ST_NONE;
          //MDS 2006-06-16 WHY?
		  //GroupDoc->ServerFirst = GroupDoc->uLoadLast;
          
		  //GroupDoc->ServerEstNum = estnum;
          FlagJumpGroup = 0;
          return;
        }
      } 
      else 
      {
        if (ShowUnreadOnly) 
        {
          DlgStatus = ID_THRESHOLD_UNREAD;
        } 
        else 
        {
          DlgStatus = ID_THRESHOLD_ALL;
        }  
      }
      switch (DlgStatus) 
      {
        case IDC_RB_EARLIEST:
        case IDC_RB_RANGE:
            GroupDoc->uLoadFirst = server_first;
            GroupDoc->uLoadLast = server_first + arts_to_retrieve;
            if (GroupDoc->uLoadLast > GroupDoc->ServerLast)
                GroupDoc->uLoadLast = GroupDoc->ServerLast;
            break;
        case ID_THRESHOLD_ALL:
          /* all of them! note this is different from retrieving GroupDoc->ServerEstNum
             headers because sometimes canceled article are NOT counted. Example
             ServerEstNum = 592 ServerFirst = 1 ServerLast = 595 because 3 articles
             are canceled on the server */
        
          GroupDoc->uLoadFirst = GroupDoc->ServerFirst;
          arts_to_retrieve = GroupDoc->ServerEstNum;
          break;
        case ID_THRESHOLD_UNREAD:
          /* retrieve from the first unread article to the end of the server */
                 
          arts_to_retrieve = 1 + GroupDoc->uLoadLast - GroupDoc->pSeendb->NextUnseen();

                //*********** to do **********
          // make min_to_retrieve depending host and newsgroup like sortoption
          // because min_to_retrieve is depending on the traffic in the group.
          // a good value should be to go back about 3 days for every group
#ifndef FIX_df__20000427
          if ((arts_to_retrieve < min_to_retrieve) && (arts_to_retrieve > 0))
#else
         /*df 2000-04-27 - fix logic to ensure uLoadFirst <= uLoadlast */
         if (arts_to_retrieve < min_to_retrieve && !IsFlagSet(CommDoc, InstantUpdate))
#endif
                 
          {
             arts_to_retrieve = min_to_retrieve;
          }
          /* calculate which article we should load first */
          GroupDoc->uLoadFirst = 1 + GroupDoc->uLoadLast - arts_to_retrieve;

          
          //MDS no no, Server Est number can exclude cancelled arts
          //so a ServerEstNum < art_to_retrieve is OK, but we dont want to get
          //previously seen arts so only set arts_to_retrieve, don't set
          //uLoadFirst.
          if (GroupDoc->ServerEstNum < arts_to_retrieve) 
          {
            arts_to_retrieve = GroupDoc->ServerEstNum;
            if (    GroupDoc->uLoadFirst < GroupDoc->ServerFirst)   //MDS added this. Shouldn't get called
              GroupDoc->uLoadFirst = GroupDoc->ServerFirst;
          }
                    

            break;
        default:
          /* if user response is a number of articles ats_to retrive is already set */
                    /* but could also be 0 which is simply pressing the OK button without entering a value */
			if (arts_to_retrieve == 0) 
			{
				arts_to_retrieve = 1 + GroupDoc->uLoadLast - GroupDoc->pSeendb->NextUnseen();
				if ((arts_to_retrieve < min_to_retrieve)) 
				{
					arts_to_retrieve = min_to_retrieve;
				}
			}
			/* calculate which article we should load first */
			GroupDoc->uLoadFirst = 1 + GroupDoc->uLoadLast - arts_to_retrieve;

			//MDS no no, Server Est number can exclude cancelled arts
			//so a ServerEstNum < art_to_retrieve is OK, but we dont want to get
			//previously seen arts so only set arts_to_retrieve, don't change
			//uLoadFirst.
			if (GroupDoc->ServerEstNum < arts_to_retrieve) 
			{
				arts_to_retrieve = GroupDoc->ServerEstNum;
				if (    GroupDoc->uLoadFirst < GroupDoc->ServerFirst)   //MDS added this. Shouldn't get called
				  GroupDoc->uLoadFirst = GroupDoc->ServerFirst;
			}

      } /* switch */

	  ClearFlags(CommDoc, InstantUpdate);
#if 0
    if (arts_to_retrieve == 0) 
    {
       /* abort the fledgling group window */
       DestroyWindow (CommDoc->hWndFrame);
       CommBusy = FALSE;
       CommState = ST_NONE;
       MessageBox (NetDoc.hDocWnd, "No Articles to Retrieve", "WinVN", MB_OK | MB_ICONHAND);
       return;
    }
     if (GroupDoc->ServerEstNum <= 0) 
     {
       /* abort the fledgling group window */
       DestroyWindow (CommDoc->hWndFrame);
       CommBusy = FALSE;
       CommState = ST_NONE;
       GroupDoc->ServerFirst = GroupDoc->uLoadLast;
       //if (GroupDoc->ServerLast) {
       //  GroupDoc->ServerFirst = GroupDoc->ServerLast;
       //}
       GroupDoc->ServerEstNum = 0;
       MessageBox (NetDoc.hDocWnd, "Empty Newsgroup", "WinVN", MB_OK | MB_ICONHAND);
       return;
    }
#endif

      
      CommDoc->TotalLines = arts_to_retrieve; //MDS may be > 64K!

      if (arts_to_retrieve > 0) 
      {

#ifdef FIX_df__20000427      
        /*df 2000-04-26 - Handle server that lies or has received new articles,
         *                by allocating (5) extra article slots.
         *                5 may be changed to some larger value if desired.
         *                Additionally, it would be good to count the _actual_
         *                number of articles retrieved and use that in place
         *                of the server estimate for a subsequent article
         *                retrieval in the same group - to do.
         */
         unsigned long my_arts_to_retrieve = arts_to_retrieve + 5;

        /*df - these GlobalAllocs ought to be encapsulated in headarry.cpp -
         * or is it just me?
         */
        header_handle =
            GlobalAlloc (GMEM_MOVEABLE | GMEM_ZEROINIT, (long)
                         ((sizeof (TypHeader)) *
                          my_arts_to_retrieve) + sizeof (thread_array *)); 

        /* allocate space for the header_array index table */
        thread_handle =
          GlobalAlloc (GMEM_MOVEABLE,
                       (long) ((sizeof (long)) * my_arts_to_retrieve));

#else
        header_handle =
          GlobalAlloc (GMEM_MOVEABLE | GMEM_ZEROINIT, (long)
                       ((sizeof (TypHeader)) *
                        arts_to_retrieve) + sizeof (thread_array *)); //s.yoshida

        /* allocate space for the header_array index table */
        thread_handle =
          GlobalAlloc (GMEM_MOVEABLE,
                       (long) ((sizeof (long)) * arts_to_retrieve));
#endif
        GroupDoc->header_handle = header_handle;
        GroupDoc->thread_handle = thread_handle;
                GroupDoc->alloc_headers = arts_to_retrieve;
        /* stick nulls and 0's, etc.. in case display code get mis-threaded */

#ifdef FIX_df__20000427
        /*df 2000-04-26 - Avoid losing the last 5 articles */
        /*df - Having allocated extra headers, ensure they are usable! */
        initialize_header_array (header_handle, thread_handle, my_arts_to_retrieve);
        CommDoc->TotalLines = my_arts_to_retrieve;
#else
        initialize_header_array (header_handle, thread_handle, arts_to_retrieve);
#endif

      }
      
      //GroupDoc = GetGroup(LinePtr);     // (didn't we already do this?)
      //GroupDoc->ServerEstNum = estnum;
      //GroupDoc->ServerFirst = GroupDoc->uLoadFirst;
          
      GroupDoc->Threaded = ((xoverp || threadp) && (GroupDoc->thread_handle));
      ClearLatestInfo();
#ifdef FIX_df__20000427
     /*df 2000-04-26 - Initialise to _not saving_ state.
      *                This makes return codes 4xx (except 400) recoverable
      *                in response to XOVER, by fall-back to XHDR. Arguably
      *                412 and 420 should not be recoverable, but they "can't
      *                happen" by design with WinVN (oh yeah?)
      */
     GroupDoc->SubjDoc->savingArtIndex = -1;
#endif


      if (xoverp) 
      {
        mylen = _sprintf (mybuf, "XOVER %ld-%ld", GroupDoc->uLoadFirst, GroupDoc->uLoadLast);
        CommState = ST_XOVER_START;
      }
      else 
      {
        mylen = _sprintf (mybuf, "XHDR from %ld-%ld", GroupDoc->uLoadFirst, GroupDoc->uLoadLast);
        CommState = ST_XHDR_FROM_START;
      }
      PutCommLine (mybuf);
      GlobalUnlock (BlockPtr->hCurBlock);

      break;

      /* use XOVER if its available */
    case ST_XOVER_START:
      retcode = 0;
      *LastCommLine = '\0';
	  
      sscanf (CommLineIn, "%d", &retcode);
      if (retcode == 224) 
      {
            CommState = ST_XOVER_DATA;
            CommDoc->ActiveLines = 0;
            total_xhdrs = 1; //MDS set here in case killfile not enabled
            if (bEnableArticleAction) 
            {
              if(LockLine(CommDoc->hParentBlock, CommDoc->ParentOffset,
					CommDoc->ParentLineID, &BlockPtr, &LinePtr))
              {
                GroupDoc = GetGroup(LinePtr);
                //aa.ReadActions(CurrentGroup);      /* Get article action list */
                uLastArtNum = GroupDoc->uLoadFirst;
                total_xhdrs = 1;
                if ((g_pAction->NeedNgroupHeader() || GroupDoc->pAA->NeedNgroupHeader()) && !g_bHaveGroupsHeader)
                {       /* Find out if we need newsgroups */
                    total_xhdrs++;
                }
                if (g_pAction->NeedNNTPHost() || GroupDoc->pAA->NeedNNTPHost() )
                {
                    total_xhdrs++; //we need NNTP-Posting-Host too
                }
                if (!g_bUseXHDR)
                    total_xhdrs = 1;

                GlobalUnlock (BlockPtr->hCurBlock);
              }
            }
      }
      else 
      {
        if (check_server_code (retcode)) 
        {
            //abort group window
            SendMessage (CommDoc->hWndFrame,  WM_CLOSE, 0, 0L);

          break;
        }

        LockLine(CommDoc->hParentBlock, CommDoc->ParentOffset,
          CommDoc->ParentLineID, &BlockPtr, &LinePtr);
        GroupDoc = GetGroup(LinePtr);
        mylen = _sprintf (mybuf, "XHDR from %ld-%ld", GroupDoc->uLoadFirst, GroupDoc->uLoadLast);
        CommState = ST_XHDR_FROM_START;
        PutCommLine (mybuf);
        GlobalUnlock (BlockPtr->hCurBlock);
      }
      break;

    case ST_XOVER_DATA:
      char * refer;

    if(strcmp(CommLineIn, ".") == 0)
    {
      if(LockLine(CommDoc->hParentBlock, CommDoc->ParentOffset,
                 CommDoc->ParentLineID, &BlockPtr, &LinePtr))
      {
        GroupDoc = GetGroup(LinePtr);
        GroupDoc->total_headers = CommDoc->ActiveLines;
                
        //first = GroupDoc->ServerFirst;      
        //last = GroupDoc->ServerLast;
      
        if(uLastArtNum < GroupDoc->uLoadLast)
        {
          GroupDoc->pSeendb->MarkSeen(uLastArtNum + 1, GroupDoc->uLoadLast);
        }
        if(bEnableArticleAction && CommDoc->ActiveLines)/* Find out if we need newsgroups */
        { //was anything there to load?
            if ((g_pAction->NeedNgroupHeader() || GroupDoc->pAA->NeedNgroupHeader()) && !g_bHaveGroupsHeader)
            {
              /* Now ask for the newsgroups */
              if (  g_bUseXHDR)
              {
                  mylen = _sprintf (mybuf, "XHDR newsgroups %ld-%ld", GroupDoc->uLoadFirst, GroupDoc->uLoadLast);
                  CommState = ST_XHDR_NEWSGROUPS_START;
              }
              else
              {
                 art_num = GroupDoc->uLoadFirst;
                 mylen = _sprintf (mybuf, "HEAD %ld", art_num);
                 CommState = ST_HEAD_NEWSGROUPS_START;
                 CommDoc->ActiveLines = 0;
              }
              PutCommLine (mybuf);
            }
            else if (g_pAction->NeedNNTPHost() || GroupDoc->pAA->NeedNNTPHost())
            {
              if (  g_bUseXHDR)
              {

                mylen = _sprintf (mybuf, "XHDR NNTP-Posting-Host %ld-%ld", GroupDoc->uLoadFirst, GroupDoc->uLoadLast);
                CommState = ST_XHDR_NNTP_HOST_START;
              }
              else
              {
                 art_num = GroupDoc->uLoadFirst;
                 mylen = _sprintf (mybuf, "HEAD %ld", art_num);
                 CommState = ST_HEAD_NEWSGROUPS_START;
                 CommDoc->ActiveLines = 0;

              }
              PutCommLine (mybuf);

            }
        }


        if (CommState == ST_XOVER_DATA)
        {
          CommState = ST_IN_GROUP;
          CommBusy = FALSE;
          finish_header_retrieval();
          SetStatbarText (CommDoc->hWndFrame,  "", CommDoc, TRUE, TRUE);
        }
        GlobalUnlock (BlockPtr->hCurBlock);
      }
    }
    else 
    {
      char *this_hop, *next_hop;
      int   tabCount;
      int   nXoverType;

      //MDS 19991211
      //added support for bizzare multiline XOVER data from server at clio.lyris.net
      //group lyris-discuss which splits long lines, and adds extra tabs between fields.
      //we accumulate partial lines until we get a complete valid line or a part which
      //completes this XOVER line.
      //According to IETF nntp docs, XOVER doesn't preclude embedded newlines.
      //However changing field order is not considered normal. Since this doesn't break
      //old servers, it shouldn't hurt to have this support in here.
      if ( *LastCommLine != '\0')
      {
        //acummulating responses
        if (IsValidXOVER(CommLineIn, MAXCOMMLINE)) 
		{
          //skip old data
          //TRACE1("Dropping %s\n", LastCommLine);
          *LastCommLine = '\0';
        }
        else {
          //XOVER data is probably wrapped
          strncat(LastCommLine, CommLineIn, MAXCOMMLINE - strlen( LastCommLine) - 1);
          strncpy(CommLineIn, LastCommLine, MAXCOMMLINE -1);
          StripNewline(CommLineIn, MAXCOMMLINE-1);
          *LastCommLine = '\0';

        }
      }

      /* the XOVER-line should contain the following fields: */
      /* articlenumber, subject, author, date, message-id, references, bytes, lines */
      /* this means there must be at least 7 tabs */
      /* this also avoids empty XOVER-lines */
      /* formerly code from (peterk@borland.com) only checks for real empty lines */
      
      tabCount = CountTabs(CommLineIn, MAXCOMMLINE);
      nXoverType = IsValidXOVER(CommLineIn, MAXCOMMLINE);
      if (nXoverType) 
	  {
        if(LockLine (CommDoc->hParentBlock, CommDoc->ParentOffset,
          CommDoc->ParentLineID, &BlockPtr, &LinePtr))
		{
          GroupDoc = GetGroup(LinePtr);

          if(CommDoc->ActiveLines >= CommDoc->TotalLines)
          {
            GroupDoc->uLoadLast = min(GroupDoc->uLoadLast, uLastArtNum - 1);
          }
          else
          {
          header_handle = GroupDoc->header_handle;
          thread_handle = GroupDoc->thread_handle;
          //header->sNewsgroups = "";

          /* Lock the header data */
          headers = lock_headers (header_handle, thread_handle);
          header = header_elt (headers, CommDoc->ActiveLines);
        
          this_hop = CommLineIn;

          /* article number */
          next_hop = strchr (this_hop, '\t');
          *(next_hop++) = (char) NULL;
          header->number = atol (this_hop);
          if(header->number && uLastArtNum < header->number - 1)
            GroupDoc->pSeendb->MarkSeen(uLastArtNum + 1, header->number - 1);
          uLastArtNum = header->number;

          /* subject */
          this_hop = next_hop;
          next_hop = strchr (this_hop, '\t');
          *(next_hop++) = (char) NULL;

          header->pHD->sSubject = this_hop;
#ifdef RFC1522
          decodeRFC1522(header->pHD->sSubject);
#endif         

          CommDoc->LongestLine = max (CommDoc->LongestLine,
                                      ArtListSeparator3 / CharWidth +
                                      (unsigned) lstrlen (header->pHD->sSubject));

          /* author */
          this_hop = next_hop;
          next_hop = strchr (this_hop, '\t');
          *(next_hop++) = (char) NULL;

          ParseAddress (this_hop,
                        AddressString, MAXDIALOGSTRING,
                        NameString, MAXDIALOGSTRING);

          if (FullNameFrom) {
            header->pHD->sFrom = NameString;
#ifdef RFC1522
            decodeRFC1522(header->pHD->sFrom);
#endif
          }
          else
            header->pHD->sFrom = this_hop;
          header->pHD->sEmail = AddressString;

          /* date */
          this_hop = next_hop;
          next_hop = strchr (this_hop, '\t');
          *(next_hop++) = (char) NULL;

          /* Ray Dunn:99/10: added datetime arg for date only (day/month) sorting */
          header->pHD->date = parse_usenet_date (&header->pHD->datetime, this_hop);

          /* message-id */
          this_hop = next_hop;
          next_hop = strchr (this_hop, '\t');
          *(next_hop++) = (char) NULL;

          if(this_hop+1)
            header->pHD->sMessage_id = message_id_with_frob(this_hop+1);
          else
            header->pHD->sMessage_id = "";

          /* references */
          this_hop = next_hop;
          if (tabCount > 7 && nXoverType != 1)
          {
            //more fields than expected, expected last two fields are bytes & lines
            next_hop =  GetField(this_hop, (CommLineIn +MAXCOMMLINE) - this_hop, tabCount-6 ); //five fields already read
            if (next_hop)
              *(next_hop -1) = '\0';
            else
              next_hop = strchr (this_hop, '\t'); //should never happen 
          }
          else
          {
            next_hop = strchr (this_hop, '\t');
            *(next_hop++) = (char) NULL;
          }
          // stop GPF's when reference line is messed up  (JD 2/17/97) 
          // header->pHD->sBest_ref = message_id_with_frob(get_best_reference(this_hop));
          refer = get_best_reference(this_hop,header->pHD->sMessage_id);
          if (refer) 
		  {
            header->pHD->sBest_ref = message_id_with_frob(refer);
            build_ref_list(header->pHD->sFrob_ref_list, this_hop,header->pHD->sMessage_id);
          }

          /* bytes (ignored) */
          this_hop = next_hop;
          next_hop = strchr (this_hop, '\t');
          *(next_hop++) = (char) NULL;

          /* lines (last one doesn't have to have the tab) */
          this_hop = next_hop;
		  next_hop = strchr (this_hop, '\t');
		  if (next_hop)
			*(next_hop++) = (char) NULL;
          
          //MDS some microsoft posts now set total number of lines for a multipart
          //message in each sub-part, thus overflowing the #lines conversion.
          //instead of changing the TypDoc struct, we'll trim the linecount, since 
          //the line count is wrong anyway.   20000401
#ifdef _WIN32
          header->lines =   (int) atol(this_hop);
#else
          long nLines = atol(this_hop);
          if (nLines > 0xFFFF)    
            nLines = 0xFFFF;
          header->lines = (unsigned short) nLines;
#endif
          /* optional Xref: line*/
		  //Xref: server-name group-name:<number> group-name:<number> (repeats)
		  this_hop = next_hop;
		  if (this_hop && _strnicmp("xref: ", this_hop, 6) == 0 )
		  {

				char *mybuf = new char[strlen(this_hop)];
				char *p = mybuf, *ptr, *ptr1;
				*p = '\0';

				ptr = this_hop+6;
				while((*ptr != 0) && (*ptr != ' ')) (ptr)++;	 //skip server name
				if (*ptr)
					ptr++;
                while (*ptr != 0) 
                {
                    ptr1 = ptr;
                    while((*ptr1 != 0) && (*ptr1 != ':')) (ptr1)++;  /* HB 1996/03/23 */
                    if (*mybuf)
						strcat(mybuf,",");
					strncat(mybuf,ptr,ptr1-ptr);                     /* save groupname */
                    
                    NextToken(&ptr);                                  /* skip over artindex */
                    ptr1++;
                    //strntcpy(mybuf2,ptr1,ptr-ptr1);                   /* save artindex */
                    //if (stricmp(mybuf,CurrentGroup) != 0)         /* mark if not our group */ 
                    //    mark_article_seen(&NetDoc, mybuf,atol(mybuf2));
                }
				
			  if (*mybuf)
			  {
				  
				  g_bHaveGroupsHeader = TRUE;
				  header->pHD->sNewsgroups = mybuf;
				  header->pHD->iGroups = 1;
				  while(p = strchr(p, ','))
				  {
					header->pHD->iGroups++;
					p++;
				  }
			  }
			  delete [] mybuf;					
				
			/*	
				
				syncnum = sync_artnum(atol(CommLineIn),
									  (header_elt(headers, CommDoc->ActiveLines))->number,
									  headers, GroupDoc);
				if(syncnum >= 0)
				{
				  const char* sTmp;
				  header = header_elt (headers, syncnum);
				  sTmp = get_xhdr_line(CommLineIn);

				  if (strcmp(sTmp, "(none)") )
				  {
					  header->pHD->sNewsgroups = sTmp;
					  header->pHD->iGroups = 1;
					  while(sTmp = strchr(sTmp, ','))
					  {
						header->pHD->iGroups++;
						sTmp++;
					  }

					  CommDoc->ActiveLines++;

				  }



				}
				unlock_headers(header_handle, thread_handle);

				update_window_title(CommDoc->hWndFrame, group,
									RcvLineCount++, CommDoc->TotalLines * total_xhdrs);
			  */
		}
			  
		  /* set other header fields */
          header->ArtDoc = (TypDoc *) NULL;
          header->SeenSelected = WasArtSeen (header->number, GroupDoc) ? ART_SEEN : ART_UNSEEN_UNSELECTED;

          unlock_headers (header_handle, thread_handle);

          CommDoc->ActiveLines++;

          update_window_title (CommDoc->hWndFrame, group,
                               RcvLineCount++,
                               CommDoc->TotalLines * total_xhdrs);
          }
          GlobalUnlock (BlockPtr->hCurBlock);
        }
        LastCommLine[0] = '\0';
     }
     else
     {
        if (*LastCommLine) 
		{
          //TRACE1("Ignoring data %s\n", LastCommLine);
          *LastCommLine = '\0';
        }
        else 
		{
          strncpy(LastCommLine, CommLineIn, MAXCOMMLINE-1);
        }
     }
    }
    break;


      /* The next few cases handle retrieval of XHDR information for display */
      /* in the group window.  If you change the number of XHDR's retrieved */
      /* (such as adding 'XHDR References' back into the state machine), you */
      /* need to reflect that change in the variable total_xhdrs. */

      /* the current flow is FROM -> DATE -> LINES -> SUBJECT -> NEWSGROUPS */
      /* (threadp) FROM -> DATE -> LINES -> ID -> REF -> SUBJECT -> NEWSGROUPS */
      /* The NEWSGROUPS state is skipped if there are no NEWSGROUPS kill filters */

      /* this will now be done dynamically, depending on the state of */
      /* the 'threadp' variable */

    case ST_XHDR_FROM_START:
      retcode = 0;
      sscanf (CommLineIn, "%d", &retcode);

      if (retcode < 100)
        break;
      if (check_server_code (retcode)) {
        //abort group window
        SendMessage (CommDoc->hWndFrame,  WM_CLOSE, 0, 0L);
        break;
      }


      CommState = ST_XHDR_FROM_DATA;
      CommDoc->ActiveLines = 0;
      break;

    case ST_XHDR_FROM_DATA:
      if (strcmp (CommLineIn, ".") == 0) {
        LockLine (CommDoc->hParentBlock, CommDoc->ParentOffset,
                  CommDoc->ParentLineID, &BlockPtr, &LinePtr);

        GroupDoc = GetGroup(LinePtr);
        /* we do this here to allow mid-session change-of-mind */
        total_xhdrs = threadp ? 6 : 4;
        if (bEnableArticleAction &&  /* Find out if we need newsgroups */
          (g_pAction->NeedNgroupHeader() || GroupDoc->pAA->NeedNgroupHeader()))
        {
          total_xhdrs++;
        }
        if(uLastArtNum < GroupDoc->uLoadLast)
        {
          GroupDoc->pSeendb->MarkSeen(uLastArtNum + 1, GroupDoc->uLoadLast);
        }
        GroupDoc->total_headers = CommDoc->ActiveLines;

        //first = GroupDoc->ServerFirst;
        //last = GroupDoc->ServerLast;

        CommDoc->ActiveLines = 0;

        /* Now ask for the date lines */
        mylen = _sprintf (mybuf, "XHDR date %ld-%ld", GroupDoc->uLoadFirst, GroupDoc->uLoadLast);
        CommState = ST_XHDR_DATE_START;
        PutCommLine (mybuf);
        GlobalUnlock (BlockPtr->hCurBlock);
      }
      else {

        /* Access the Group struct, get HANDLE for header data */
        LockLine (CommDoc->hParentBlock, CommDoc->ParentOffset,
                  CommDoc->ParentLineID, &BlockPtr, &LinePtr);

        GroupDoc = GetGroup(LinePtr);

        if(CommDoc->ActiveLines > CommDoc->TotalLines)
        {
          GroupDoc->uLoadLast = min(GroupDoc->uLoadLast, uLastArtNum - 1);
        }
        else
        {

        header_handle = GroupDoc->header_handle;
        thread_handle = GroupDoc->thread_handle;

        /* Lock the header data */
        headers = lock_headers (header_handle, thread_handle);
        sscanf (CommLineIn, "%ld", &art_num);
        header = header_elt (headers, CommDoc->ActiveLines);
        header->number = art_num;
        if(uLastArtNum < header->number - 1)
          GroupDoc->pSeendb->MarkSeen(uLastArtNum + 1, header->number - 1);
        uLastArtNum = header->number;
        
        //header->sSubject = header->sFrom = header->sEmail = header->sMessage_id =
        // header->sBest_ref = header->sFrob_ref_list = header->sNewsgroups = "";
        header->ArtDoc = (TypDoc *) NULL;
        header->SeenSelected = WasArtSeen (art_num, GetGroup(LinePtr)) ? ART_SEEN : ART_UNSEEN;
        
        /* now use some of our nice formatting of email addresses */
        ParseAddress (get_xhdr_line (CommLineIn), header->pHD->sEmail, header->pHD->sFrom);
        
        if(!FullNameFrom)
          header->pHD->sFrom = header->pHD->sEmail;
#ifdef RFC1522
        else
          decodeRFC1522(header->pHD->sFrom);
#endif                   
        
        unlock_headers (header_handle, thread_handle);
        CommDoc->ActiveLines++;
        update_window_title (CommDoc->hWndFrame, group,
          RcvLineCount++,
          CommDoc->TotalLines * total_xhdrs);
        GlobalUnlock (BlockPtr->hCurBlock);
        }
      }

      break;

    case ST_XHDR_DATE_START:
      retcode = 0;
      sscanf (CommLineIn, "%d", &retcode);
      if (check_server_code (retcode)) {
        CommBusy = FALSE;
        CommState = ST_IN_GROUP;
        break;
      }
      CommState = ST_XHDR_DATE_DATA;
      CommDoc->ActiveLines = 0;
      break;

    case ST_XHDR_DATE_DATA:
      if ((strcmp (CommLineIn, ".") == 0) &&
          LockLine (CommDoc->hParentBlock, CommDoc->ParentOffset,
                    CommDoc->ParentLineID, &BlockPtr, &LinePtr))     {

        GroupDoc = GetGroup(LinePtr);
        GroupDoc->total_headers = CommDoc->ActiveLines;
            
        //first = GroupDoc->ServerFirst;
        //last = GroupDoc->ServerLast;
        CommDoc->ActiveLines = 0;

            /* Now ask for the #of lines */
        mylen = _sprintf (mybuf, "XHDR lines %ld-%ld", GroupDoc->uLoadFirst, GroupDoc->uLoadLast);
        CommState = ST_XHDR_LINES_START;
        PutCommLine (mybuf);
        GlobalUnlock (BlockPtr->hCurBlock);
      }
      else {

        /* Access the Group struct, get HANDLE for header data */
        if (LockLine (CommDoc->hParentBlock, CommDoc->ParentOffset,
                      CommDoc->ParentLineID, &BlockPtr, &LinePtr)) {
          GroupDoc = GetGroup(LinePtr);
          header_handle = GroupDoc->header_handle;
          thread_handle = GroupDoc->thread_handle;
          GlobalUnlock (BlockPtr->hCurBlock);

          /* Lock the header data */
          headers = lock_headers (header_handle, thread_handle);
          syncnum = sync_artnum (atol (CommLineIn),
                         (header_elt (headers, CommDoc->ActiveLines))->number,
                                 headers,
                                 GroupDoc);
          if (syncnum >= 0)
          {
/* Ray Dunn:99/10: added datetime arg for date sorting */
           (header_elt (headers, syncnum))->pHD->date
              = parse_usenet_date (&(header_elt (headers, syncnum))->pHD->datetime, get_xhdr_line (CommLineIn));          
            CommDoc->ActiveLines++;
          }

          unlock_headers (header_handle, thread_handle);

          update_window_title (CommDoc->hWndFrame, group,
                               RcvLineCount++,
                               CommDoc->TotalLines * total_xhdrs);
        }
      }

      break;

    case ST_XHDR_LINES_START:
      retcode = 0;
      sscanf (CommLineIn, "%d", &retcode);
      if (check_server_code (retcode)) {
        CommBusy = FALSE;
        CommState = ST_IN_GROUP;
        break;
      }
      CommState = ST_XHDR_LINES_DATA;
      CommDoc->ActiveLines = 0;
      break;

    case ST_XHDR_LINES_DATA:
      if ((strcmp (CommLineIn, ".") == 0) &&
          LockLine (CommDoc->hParentBlock, CommDoc->ParentOffset,
                    CommDoc->ParentLineID, &BlockPtr, &LinePtr))  {

        GroupDoc = GetGroup(LinePtr);
        GroupDoc->total_headers = CommDoc->ActiveLines;

        //first = GroupDoc->ServerFirst;
        //last = GroupDoc->ServerLast;

        CommDoc->ActiveLines = 0;

        /* Check for threading option, if enabled, go to ID & REF */
        /* states first */

        if (threadp) {
          /* Now ask for the message-id lines */
          mylen = _sprintf (mybuf, "XHDR message-id %ld-%ld", GroupDoc->uLoadFirst, GroupDoc->uLoadLast);
          CommState = ST_XHDR_MID_START;
          PutCommLine (mybuf);
        }
        else {
          CommState = ST_XHDR_SUBJECT_START;
          mylen = _sprintf (mybuf, "XHDR subject %ld-%ld", GroupDoc->uLoadFirst, GroupDoc->uLoadLast);
          PutCommLine (mybuf);
        }
        GlobalUnlock (BlockPtr->hCurBlock);
      }

      else {

        /* Access the Group struct, get HANDLE for header data */
        if (LockLine (CommDoc->hParentBlock, CommDoc->ParentOffset,
                      CommDoc->ParentLineID, &BlockPtr, &LinePtr)) {
          GroupDoc = GetGroup(LinePtr);
          header_handle = GroupDoc->header_handle;
          thread_handle = GroupDoc->thread_handle;
          GlobalUnlock (BlockPtr->hCurBlock);

          /* Lock the header data */
          headers = lock_headers (header_handle, thread_handle);

          syncnum = sync_artnum (atol (CommLineIn),
                         (header_elt (headers, CommDoc->ActiveLines))->number,
                                 headers,
                                 GroupDoc);
      if (syncnum >= 0)
      {
        //MDS some microsoft posts now set total number of lines for a multipart
        //message in each sub-part, thus overflowing the #lines conversion.
        //instead of changing the TypDoc struct, we'll trim the linecount, since 
        //the line count is wrong anyway.   20000401

        long nLines = 0;
        sscanf (CommLineIn, "%ld %ld", &art_num, &nLines);
#ifdef _WIN32
        (header_elt (headers, syncnum))->lines = (int) nLines;
#else
        if (nLines > 0xFFFF)
          nLines = 0xFFFF;
        (header_elt (headers, syncnum))->lines = (short) nLines;
#endif
        CommDoc->ActiveLines++;
      }
          unlock_headers (header_handle, thread_handle);
          update_window_title (CommDoc->hWndFrame, group,
                               RcvLineCount++,
                               CommDoc->TotalLines * total_xhdrs);
        }
      }

      break;


    case ST_XHDR_MID_START:
      retcode = 0;
      sscanf (CommLineIn, "%d", &retcode);
      if (check_server_code (retcode)) {
        CommBusy = FALSE;
        CommState = ST_IN_GROUP;
        break;
      }
      CommState = ST_XHDR_MID_DATA;
      CommDoc->ActiveLines = 0;
      break;

    case ST_XHDR_MID_DATA:
      if ((strcmp (CommLineIn, ".") == 0) &&
          (LockLine (CommDoc->hParentBlock, CommDoc->ParentOffset,
                     CommDoc->ParentLineID, &BlockPtr, &LinePtr))) {

        GroupDoc = GetGroup(LinePtr);
        GroupDoc->total_headers = CommDoc->ActiveLines;
        //first = GroupDoc->ServerFirst;
        //last = GroupDoc->ServerLast;
        CommDoc->ActiveLines = 0;

        /* Now ask for the references lines */
        CommState = ST_XHDR_REF_START;
        mylen = _sprintf (mybuf, "XHDR references %ld-%ld", GroupDoc->uLoadFirst, GroupDoc->uLoadLast);
        PutCommLine (mybuf);
        GlobalUnlock (BlockPtr->hCurBlock);
      }
      else {
        /* Access the Group struct, get HANDLE for header data */
        if (LockLine (CommDoc->hParentBlock, CommDoc->ParentOffset,
                      CommDoc->ParentLineID, &BlockPtr, &LinePtr)) {

            GroupDoc = GetGroup(LinePtr);
            header_handle = GroupDoc->header_handle;
            thread_handle = GroupDoc->thread_handle;
            GlobalUnlock (BlockPtr->hCurBlock);

        /* Lock the header data */
            headers = lock_headers (header_handle, thread_handle);
            syncnum = sync_artnum (atol (CommLineIn),
                        (header_elt (headers, CommDoc->ActiveLines))->number,
                                headers,
                                GroupDoc);
            if (syncnum >= 0) {
        (header_elt (headers, syncnum))->pHD->sMessage_id =
          message_id_with_frob((char far *) (get_xhdr_line (CommLineIn)+1));
        CommDoc->ActiveLines++;
            }
            unlock_headers (header_handle, thread_handle);

            update_window_title (CommDoc->hWndFrame, group,
                                 RcvLineCount++,
                                 CommDoc->TotalLines * total_xhdrs);
          }

        }

      break;


    case ST_XHDR_REF_START:
      retcode = 0;
      sscanf (CommLineIn, "%d", &retcode);
      if (check_server_code (retcode)) {
        CommBusy = FALSE;
        CommState = ST_IN_GROUP;
        break;
      }
      CommState = ST_XHDR_REF_DATA;
      CommDoc->ActiveLines = 0;
      break;

    case ST_XHDR_REF_DATA:
      if ((strcmp (CommLineIn, ".") == 0) &&
          (LockLine (CommDoc->hParentBlock, CommDoc->ParentOffset,
                     CommDoc->ParentLineID, &BlockPtr, &LinePtr))) {

        GroupDoc = GetGroup(LinePtr);
        GroupDoc->total_headers = CommDoc->ActiveLines;
        //first = GroupDoc->ServerFirst;
        //last = GroupDoc->ServerLast;
        GlobalUnlock (BlockPtr->hCurBlock);
        CommDoc->ActiveLines = 0;

       
        /* Now ask for the subject lines */
        mylen = _sprintf (mybuf, "XHDR subject %ld-%ld", GroupDoc->uLoadFirst, GroupDoc->uLoadLast);
        CommState = ST_XHDR_SUBJECT_START;
        PutCommLine (mybuf);
      }
      else {
        header_p header;
        char far *sLine;
        char * refer;
        
        /* Access the Group struct, get HANDLE for header data */
        if (LockLine (CommDoc->hParentBlock, CommDoc->ParentOffset,
                      CommDoc->ParentLineID, &BlockPtr, &LinePtr)){
          GroupDoc = GetGroup(LinePtr);
          header_handle = GroupDoc->header_handle;
          thread_handle = GroupDoc->thread_handle;
          GlobalUnlock (BlockPtr->hCurBlock);
          
          /* Lock the header data */
          headers = lock_headers (header_handle, thread_handle);
          
          /* for now, we only pay attention to first (whole) referral */
          sLine = get_xhdr_line (CommLineIn);
          
          /* Patch to check for bad info from server JD 6/19/93 */
          syncnum = sync_artnum (atol (CommLineIn),
            (header_elt (headers, CommDoc->ActiveLines))->number,
            headers, GroupDoc);
          if (syncnum >= 0) {
            header = header_elt (headers, syncnum);
            refer = get_best_reference(sLine,header->pHD->sMessage_id);
            if (refer) {
                 header->pHD->sBest_ref = message_id_with_frob(refer);
                 build_ref_list(header->pHD->sFrob_ref_list, sLine,header->pHD->sMessage_id);
            }
            CommDoc->ActiveLines++;
          }
          
          unlock_headers (header_handle, thread_handle);
          
          update_window_title (CommDoc->hWndFrame, group,
            RcvLineCount++,
            CommDoc->TotalLines * total_xhdrs);
        }
        
      }
      
      break;



    case ST_XHDR_SUBJECT_START:
      retcode = 0;
      sscanf (CommLineIn, "%d", &retcode);
      if (check_server_code (retcode)) 
      {
        CommBusy = FALSE;
        CommState = ST_IN_GROUP;
        break;
      }
      CommState = ST_XHDR_SUBJECT_DATA;
      break;

    case ST_XHDR_SUBJECT_DATA:
      if(strcmp(CommLineIn, ".") == 0)
      {
        if(LockLine(CommDoc->hParentBlock, CommDoc->ParentOffset,
                     CommDoc->ParentLineID, &BlockPtr, &LinePtr))
        {
          GroupDoc = GetGroup(LinePtr);
          GroupDoc->total_headers = CommDoc->ActiveLines;

          if (bEnableArticleAction &&
             (g_pAction->NeedNgroupHeader() || GroupDoc->pAA->NeedNgroupHeader())) 
          {             /* Find out if we need newsgroups */
              /* Now ask for the newsgroups */
            // CommDoc->ActiveLines = 0; unneeded set in ST_XHDR_NEWSGROUPS_START
            mylen = _sprintf (mybuf, "XHDR newsgroups %ld-%ld", GroupDoc->uLoadFirst, GroupDoc->uLoadLast);
            CommState = ST_XHDR_NEWSGROUPS_START;
            PutCommLine (mybuf);
          }
          else if (bEnableArticleAction && 
                  (g_pAction->NeedNNTPHost() || GroupDoc->pAA->NeedNNTPHost()) )
          {
            mylen = _sprintf (mybuf, "XHDR NNTP-Posting-Host %ld-%ld", GroupDoc->uLoadFirst, GroupDoc->uLoadLast);
            CommState = ST_XHDR_NNTP_HOST_START;
            PutCommLine (mybuf);

          }
          else 
          {
            CommState = ST_IN_GROUP;
            CommBusy = FALSE;
            finish_header_retrieval();
            SetStatbarText (CommDoc->hWndFrame, "", CommDoc, TRUE, TRUE);
          }
          GlobalUnlock (BlockPtr->hCurBlock);  
        }
      }
      else 
      {

        if(LockLine(CommDoc->hParentBlock, CommDoc->ParentOffset,
                     CommDoc->ParentLineID, &BlockPtr, &LinePtr))
        {
          GroupDoc = GetGroup(LinePtr);

          art_num = 0;
          sscanf (CommLineIn, "%ld", &art_num);
          if (art_num) 
          {
            header_handle = GroupDoc->header_handle;
            thread_handle = GroupDoc->thread_handle;
            GlobalUnlock (BlockPtr->hCurBlock);

            /* Lock the header data */
            headers = lock_headers (header_handle, thread_handle);

            /* update the seen thing. */
            syncnum = sync_artnum (atol (CommLineIn),
                      (header_elt (headers, CommDoc->ActiveLines))->number,
                                   headers, GroupDoc);
            if(syncnum >= 0)
            {
              header = header_elt (headers, syncnum);
              header->pHD->sSubject = get_xhdr_line (CommLineIn);
#ifdef RFC1522
              decodeRFC1522(header->pHD->sSubject);
#endif      
              //CommDoc->LongestLine = max (CommDoc->LongestLine,
              //                            ARTICLE_SUBJECT_OFFSET +
              //                            (unsigned) lstrlen (header->subject));
              CommDoc->LongestLine = max (CommDoc->LongestLine,
                                      ArtListSeparator3 / CharWidth +
                                      (unsigned) lstrlen (header->pHD->sSubject));
              CommDoc->ActiveLines++;
            }
            UnlockLine (BlockPtr, LinePtr, &(CommDoc->hParentBlock),
                      &(CommDoc->ParentOffset), &(CommDoc->ParentLineID));

            unlock_headers (header_handle, thread_handle);
            update_window_title (CommDoc->hWndFrame, group,
                               RcvLineCount++,
                               CommDoc->TotalLines * total_xhdrs);

          }
        }
      }
      break;

#ifdef OFFLINE
   case ST_HEAD2_NEWSGROUPS_START:
#endif

   case ST_HEAD_NEWSGROUPS_START:
      retcode = 0;
      sscanf (CommLineIn, "%d", &retcode);
#ifdef OFFLINE
      if (((retcode/100) == 5) && (CommState == ST_HEAD_NEWSGROUPS_START)) {
          _sprintf (mybuf, "ARTICLE %ld", art_num);
          CommState = ST_HEAD2_NEWSGROUPS_START;
          PutCommLine (mybuf);
          break;
      }
#endif
      if ((retcode/100) != 4) 
      {
         /* class 4 return-codes are handled by ST_HEAD_NEWSGROUPS_DATA */
         if(check_server_code(retcode)) 
         {
            CommBusy = FALSE;
            CommState = ST_IN_GROUP;
            break;
         }
         CommState = ST_HEAD_NEWSGROUPS_DATA;
         break;
      }

   case ST_HEAD_NEWSGROUPS_DATA:
    if(LockLine(CommDoc->hParentBlock, CommDoc->ParentOffset,
            CommDoc->ParentLineID, &BlockPtr, &LinePtr))
    {
        GroupDoc = GetGroup(LinePtr);

        if ((strcmp (CommLineIn,".") == 0) || ((retcode/100) == 4)) 
        {
            if ((retcode/100) == 4) //MDS we add empty lines if ActiveLines is always incr
                CommDoc->TotalLines--;
            else
                CommDoc->ActiveLines++;
            if (CommDoc->ActiveLines >= CommDoc->TotalLines 
                   || (total_xhdrs == 2 && art_num >= GroupDoc->uLoadLast)) 
            {   //MDS don't get more than expected if AA got us here
               CommState = ST_IN_GROUP;
               CommBusy = FALSE;
               finish_header_retrieval();
               SetStatbarText (CommDoc->hWndFrame, "", CommDoc, TRUE, TRUE);
            } 
            else 
            {
               art_num++;
               _sprintf (mybuf, "HEAD %ld", art_num);
               CommState = ST_HEAD_NEWSGROUPS_START;
               PutCommLine (mybuf);
            }
  
        } 
        else 
        {
           if (_strnicmp("newsgroups:",CommLineIn,11) == 0) 
           {
              header_handle = GroupDoc->header_handle;
              thread_handle = GroupDoc->thread_handle;

              /* Lock the header data */
              headers = lock_headers(header_handle, thread_handle);
              syncnum = sync_artnum(art_num,
                              (header_elt(headers, CommDoc->ActiveLines))->number,
                              headers, GroupDoc);
              if(syncnum >= 0)
              {
                const char* sTmp, * slTmp;
                header = header_elt (headers, syncnum);
                sTmp = CommLineIn + 11; /* points after "newsgroups:" */
                while (*sTmp == ' ')
                sTmp++;
                header->pHD->sNewsgroups = slTmp = sTmp;
                header->pHD->iGroups = 1;
                while(sTmp = strchr(sTmp, ','))
                {
                  if ((sTmp - slTmp) > 3)
                  {
                    //MDS skip bogus group names of i.e. '.'
                    header->pHD->iGroups++;
                  }
                  sTmp++;
                  slTmp = sTmp;
                }
                update_window_title (CommDoc->hWndFrame, group,
                CommDoc->ActiveLines,
                CommDoc->TotalLines);
              }
              unlock_headers(header_handle, thread_handle);
           }
           if (_strnicmp("nntp-posting-host:",CommLineIn,18) == 0)
           {
              header_handle = GroupDoc->header_handle;
              thread_handle = GroupDoc->thread_handle;

              /* Lock the header data */
              headers = lock_headers(header_handle, thread_handle);
              syncnum = sync_artnum(art_num,
                              (header_elt(headers, CommDoc->ActiveLines))->number,
                              headers, GroupDoc);
              if(syncnum >= 0)
              {
                
                
                const char* sTmp, * slTmp;
                header = header_elt (headers, syncnum);
                sTmp = CommLineIn + 18; /* points after "newsgroups:" */
                while (*sTmp == ' ')
                    sTmp++;
                header->pHD->sNNTPPostHost = slTmp = sTmp;

              }
              unlock_headers(header_handle, thread_handle);

           }

        }
        GlobalUnlock (BlockPtr->hCurBlock);  
    }
    break;
    case ST_XHDR_NNTP_HOST_START:
    // This is called to retrieve NNTP-Posting-Host if Article Action needs it
        retcode = 0;
        sscanf(CommLineIn, "%d", &retcode);
        if (xoverp && ((retcode/100) == 5))
        {
             if (GetNewsgroupsViaNNTP) 
             {
                if(LockLine(CommDoc->hParentBlock, CommDoc->ParentOffset,
                     CommDoc->ParentLineID, &BlockPtr, &LinePtr))
                   {
                     GroupDoc = GetGroup(LinePtr);
                     art_num = GroupDoc->uLoadFirst;

                     _sprintf (mybuf, "HEAD %ld", art_num);
                     CommState = ST_HEAD_NEWSGROUPS_START;
                     CommDoc->ActiveLines = 0;
                     PutCommLine (mybuf);
                     GlobalUnlock (BlockPtr->hCurBlock);  
                   }
             } 
             else 
             {
                CommState = ST_IN_GROUP;
                CommBusy = FALSE;
                finish_header_retrieval();
                SetStatbarText (CommDoc->hWndFrame, "", CommDoc, TRUE, TRUE);
             }
             break;
        }
        if(check_server_code(retcode)) 
        {
          CommBusy = FALSE;
          CommState = ST_IN_GROUP;
          break;
        }
        CommState = ST_XHDR_NNTP_HOST_DATA;
        CommDoc->CountedLines = CommDoc->ActiveLines; //MDS save the old count in case HEAD command returns empty list
        CommDoc->ActiveLines = 0;
        break;

    case ST_XHDR_NNTP_HOST_DATA:
        if (strcmp (CommLineIn, ".") == 0)
        {
          if (CommDoc->ActiveLines == 0 &&  CommDoc->CountedLines != 0)
          {
            //Fix for needgroupheaders on NewsGuy which returns empty XHDR newsgroups list
            CommDoc->ActiveLines = CommDoc->CountedLines;
            CommDoc->CountedLines = 0;
            g_bUseXHDR = FALSE;

            if ( (GetNewsgroupsViaNNTP) ) 
            {
              //try again...
              if(LockLine(CommDoc->hParentBlock, CommDoc->ParentOffset,
                              CommDoc->ParentLineID, &BlockPtr, &LinePtr)) 
              {
                GroupDoc = GetGroup(LinePtr);
                art_num = GroupDoc->uLoadFirst;

                _sprintf (mybuf, "HEAD %ld", art_num);
                CommState = ST_HEAD_NEWSGROUPS_START;
                CommDoc->ActiveLines = 0;
                PutCommLine (mybuf);
                GlobalUnlock (BlockPtr->hCurBlock);
                break;
              }
            }  

          }
          CommState = ST_IN_GROUP;
          CommBusy = FALSE;
          finish_header_retrieval();
          SetStatbarText (CommDoc->hWndFrame, "", CommDoc, TRUE, TRUE);
        }
        else
        {
          /* Access the Group struct, get HANDLE for header data */
          if(LockLine(CommDoc->hParentBlock, CommDoc->ParentOffset,
             CommDoc->ParentLineID, &BlockPtr, &LinePtr))
          {
            GroupDoc = GetGroup(LinePtr);
            header_handle = GroupDoc->header_handle;
            thread_handle = GroupDoc->thread_handle;
            GlobalUnlock(BlockPtr->hCurBlock);

            /* Lock the header data */
            headers = lock_headers(header_handle, thread_handle);
            syncnum = sync_artnum(atol(CommLineIn),
                                  (header_elt(headers, CommDoc->ActiveLines))->number,
                                  headers, GroupDoc);
            if(syncnum >= 0)
            {
              const char* sTmp;
              header = header_elt (headers, syncnum);
              sTmp = get_xhdr_line(CommLineIn);
              if (strcmp(sTmp, "(none)") )
              {
                header->pHD->sNNTPPostHost = sTmp;
                CommDoc->ActiveLines++;
              }
            }
            unlock_headers(header_handle, thread_handle);

            update_window_title(CommDoc->hWndFrame, group,
                                RcvLineCount++, CommDoc->TotalLines * total_xhdrs);
          }
        }
    break;

   case ST_XHDR_NEWSGROUPS_START:
        /* this is called also for retrieving newsgroups if Articleaction need this */
        retcode = 0;
        sscanf(CommLineIn, "%d", &retcode);
        if (xoverp && ((retcode/100) == 5)) 
        {
             if (GetNewsgroupsViaNNTP) 
             {
                if(LockLine(CommDoc->hParentBlock, CommDoc->ParentOffset,
                     CommDoc->ParentLineID, &BlockPtr, &LinePtr))
                   {
                     GroupDoc = GetGroup(LinePtr);
                     art_num = GroupDoc->uLoadFirst;

                     _sprintf (mybuf, "HEAD %ld", art_num);
                     CommState = ST_HEAD_NEWSGROUPS_START;
                     CommDoc->ActiveLines = 0;
                     PutCommLine (mybuf);
                     GlobalUnlock (BlockPtr->hCurBlock);  
                   }
             } 
             else 
             {
                CommState = ST_IN_GROUP;
                CommBusy = FALSE;
                finish_header_retrieval();
                SetStatbarText (CommDoc->hWndFrame, "", CommDoc, TRUE, TRUE);
             }
             break;        
        }
        if(check_server_code(retcode)) 
        {
          CommBusy = FALSE;
          CommState = ST_IN_GROUP;
          break;
        }
        CommState = ST_XHDR_NEWSGROUPS_DATA;
        CommDoc->CountedLines = CommDoc->ActiveLines; //MDS save the old count in case HEAD command returns empty list
        CommDoc->ActiveLines = 0;
    break;

    case ST_XHDR_NEWSGROUPS_DATA:
    if (strcmp (CommLineIn, ".") == 0)
    {
      if (CommDoc->ActiveLines == 0 &&  CommDoc->CountedLines != 0)
      {
        //Fix for needgroupheaders on NewsGuy which returns empty XHDR newsgroups list
        CommDoc->ActiveLines = CommDoc->CountedLines;
        CommDoc->CountedLines = 0;
        g_bUseXHDR = FALSE;

        if ( (GetNewsgroupsViaNNTP) ) 
        {
          //try again...
          if(LockLine(CommDoc->hParentBlock, CommDoc->ParentOffset,
                          CommDoc->ParentLineID, &BlockPtr, &LinePtr)) 
          {
            GroupDoc = GetGroup(LinePtr);
            art_num = GroupDoc->uLoadFirst;

            _sprintf (mybuf, "HEAD %ld", art_num);
            CommState = ST_HEAD_NEWSGROUPS_START;
            CommDoc->ActiveLines = 0;
            PutCommLine (mybuf);
            GlobalUnlock (BlockPtr->hCurBlock);
            break;
          }
        }  

      }
      //Now see if we need NNTP HOST data
      if(LockLine(CommDoc->hParentBlock, CommDoc->ParentOffset,
                      CommDoc->ParentLineID, &BlockPtr, &LinePtr)) 
      {
        GroupDoc = GetGroup(LinePtr);
        if (bEnableArticleAction && 
            (g_pAction->NeedNNTPHost() || GroupDoc->pAA->NeedNNTPHost()) )
        {
          mylen = _sprintf (mybuf, "XHDR NNTP-Posting-Host %ld-%ld", GroupDoc->uLoadFirst, GroupDoc->uLoadLast);
          CommState = ST_XHDR_NNTP_HOST_START;
          PutCommLine (mybuf);

        }
        GlobalUnlock (BlockPtr->hCurBlock);
      }

      if (CommState == ST_XHDR_NEWSGROUPS_DATA)
      {
        //we're done
        CommState = ST_IN_GROUP;
        CommBusy = FALSE;
        finish_header_retrieval();
        SetStatbarText (CommDoc->hWndFrame, "", CommDoc, TRUE, TRUE);
      }
    }
    else
    {
      /* Access the Group struct, get HANDLE for header data */
      if(LockLine(CommDoc->hParentBlock, CommDoc->ParentOffset,
         CommDoc->ParentLineID, &BlockPtr, &LinePtr))
      {
        GroupDoc = GetGroup(LinePtr);
        header_handle = GroupDoc->header_handle;
        thread_handle = GroupDoc->thread_handle;
        GlobalUnlock(BlockPtr->hCurBlock);

        /* Lock the header data */
        headers = lock_headers(header_handle, thread_handle);
        syncnum = sync_artnum(atol(CommLineIn),
                              (header_elt(headers, CommDoc->ActiveLines))->number,
                              headers, GroupDoc);
        if(syncnum >= 0)
        {
          const char* sTmp;
          header = header_elt (headers, syncnum);
          sTmp = get_xhdr_line(CommLineIn);

          if (strcmp(sTmp, "(none)") )
          {
              header->pHD->sNewsgroups = sTmp;
              header->pHD->iGroups = 1;
              while(sTmp = strchr(sTmp, ','))
              {
                header->pHD->iGroups++;
                sTmp++;
              }

              CommDoc->ActiveLines++;

          }



        }
        unlock_headers(header_handle, thread_handle);

        update_window_title(CommDoc->hWndFrame, group,
                            RcvLineCount++, CommDoc->TotalLines * total_xhdrs);
      }
    }
    break;

    case ST_IN_GROUP:
#ifdef _DEBUG
	///hmmm not expecting messages are we?
	MessageBeep( (UINT) -1);
	OutputDebugString(CommLineIn);
	OutputDebugString("\r\n");	
#endif

        break;

    case ST_IS_ARTICLE_ON_SERVER:  // Heirich 19961015
        retcode = 0;
        sscanf(CommLineIn, "%d", &retcode);
        IsArticleAvailable =  (retcode != 430);
        CommState = ST_IN_GROUP; /* STAT Command completed */
        CommBusy = FALSE;
        break;

    case ST_ARTICLE_RESP:
        retcode = 0;
        sscanf (CommLineIn, "%d", &retcode);
        if (check_server_code (retcode)) 
        {
            //TODO: MDS check for GPF when multiple bad message windows
            /* abort communication, because there doesn't come any data from server */
            CommState = ST_IN_GROUP;
            CommBusy = FALSE;
            /* tell the Doc retrieving is complete as a fake ! */
            if (CommDecoding && currentCoded) 
            {
                SendMessage (currentCoded->hParentWnd, (UINT) WM_COMMAND,
                            (WPARAM) ID_ARTICLE_ERROR, 0L);
            } 
            else 
            {
                SendMessage (CommDoc->hWndFrame, (UINT) WM_COMMAND,
                        (WPARAM) ID_ARTICLE_ERROR, 0L);
            }
            break;
        }
        ClearLatestInfo();
#if defined( JAPANESE ) && defined( RFC1522 )
        decodeMLStrRFC1522(MLTLN_INIT);
#endif
        CommState = ST_REC_ARTICLE_HEADER;
        UsingMIME = FALSE;        /* new article, init to no MIME */
        WithinSignature = FALSE; /* new article, init to no signature */
        ClearFlags(CommDoc, (yEncoded | BNewsEnc));
        break;

    case ST_REC_ARTICLE_HEADER:
#ifdef OFFLINE
        // S.Yoshida
        if (strcmp (CommLineIn, ".") == 0) 
        {
            goto nntpArticleERR;    /* error: empty article (end in middle of header) */
        }
#else
        //    if (strcmp (CommLineIn, ".") == 0) 
        //    {
        //      ; /* error: empty article (end in middle of header) */
        //    }
#endif
        if (IsBlankStr (CommLineIn)) 
        {    /* headers end in blank line */
            Latest.cAddWrap = 0;
#if defined( JAPANESE ) && defined( RFC1522 )
            strcpy (CommLineIn, ".");
            while (decodeMLStrRFC1522(CommLineIn)) 
            {
                WrapAddCommLineToDoc (CommLineIn);
                StoreHeaderInfo(CommLineIn);
                CommLineIn[0] = '\0';
            }
#endif // JAPANESE
            WrapAddCommLineToDoc (CommLineIn);
            CommState = ST_REC_ARTICLE;
            CommDoc->HeaderLines = CommDoc->ActiveLines;  //Heirich 19960728
        } 
        else 
        {
            if (CommDecoding) 
            {
                /* find out if this article is crossposted to other groups */
                if (((TrackSubscribedCrossposts == TRUE) || 
                        (TrackUnsubscribedCrossposts == TRUE)) &&  
                        (_strnicmp("xref:", CommLineIn, 5) == 0 )) 
                {
                    //header line looks like:
					//xref: myserver.somewhere agroup:1234 bgroup:2345
					ptr = CommLineIn+6;
					while (*ptr && *ptr != ' ') //MDS skip server name and space
						ptr++;
					while (*ptr && *ptr == ' ')
						ptr++;

                    while (*ptr != 0) 
                    {
                        ptr1 = ptr;
                        while((*ptr1 != 0) && (*ptr1 != ':')) 
							(ptr1)++;  /* HB 1996/03/23 */
                        strntcpy(mybuf,ptr,ptr1-ptr);                     /* save groupname */
                        
                        NextToken(&ptr);                                  /* skip over artindex */
                        ptr1++;
                        strntcpy(mybuf2,ptr1,ptr-ptr1);                   /* save artindex */
                        if (stricmp(mybuf,CurrentGroup) != 0)         /* mark if not our group */ 
                            mark_article_seen(&NetDoc, mybuf,atol(mybuf2));
                    }
                }
            }
        }
        if (!TrimHeader (CommLineIn))
        {
#if defined( JAPANESE ) && defined( RFC1522 )
            while (decodeMLStrRFC1522(CommLineIn)) 
            {
                WrapAddCommLineToDoc (CommLineIn);
                StoreHeaderInfo(CommLineIn);
                CommLineIn[0] = '\0';
            }
#else
#ifdef RFC1522
            decodeStrRFC1522(CommLineIn);
#endif                            
            WrapAddCommLineToDoc (CommLineIn);
            StoreHeaderInfo(CommLineIn);
#endif  // JAPANESE
        }
        break;

    case ST_REC_ARTICLE:
        if (strcmp (CommLineIn, ".") != 0) 
        {
#ifdef JAPANESE
            decodeNetworkCoding (CommLineIn, WVCDS_ISO2022);
#endif
            WrapAddCommLineToDoc (CommLineIn);
        }
        else 
        {
nntpArticleERR:;
            if (CommDoc) 
            {          
                /* Handle aborted condition JD 10/26/95 */
                /* article receive complete */
                CommState = ST_IN_GROUP;
                CommBusy = FALSE;
#ifdef VISUAL1521                       
                Visual1521InputArticle(CommDoc); //Heirich 19960827
#endif          
                if (CommDecoding && currentCoded) 
                {
                    SendMessage (currentCoded->hParentWnd, (UINT) WM_COMMAND,
                                (WPARAM) ID_ARTICLE_RETRIEVE_COMPLETE, 0L);
                    break;
                }
                else 
                {
                    SendMessage (CommDoc->hWndFrame, (UINT) WM_COMMAND,
                                (WPARAM) ID_ARTICLE_RETRIEVE_COMPLETE, 0L);
                }

                if ((CommDoc->ParentDoc) &&
                    (LockLine (CommDoc->ParentDoc->hParentBlock,
                    CommDoc->ParentDoc->ParentOffset,
                    CommDoc->ParentDoc->ParentLineID,
                    &BlockPtr, &LinePtr))) 
                {

                    GroupDoc = GetGroup(LinePtr);
                    header_handle = GroupDoc->header_handle;
                    thread_handle = GroupDoc->thread_handle;

                    /* Ray Dunn:99/10 Allow optional Author name in title bar */
                    headers = lock_headers (header_handle, thread_handle);
                    lpsz = (char far *) (const char far*)((header_elt (headers, CommDoc->LastSeenLineID))->pHD->sSubject);
                    mylstrncpy (group, lpsz, MAXGROUPNAME);

                    lpsz = (char far *) (const char far*)((header_elt (headers, CommDoc->LastSeenLineID))->pHD->sFrom);
                    mylstrncpy (mybuf2, lpsz, MAXGROUPNAME);

                    unlock_headers (header_handle, thread_handle);

                    _sprintf(mybuf,"%lu", CommDoc->TotalLines);
                    SetStatbarText (CommDoc->hWndFrame, "", CommDoc, TRUE, TRUE);
                    // easy way out, if too big, don't do it
                    if (ShowAuthorName && strlen(group) + strlen(mybuf2) + 20 < sizeof(mybuf))
                        WVArg3SetWindowText(CommDoc->hWndFrame, IDS_ARTICLE_TITLE2 ,group, mybuf2, mybuf);
                    else
                        WVArg2SetWindowText(CommDoc->hWndFrame, IDS_ARTICLE_TITLE ,group, mybuf);

                    /* Ray Dunn:99/10:End */

                    //make xface here or...
                    
					if ( ((LPCSTR)Latest.sFace) != "")
                    {
                        
						MakeFaceWnd(Latest.sFace);
                        SetFocus(CommDoc->hDocWnd);
                    }
					

                    InvalidateRect (CommDoc->hDocWnd, NULL, FALSE);
                    GlobalUnlock (BlockPtr->hCurBlock);


                    /* Skip to the first line of the text of the article
                    * and make sure it's visible on the screen.  This is
                    * so that the user doesn't have to have the first
                    * screen filled with a lengthy, worthless header.
                    *
                    * and save number of header lines (on display)
                    * for later (Bretherton)
                    */
                    if (TopOfDoc (CommDoc, &BlockPtr, &LinePtr) && ScrollPastHeaders) 
                    {
                        found = FALSE;
                        do 
                        {
                            lpsz = ((char far *) LinePtr + sizeof (TypLine) + sizeof (TypText));
                            if (IsBlankStr (lpsz)) 
                            {
                                found = TRUE;
                                CommDoc->HeaderLines = WhatLine (BlockPtr, LinePtr);
                                break;
                            }
                            if (!NextLine (&BlockPtr, &LinePtr))
                                break;
                        } while (!found);

                        NextLine (&BlockPtr, &LinePtr);

                        /* If the line is in the last screen's worth of lines, back
                        * up the pointer so it points to the first line of the last
                        * screen.
                        */
                        if (found && CommDoc->TotalLines > CommDoc->ScYLines &&
                            !CommDoc->TopScLineID)
                            AdjustTopSc (BlockPtr, LinePtr);

                        UnlockLine (BlockPtr, LinePtr, &hBlock, &Offset, &MyLineID);
                    }
                }
            }
        }
        break;

    case ST_POST_WAIT_PERMISSION:

        /*      WndPost = getWndEdit(WndPosts,CommWnd,MAXPOSTWNDS) ; */
        /*      found = (WndPost != NULL) ; */

        retcode = 0;
        sscanf (CommLineIn, "%u", &retcode);

        SetStatbarText (NetDoc.hWndFrame, "", &NetDoc, TRUE, TRUE);
        InvalidateRect (NetDoc.hDocWnd, NULL, TRUE);
        if (retcode == 340) 
        {
            //recommended ID <finq4h$635$1@registered.motzarella.org>
			DoSend (CONTINUE);
        }
        else 
        {
            check_server_code (retcode);
            AbortSendPost (ComposeWnd->hWnd);
            DoSend (CONTINUE);      /* allow any sendmail to continue */
        }
        break;

    case ST_POST_WAIT_END:

        /*      WndPost = getWndEdit(WndPosts,CommWnd,MAXPOSTWNDS) ; */
        /*      found = (WndPost != NULL) ; */

        /* no check for failure to find posting documents */
        retcode = 0;
        sscanf (CommLineIn, "%d", &retcode);
        if (retcode == 240) 
        {
            DoSend (CONTINUE);
        }
        else if (check_server_code (retcode)) 
        {
            /* cut down on winvn mailing list torture */
            if (strcmp (CommLineIn,
                    "441 Article not posted -- more included text than new text") == 0) 
            {
                //resource string > 250 chars, build error in MSCV1.5, thus split in two
                //WVMessageBox (NetDoc.hDocWnd, IDS_WVUTIL_REJECTED, IDS_WVUTIL_WINVN5, MB_OK | MB_ICONHAND);
                LoadString(hInst,IDS_WVUTIL_REJECTED2, mybuf2, MAXHEADERLINE);          
                WVArgMessageBox(NetDoc.hDocWnd, IDS_WVUTIL_REJECTED, IDS_WVUTIL_WINVN5, MB_OK | MB_ICONHAND, mybuf2);
            }
        
            /*
             * if we have an error, and did not lose the connection, abort send.
             * (if connection was lost the send was already aborted)
             */

            if (Initializing != INIT_NOT_CONNECTED) 
            {
                AbortSendPost (ComposeWnd->hWnd);
                DoSend (CONTINUE);    /* allow any sendmail to continue */
            }
        }
        break;
    
    case ST_GROUP_RETURN: //return after inline group query or reconnect
		CommState = ST_IN_GROUP;
        CommBusy = FALSE;

        statDoc = CommDoc ? CommDoc : &NetDoc;
        SetStatbarText (statDoc->hWndFrame, "", statDoc, TRUE, TRUE);

		break;

      /* the following code is for an MRR-hacked nntp server */

    case ST_GROUP_REJOIN:
      CommState = ST_ARTICLE_RESP;
      break;
#ifdef OFFLINE  // TANAKA Goh
    case ST_DELETE:
      break;

    case ST_WRITE:
        if (strcmp(CommLineIn, ".") == 0) 
        {
            CommState = ST_IN_GROUP;
            CommBusy = FALSE;
            SetStatbarPercent(CommDoc->ParentDoc->hWndFrame, 100, CommDoc->ParentDoc, TRUE);
            //SetStatbarPercent(NetDoc.hWndFrame, 0, &NetDoc, TRUE);
            SetStatbarText(CommDoc->ParentDoc->hWndFrame, "", CommDoc->ParentDoc, TRUE, TRUE);
            //SetStatbarText(NetDoc.hWndFrame, "", &NetDoc, TRUE, TRUE);
            SetStatbarPercent(CommDoc->ParentDoc->hWndFrame, 0, CommDoc->ParentDoc, TRUE);
            //SetStatbarPercent(NetDoc.hWndFrame, 0, &NetDoc, TRUE);
            PostMessage(CommDoc->ParentDoc->hDocWnd, WM_TIMER, 0, 0);
        } 
        else 
        {
            static int  prevPercent = 100, newPercent = 0;
            unsigned long   number = atol(CommLineIn) * 100;

            if (newPercent < prevPercent) 
            {
                CommBusy = TRUE;
                prevPercent = 0;
                WVSetStatbarText(CommDoc->ParentDoc->hWndFrame, IDS_STBAR_SAVE_SOUP, CommDoc->ParentDoc, TRUE, TRUE);
                //SetStatbarText(NetDoc.hWndFrame, "Saving SOUP...", &NetDoc, TRUE, TRUE);
                SetStatbarPercent(CommDoc->ParentDoc->hWndFrame, 0, CommDoc->ParentDoc, TRUE);
                //SetStatbarPercent(NetDoc.hWndFrame, 0, &NetDoc, TRUE);
            }

            if ((number % UPDATE_TITLE_FREQ) == 0) 
            {
                newPercent = (int)(number / CommDoc->TotalLines);

                if (newPercent != prevPercent && newPercent - prevPercent > 1) 
                {
                    SetStatbarPercent(CommDoc->ParentDoc->hWndFrame, newPercent, CommDoc->ParentDoc, TRUE);
                    //SetStatbarPercent(NetDoc.hWndFrame, newPercent, &NetDoc, TRUE);
                    prevPercent = newPercent;
                }
            }
        }

        break;
#endif
    }
  }
}



BOOL isLineQuotation (char *textptr)
{
    char *loc;
    loc = (char*) memchr (textptr, QuoteLineInd, 2);
    if (!loc)
        loc = (char*) memchr (textptr, '|', 2);
    if (!loc)
        loc = (char*) memchr (textptr, ':', 2);
    return (loc != NULL);
}


BOOL isSignatureDelimiter(char *textptr)
{
    if (strcmp(textptr,SIG_DELIMITER) == 0)
        return TRUE;
    if ((TruncateSignatures > 1) && (strcmp(textptr,"- -- ") == 0)) /* PGP signed articles */
        return TRUE;
    if ((TruncateSignatures > 2) && (strncmp(textptr,SIG_DELIMITER,3)==0))
        return TRUE;
    if ((TruncateSignatures > 3) && (strcmp(textptr,"--")==0))
        return TRUE;
    if ((TruncateSignatures > 4) && (strncmp(textptr,"--",2)==0))
        return TRUE;
    return FALSE;
}

/*-- function AddCommLineToDoc ---------------------------------------
 *  Adds the given line to the comm doc, wrapping if necessary
 *  uses AddCommLineToDocHelp
 */
void AddCommLineToDoc (char *line)
{
    TypLine far *LinePtr;
    TypBlock far *BlockPtr;
    char *cptr, *cdest;
    int mylen;
    unsigned int col, maxlen;
    char artline[MAXHEADERLINE];
    int PercentDone;

    if (CommState == ST_REC_ARTICLE_HEADER) 
    {
        maxlen = MAXHEADERLINE;
    } 
    else 
    {
        maxlen = MAXINTERNALLINE;
    }

    /* Copy this line into an image of a textblock line,
    * expanding tabs.
    */
    cptr = line;
    cdest = artline + sizeof (TypLine) + sizeof (TypText);

    if (CommDoc && ( (CommDoc->ParentDoc && CommDoc->ParentDoc->savingArtIndex != (UINT)-1) || IsFlagSet(CommDoc, (yEncoded | BNewsEnc)) ) )
    {
        //saving art? always store long line
        maxlen = MAXHEADERLINE;

        for (col = 0; *cptr && col < (maxlen - 3 * sizeof (TypLine) - sizeof (TypText)); cptr++, col++) 
        {
            *(cdest++) = *cptr;
        }
    }
    else
    {
        for (col = 0;
            *cptr && col < (maxlen - 3 * sizeof (TypLine) - sizeof (TypText));
            cptr++) 
        {
            if (*cptr == '\t') 
            {
                do 
                {
                    *(cdest++) = ' ';
                } while (++col & 7);
            }
            else 
            {
                *(cdest++) = *cptr;
                col++;
            }
        }
    }
    *(cdest++) = '\0';

    ((TypLine *) artline)->LineID = NextLineID++;
    if (LockLine (CommDoc->hCurAddBlock, CommDoc->AddOffset, 
        CommDoc->AddLineID, &BlockPtr, &LinePtr)) 
    {
        mylen = (cdest - artline) + sizeof (int);

        // changed by Holger.Liebig@mch.sni.de to have correct
        // allignment on MIPS processors.
        // was:  mylen += mylen % 2;
        if (mylen % sizeof(int))
        {
            mylen += sizeof(int) - (mylen % sizeof(int));
        }

        ((TypText *) (artline + sizeof (TypLine)))->NameLen =
            (cdest - 1) - (artline + sizeof (TypLine) + sizeof (TypText));
        ((TypLine *) artline)->length = mylen;
        ((TypLine *) artline)->active = TRUE;
        *((int *) (artline + mylen - sizeof (int))) = mylen;
        AddLine ((TypLine *) artline, &BlockPtr, &LinePtr);
        CommDoc->LongestLine = max (CommDoc->LongestLine, (unsigned int) col);
        CommDoc->ActiveLines++;
        UnlockLine (BlockPtr, LinePtr, &(CommDoc->hCurAddBlock),
            &(CommDoc->AddOffset), &(CommDoc->AddLineID));
    }

    if (CommDoc->CountedLines > 0) 
    {
        //  PercentDone = MulDiv(CommDoc->ActiveLines, 100, CommDoc->CountedLines);
        PercentDone = (int)((CommDoc->ActiveLines * 100) / CommDoc->CountedLines);
        SetStatbarPercent(CommDoc->hWndFrame, (int) PercentDone, CommDoc, TRUE);
    }

    if ((CommDoc->TotalLines % UPDATE_ART_FREQ) == 0)
        InvalidateRect (CommDoc->hDocWnd, NULL, FALSE);

    if (*cptr)
    {
        //if line is too long, don't toss it, keep adding
        AddCommLineToDoc(cptr);
    }
}


#define MAXQUOTATIONLENGTH 30
void WrapAddCommLineToDoc (char *line)
{
    char *start, *end;
    int thisLen, lenToGo;
    char saveChar;
    char AppendLine[MAXHEADERLINE];
    char Quotations[MAXQUOTATIONLENGTH];
    unsigned int QuoteLength;
    BOOL FillInQuotations;
    char *p;
    int result;
    int SaveCommState;

    /* special case for lines starting with '..' */
    if (strncmp (CommLineIn, "..", 2)) 
    {
        start = CommLineIn;
    }
    else 
    {
        start = CommLineIn + 1;
    }
    if (CommDecoding) 
    {
        SaveCommState = CommState;
        CommState = ST_WAIT_COMM;
        /* Here we have to stop receiving, because recursive call of
        DecodeLine(), CompleteThisDecode() or InitCoded() causes a lot of problems.
        This happens if during this functions a Windowsfunction, like MessageBox()
        is called. During Windows wait for a Input to MessageBox() Windows continues
        processing WM_TIMER messages to WVUSENET which calls WVUTIL DoCommState()
        and this call WrapAddCommLineToDoc again. And so we have a recursion.
        As long ST_WAIT_COMM is set DoCommState isn't called in WVUSENET
        */
        result = DecodeLine (currentCoded, start);
        if (result == FAIL) 
        {
            MessageBox (CommDoc->hDocWnd, "Aborted decode", "Problems during decode", MB_OK | MB_ICONEXCLAMATION);
            //MDS
            if (CompleteThisDecode () == FAIL) 
            {
                MessageBox (CommDoc->hDocWnd, "Aborted decode", "Problems during decode", MB_OK | MB_ICONEXCLAMATION);
            }
            if ((currentCoded = InitCoded (CommDoc->hDocWnd)) == NULL) 
            {
                MessageBox (CommDoc->hDocWnd, "Unable to continue due to memory constraints.  Aborted", "Init Coded Object Error", MB_OK);
            }
        }
        if (result == END_BLOCK) 
        {
            if (CompleteThisDecode () == FAIL) 
            {
                MessageBox (CommDoc->hDocWnd, "Aborted decode", "Problems during decode", MB_OK | MB_ICONEXCLAMATION);
            }
            if ((currentCoded = InitCoded (CommDoc->hDocWnd)) == NULL) 
            {
                MessageBox (CommDoc->hDocWnd, "Unable to continue due to memory constraints.  Aborted", "Init Coded Object Error", MB_OK);
            }
        }
        CommState = SaveCommState; /* Continue Receiving */
    } 
    else 
    {
        if ((CommState == ST_REC_ARTICLE) && (isSignatureDelimiter(start)))
            WithinSignature = TRUE;

        if (IsFlagSet(CommDoc,yEncoded) && strncmp( CommLineIn, "=yend ",6) == 0)
            ClearFlags(CommDoc,yEncoded);
        else if (strncmp( CommLineIn, "=ybegin ",8) == 0)
            SetFlags(CommDoc, yEncoded);
        else if ( !strncmp( CommLineIn, "<<BPST ", 7) && strstr(CommLineIn, "!BPST>>") )
            SetFlags(CommDoc, BNewsEnc);
        else if ( IsFlagSet(CommDoc,BNewsEnc) && (!*CommLineIn || !*(CommLineIn+1)) )
            ClearFlags(CommDoc,BNewsEnc);


        if ((!WithinSignature) || (!TruncateSignatures))
        {  

            if (!WrapIncomingArticleText || CommState == ST_REC_ARTICLE_HEADER || *start == '\0'
                    || (CommDoc && CommDoc->ParentDoc && CommDoc->ParentDoc->savingArtIndex != (UINT)-1 ) 
                    || (IsFlagSet(CommDoc, (yEncoded | BNewsEnc)) ) )
            {
                AddCommLineToDoc (start);
            }
            else 
            {
                lenToGo = strlen (start);
                Quotations[0] = '\0';
                FillInQuotations = TRUE;
                while (*start) 
                {
                    thisLen = min (lenToGo, WrapIncomingArticleTextLength);
                    end = start + thisLen;
                    while ((end >start) && (*end != '\0') && (*end != ' ') && (*end != '\t')) 
                    {
                        end--;
                        thisLen--;
                    }
                    if (end == start) 
                    {
                        /* there is no space or tab from start of line (or the remaining part of it) to the place where we want to insert a line break */
                        thisLen = min(lenToGo,WrapIncomingArticleTextLength);
                        end = start + thisLen;  /* prevent that we break in a =XY sequence of QP encoding */
                        if (*(end-1) == '=')
                            thisLen--;
                        else if  (*(end-2) == '=')
                            thisLen -= 2; 
                    }

                    lenToGo -= thisLen;

                    end = start + thisLen;        /* save char at end of this line */
                    saveChar = *end;
                    *end = '\0';

                    strcpy(AppendLine,Quotations);
                    strcat(AppendLine,start);
                    AddCommLineToDoc (AppendLine);

                    if ( FillInQuotations) 
                    {
                        p = Quotations;
                        QuoteLength = MAXQUOTATIONLENGTH;
                        while ((--QuoteLength) && ((isLineQuotation(start)) || (*start == ' ') || (*start == '\t'))) 
                        {
                            *p++ = *start++;
                        }
                        *p = '\0';
                        FillInQuotations = FALSE;
                    }

                    *end = saveChar;   /* restore char at end of this line */
                    if ((*end == ' ') || (*end == '\t')) 
                    {
                        end++; /* skip over a blank or tab, if we insert a linebreak instead */
                        lenToGo--; /* this blank or tab isn't counted anymore */
                    }
                    start = end;
                }
            }
        }
    }
}

/*-- function TrimHeader ---------------------------------------------
 *
 * If TrimHeaders is deactivated, or if we are decoding, then 
 * always return false (don't trim header).
 * else, if this header line should be skipped, return TRUE
 *
 * jsc 9/24/94
 */
BOOL TrimHeader (char *line)
{
    if (TrimHeaders && !CommDecoding) 
    {
        if (!IsBlankStr (line) &&
#ifdef VISUAL1521
            ( !UseVISUAL1521 || _strnicmp ("content-type:", line, 13) ) &&
            ( !UseVISUAL1521 || _strnicmp ("content-transfer-encoding:", line, 26) ) &&
#endif  
            _strnicmp ("to:", line, 3) &&
            _strnicmp ("subject:", line, 8) &&
            _strnicmp ("date:", line, 5) &&
            _strnicmp ("from:", line, 5) &&
            _strnicmp ("reply-to:", line, 9) &&
            _strnicmp ("newsgroups:", line, 11) &&
            _strnicmp ("references:", line, 11) &&
            _strnicmp ("summary:", line, 8) &&
            _strnicmp ("distribution:", line, 13) &&
            _strnicmp ("message-id:", line, 11) &&  /* need this for article replies */
            _strnicmp ("followup-to:", line, 12) &&
            _strnicmp ("keywords:", line, 9) &&
            _strnicmp ("x-face:", line, 7) &&
			_strnicmp ("face: ", line, 6)) 
        {
            return TRUE;  /* matches none of above, so trim it */
        }
    }
    return FALSE;         /* don't trim it */
}

/*-- function WasArtSeen ---------------------------------------------
 *
 *  Determines whether (according to the information in a TypGroup entry)
 *  a given article number was seen.
 *
 *  Returns ART_SEEN iff the article has been seen.
 */
char WasArtSeen(unsigned long ArtNum, TypGroup far *GroupPtr)
{
    return GroupPtr->pSeendb->WasSeen(ArtNum) ? ART_SEEN : ART_UNSEEN;
}

/*-- function CalcNumUnread ---------------------------------------------
 *
 *  Determines number of unseen articles in a group
 *
 *  jsc 11/12/94
 */
unsigned long CalcNumUnread(TypGroup far *GroupPtr)
{
    if (GroupPtr->pSeendb)
		return GroupPtr->pSeendb->NumUnseen();
	else 
		return GroupPtr->ServerLast - GroupPtr->ServerFirst;
}

/*--- function mylstrncmp -----------------------------------------------
 *
 *   Just like strncmp, except takes long pointers.
 */
int mylstrncmp(char far *ptr1, char far *ptr2, int len)
{
    for (; len--; ptr1++, ptr2++) 
    {
        if (*ptr1 > *ptr2) 
        {
            return (1);
        }
        else if (*ptr1 < *ptr2) 
        {
            return (-1);
        }
    }
    return (0);
}

/*--- function mylstrncpy -----------------------------------------------
 *
 *   Just like strncpy, except takes long pointers.
 */
char far * mylstrncpy(char far *ptr1, const char far *ptr2, int len)
{
    char far *targ = ptr1;

    if (ptr2) 
    {
        for (; --len && *ptr2; ptr1++, ptr2++) 
        {
            *ptr1 = *ptr2;
        }
    }
    *ptr1 = '\0';
    return (targ);
}

/* this is a temporary test... */
char far * mylstrcpy(char_p ptr1, char far *ptr2)
{
    char far *targ = ptr1;
    for (; *ptr2; ptr1++, ptr2++) 
    {
        *ptr1 = *ptr2;
    }
    *ptr1 = '\0';
    return (targ);
}

// find first occurance of s2 in s1 case insensitive   MDS20000410
char *stristr(const char far *s1, const char far *s2)
{
    int n, m, l, o, bfound;
    const char *c1 = NULL;
    
    n=m=l=bfound=0;
    
    while ( *(s1+n))
        n++;
    
    while ( *(s2+m))
        m++;
    
    if (m > n) return 0;
    
    m--;
    while ( l < n - m && !bfound)
    {
        if (tolower(*(s1+l)) == tolower(*(s2)))
        {
            c1 = (const char *) s1+l;
            for (o = 1; o <= m; o++)
            {
                if ( tolower( *(c1+o)) != tolower(*(s2+o)))
                {
                    break;
                }
                if ( o == m)
                    bfound = 1;
            }
        }
        if (!bfound)
            l++;    
    }
    if (bfound)
        return (char *) c1;
    else
        return 0;
     
}

#if 0
/*--- function lstrcmpnoblank ------------------------------------------
 *
 *   Like strcmp, except takes long pointers and also stops at
 *   the first blank.
 */
int lstrcmpnoblank(char far **str1, char far **str2)
{
    register char far *s1 = *str1, far * s2 = *str2;

    for (; *s1 && *s2 && *s1 != ' ' && *s2 != ' '; s1++, s2++) 
    {
        if (*s1 > *s2) 
        {
            return (1);
        }
        else if (*s1 < *s2) 
        {
            return (-1);
        }
    }
    if (*s1 == *s2) 
    {
        return (0);
    }
    else if (*s1) 
    {
        return (1);
    }
    else 
    {
        return (-1);
    }
}
#endif

void
finish_header_retrieval ()
{
  char *sSortDef;
  TypLine far *LinePtr;
  TypBlock far *BlockPtr;
  TypGroup far *GroupDoc;
  HANDLE header_handle, thread_handle;
  thread_array thread_index;
  char mybuf[MAXINTERNALLINE];
  header_p headers, hp;
  uint32 i;
  long active_headers;
  
  /* release the mouse that is captured to the usenet window */
  ReleaseCapture ();

  CommDoc->TotalLines = CommDoc->ActiveLines;
  /* Disabled by MRR so that ActiveLines is the number of lines
   * we should display in the Group window.  Eventually, will
   * change it so that ActiveLines will count only unseen articles
   * if the user desires.
   */
  /* CommDoc->ActiveLines = 0; */
  /* Fetch this group's line in NetDoc so we can get the
   * group's name for the window's title bar.
   */
  if (LockLine (CommDoc->hParentBlock, CommDoc->ParentOffset,
                CommDoc->ParentLineID, &BlockPtr, &LinePtr)) {

    GroupDoc = GetGroup(LinePtr);
    header_handle = GroupDoc->header_handle;
    thread_handle = GroupDoc->thread_handle;
    CommDoc->KilledLines = 0;
    CommDoc->SelectedLines = 0;
    
    if (header_handle) {
      headers = lock_headers (header_handle, thread_handle);
      thread_index = *((thread_array_p) ((char_p) headers - sizeof (char_p)));

      GroupDoc->max_headers = GroupDoc->total_headers;
      GroupDoc->total_headers = CommDoc->TotalLines;  // needed in ActOnArticles()
      CommDoc->SelectedLines = 0;

      if(bEnableArticleAction)
      {
        g_pAction->ActOnArticles(GroupDoc, headers, TRUE);
        GroupDoc->pAA->ActOnArticles(GroupDoc, headers, FALSE);  /* Perform article actions */
        active_headers = move_killed_headers ( header_handle, thread_handle, GroupDoc->max_headers);

        CommDoc->TotalLines = active_headers;
        CommDoc->ActiveLines = active_headers;
        GroupDoc->total_headers = active_headers;
        CommDoc->KilledLines = GroupDoc->max_headers - active_headers;

      }

      if (CommDoc->TotalLines > 0)
      {
        for(i = 0; i < CommDoc->TotalLines; i++)
        {
          hp = header_elt(headers, i);
          if(ART_SELECTED & hp->SeenSelected)
            CommDoc->SelectedLines++;
        }
      }

//      iSortOption = IDM_SORT_ARTNUM;
      sSortDef = STR_SORT_ARTNUM;
      if (threadp) {
        WVSetWindowText (CommDoc->hWndFrame, IDS_GROUP_SORTING);
//      iSortOption = IDM_SORT_THREADSUB; // we're testing this as the default.
        sSortDef = STR_SORT_THREADSUB;
      }

/* Ray Dunn:99/11: No Sort Enhancement */
      if (InhibitSort)
      {
        iSortOption = IDM_DONT_SORT;
        InhibitSort = FALSE;
        WriteSortOption(iSortOption, CurrentGroup);
      }
      else
        iSortOption = ReadSortOption(sSortDef, CurrentGroup);


/*    TRACE2("WVUtil: Newsgroup <%s>, iSortOption: %d\n", CurrentGroup, iSortOption); */
      sort_by_option(headers, thread_index, threadp, CommDoc->TotalLines,
            header_handle, thread_handle);

      unlock_headers (header_handle, thread_handle);
    }
  }

  if (CommDoc->KilledLines) {
    WVArgXSetWindowText (CommDoc->hWndFrame, IDS_GROUP_DESCRIPTIONKILLED, CurrentGroup, 
                         CommDoc->TotalLines, CommDoc->KilledLines, CurrentDescription);
  }
    else {
        _sprintf (mybuf, "%lu", CommDoc->TotalLines);
        WVArg3SetWindowText (CommDoc->hWndFrame, IDS_GROUP_DESCRIPTION, CurrentGroup, mybuf, CurrentDescription);
    }
  /* If we have information from NEWSRC on the highest-
   * numbered article previously seen, position the window
   * so the new articles can be seen without scrolling.
   */
  {
    unsigned long i;

    if (LockLine (CommDoc->hParentBlock, CommDoc->ParentOffset,
                  CommDoc->ParentLineID, &BlockPtr, &LinePtr)) {

      /* inside the lock, can access the GroupStruct */
      GroupDoc = GetGroup(LinePtr);
      header_handle = GroupDoc->header_handle;
      thread_handle = GroupDoc->thread_handle;
      if (header_handle) {
        headers = lock_headers (header_handle, thread_handle);

    /* skip to FIRST unseen article (jsc) */
        i = 0;
        if (CommDoc->TotalLines > 0) 
        {
          for (i = 0;
             (i < CommDoc->TotalLines && (ART_SEENORKILLED & (header_elt (headers, i))->SeenSelected));
             i++);
        }

                
        //MDS fixes so initial position line doesn't scroll off bottom of view 20000426
        if (i == CommDoc->TotalLines)
                    i--;
        CommDoc->TopLineOrd =
           (CommDoc->TotalLines > i && CommDoc->ScYLines > 0 && CommDoc->TotalLines - i < CommDoc->ScYLines - 1) ? 
           (CommDoc->TotalLines - CommDoc->ScYLines) : (i > CommDoc->ScYLines && CommDoc->TotalLines > CommDoc->ScYLines) ?
            i - CommDoc->ScYLines +1 : 0; 

        CommDoc->ActiveLineID = i;
        ClearFlags(CommDoc, ThumbTracking);
        //CommDoc->ThumbTracking = FALSE;   /* if thumb tracking, release it */
    
        unlock_headers (header_handle, thread_handle);
        if (ConfigJumpGroup && (FlagJumpGroup & JUMPFLAG_GROUP)) 
        {
          if (i >= CommDoc->TotalLines) 
          {
            RetrieveNextGroup( BlockPtr, LinePtr, IDS_WVART_THATS_ALL, FALSE );
          } 
          else 
          {
            FlagJumpGroup = JUMPFLAG_ARTICLE;
          }
        }

      }
    }
  }

  SendMessage (CommDoc->hWndFrame, (UINT) WM_COMMAND, (WPARAM) ID_RETRIEVE_COMPLETE, 0L);
  InvalidateRect (CommDoc->hDocWnd, NULL, TRUE);
  UpdateWindow (CommDoc->hDocWnd);
  if (FlagJumpGroup & JUMPFLAG_ARTICLE) 
  {
    PostMessage (CommDoc->hDocWnd, WM_CHAR, '\r', 0);
    FlagJumpGroup = 0;
  }

}

/*
 * Look through the MAIL or Post edits and return the edit with
 * matching window handle Consider - centralising initial window
 * location in wvmail and wndpost using a single array (save passing
 * structure and size into this module)
 *
 */

WndEdit *
getWndEdit (WndEdit * WndEdits, HWND hWnd, int numEntries)
{
  int ih;

  for (ih = 0; ih < numEntries; ih++) {
    if (WndEdits[ih].hWnd == hWnd) {
      return &WndEdits[ih];
    }
  }

  /*MessageBox(0,"getWndEditFound Nothing","mrb debug", MB_OK | MB_ICONHAND); */

  return (WndEdit *) NULL;
}

WndEdit *
GetComposeWnd (HWND hWnd)
{
  WndEdit *compWnd;

  compWnd = getWndEdit (WndPosts, hWnd, MAXPOSTWNDS);
  if (!compWnd) {
    compWnd = getWndEdit (WndMails, hWnd, MAXMAILWNDS);
  }
  return compWnd;
}
/* ------------------------------------------------------------------------
 * Replace any white space at end of string with NULL's
 * JSC 11/1/93
 */
void RemoveTrailingWhiteSpace (char *str)
{
    register int i;

    for (i = strlen (str) - 1; i > 0 && isspace (str[i]); i--)
        str[i] = '\0';
}
/* ------------------------------------------------------------------------
 * Subsitute all occurences of character "oldchar" in string with
 * character "newchar" in string  (JD 3/27/97
 */
void SubsituteCharInString(char *str, char oldchar, char newchar)
{
    register int i;

    for (i = strlen (str) - 1; i > 0; i--)
    {
        if (str[i] == oldchar) 
            str[i] = newchar;
    }
}

/*------------------------------------------------------------------------------
 * IsBlankStr
 * Returns true if the string is entirely whitespace, else false
 * JSC 12/6/93
 */
BOOL IsBlankStr (const char *temp)
{
    register const char *ptr;
    for (ptr = temp; *ptr; ptr++)
        if (!isspace (*ptr))
            return (FALSE);
    return (TRUE);
}

/*------------------------------------------------------------------------------
 * isnumber
 * Returns true if the string is a all digits
 * JSC 12/6/93
 */
BOOL isnumber (char *str)
{
    char *ptr;

    for (ptr = str; *ptr != '\0'; ptr++)
        if (!isdigit (*ptr))
            return (FALSE);
    return (TRUE);
}

/* ------------------------------------------------------------------------
 * strntcpy is strncpy, but also terminates the dest str
 * jsc 9/28/94
 */
char * strntcpy (char *dest, char *src, int len)
{
    register char *d = dest;
    register char *s = src;
    register int l = 0;

    while (l < len && (*d++ = *s++)) 
    {
        l++;
    }
    if (l == len) 
    {
        *d = '\0';
    }
    return dest;
}


/* ------------------------------------------------------------------------
 *    Open the common font dialog
 *      Place resulting selection name and size in face,style and size
 *      Note: to select a printer font, send style as "Printer"
 *      printer font selection ignores any chosen style
 *      (JSC 1/9/94)
 */
BOOL AskForFont (HWND hParentWnd, char *face, int *size, char *style)
{
    LOGFONT lf;
    CHOOSEFONT cf;
    HDC hDC;

    memset (&lf, 0, sizeof (LOGFONT));
    strcpy (lf.lfFaceName, face);
    /* convert points to logical units (1 pt = 1/72 inch) */
    /* For printer fonts, use ScreenYPixels here anyway - the choosefont */
    /* dialog appears to require the lfHeight to be in screen units */
    /* we will convert point size to PrinterUnits in InitPrinterFonts() */
    lf.lfHeight = -MulDiv (*size, ScreenYPixels, 72);

    memset (&cf, 0, sizeof (CHOOSEFONT));
    cf.lStructSize = sizeof (CHOOSEFONT);
    cf.hwndOwner = hParentWnd;
    cf.lpLogFont = &lf;
    if (!stricmp (style, "Printer")) 
    {
        cf.nFontType = PRINTER_FONTTYPE;
        hDC = GetPrinterDC (hParentWnd);
        cf.hDC = hDC;
        cf.Flags = CF_PRINTERFONTS | CF_INITTOLOGFONTSTRUCT | CF_FORCEFONTEXIST;
    }
    else 
    {
        cf.nFontType = SCREEN_FONTTYPE;
        cf.Flags = CF_SCREENFONTS | CF_INITTOLOGFONTSTRUCT | CF_USESTYLE | CF_FORCEFONTEXIST;
        cf.lpszStyle = style;
    }
    if (!ChooseFont (&cf))
        return (FAIL);

    /*      if (!stricmp (style, "Printer"))      // commented out JD 6/17/94 */
    /*         ReleaseDC (NetDoc.hDocWnd, hDC);  */

    /*      if (!stricmp (style, "Printer")) */
    /*         DeletePrinterDC (hDC); */

    *size = cf.iPointSize / 10;   /* iPointSize is in tenths of a point */

    strcpy (face, lf.lfFaceName);
    return (SUCCESS);
}

/* ------------------------------------------------------------------------
 *    Open the common color dialog
 *      (JSC 1/9/94)
 */
BOOL AskForColor (HWND hParentWnd, COLORREF * color)
{
    CHOOSECOLOR cc;
    COLORREF nearC;
    HDC hDC;

    memset (&cc, 0, sizeof (CHOOSECOLOR));
    cc.lStructSize = sizeof (CHOOSECOLOR);
    cc.hwndOwner = hParentWnd;
    cc.rgbResult = *color;
    cc.lpCustColors = CustomColors;
    cc.Flags = CC_RGBINIT;

    if (!ChooseColor (&cc))
        return (FAIL);

    /* until we figure out how to deal with dithered colors, force */
    /* the color to the nearest physical color */
    hDC = GetDC (hParentWnd);
    nearC = GetNearestColor (hDC, cc.rgbResult);
    if (cc.rgbResult != nearC)
        MessageBox (hParentWnd, "WinVN does not currently support dithered (non-solid) colors.\nThe nearest physical solid color has been selected.",
                    "Sorry", MB_OK | MB_ICONINFORMATION);
    *color = nearC;
    ReleaseDC (hParentWnd, hDC);
    return (SUCCESS);
}

/* ------------------------------------------------------------------------
 * This should be used instead of EM_GETHANDLE on global edit buf
 * Returns a string containing the contents of an edit wnd
 * It is the reponsibility of the caller to GlobalFreePtr the string
 * (JSC)
 */
char * GetEditText (HWND hWndEdit)
{
  unsigned int size;
  char *newText;
  BOOL bRichEdit = FALSE;
  WndEdit *compWnd = GetComposeWnd (GetParent(hWndEdit));
  
  if (compWnd && compWnd->richedit)
    bRichEdit = TRUE;


#define EDIT_PAD 2
    if (!bRichEdit || !WordWrap)
    {
        SendMessage (hWndEdit, EM_FMTLINES, (WPARAM) WordWrap, 0L);

        size = (unsigned int) SendMessage (hWndEdit, WM_GETTEXTLENGTH, 0, 0L) + EDIT_PAD;

        if ((newText = (char *) GlobalAllocPtr (GMEM_MOVEABLE, size * sizeof (char))) == NULL) {
            MessageBox (hWndEdit, "Memory allocation failure", "Edit Text", MB_OK);
            return (NULL);
        }
        *newText = '\0';

        if (SendMessage (hWndEdit, WM_GETTEXT, size, (LPARAM) (LPCSTR) newText) != (long) (size - EDIT_PAD)) {
            MessageBox (hWndEdit, "Failed to get text", "Edit Text", MB_OK);
            return (NULL);
        }

    }
    else
    {
        //since EM_FMTLINES is not supported in rich edit controls
        //we get line wraps iteratively
        
        long nLines = 0;
        long nCount = 0;
        long  nLen=0;
        char *p = NULL;

        nLines = SendMessage (hWndEdit, EM_GETLINECOUNT, 0L, 0L);
        size = (unsigned int) SendMessage (hWndEdit, WM_GETTEXTLENGTH, 0, 0L) + EDIT_PAD;
        size += (unsigned int)nLines * 2; //assume all softwrapped

        if ((newText = (char *) GlobalAllocPtr (GMEM_MOVEABLE, size * sizeof (char))) == NULL) {
            MessageBox (hWndEdit, "Memory allocation failure", "Edit Text", MB_OK);
            return (NULL);
        }
        *newText = '\0';
        
        p = newText;
        for (nCount = 0; nCount < nLines; nCount++)
        {
            //set buffer size
            WORD *pSize = (WORD *)p;
            if ( size - (p - newText) > 65535)
                *pSize = 65535;
            else 
                *pSize = size - (p - newText);

            nLen = SendMessage (hWndEdit, EM_GETLINE, (WPARAM)nCount, (LPARAM) p);
            p += nLen;
            if (nLen)
            {
                if (*(p-1) != '\n')
                {
                    *p++ = '\r';
                    *p++ = '\n';
                }
            }
        }
        *p = '\0';
    }
  return (newText);
}
LRESULT SetEditText (HWND hWndEdit, char *editMem)
{
    return (SendMessage (hWndEdit, WM_SETTEXT, 0, (LPARAM) (LPCSTR) editMem));
}

/* ------------------------------------------------------------------------
 *    Write an integer to the private profile
 */
BOOL WritePrivateProfileInt (const char far * lpAppName, const char far * lpKeyName, int val, const char far * lpProFile)
{
    char buf[20];

    itoa (val, buf, 10);
    return (WritePrivateProfileString (lpAppName, lpKeyName, buf, lpProFile));
}

/* ------------------------------------------------------------------------
 *    Get/Write an unsigned integers and longs to the private profile
 *      (JSC 1/8/94)
 */
BOOL WritePrivateProfileUInt (const char far * lpAppName, const char far * lpKeyName, unsigned int val, const char far * lpProFile)
{
    char buf[20];
#ifdef __GNUC__
      //no ultoa in Mingw  GCC 2.95.
      _i64toa((unsigned __int64)val, buf, 10);
#else
    ultoa ((unsigned long)val, buf, 10);
#endif
  return (WritePrivateProfileString (lpAppName, lpKeyName, buf, lpProFile));
}

unsigned int GetPrivateProfileUInt (char far * lpAppName, char far * lpKeyName, unsigned int val, char far * lpProFile)
{
    char buf[20];

    GetPrivateProfileString (lpAppName, lpKeyName, "", buf, 20, lpProFile);

    if (*buf)
        return (atoui (buf));
    else
        return (val);
}

BOOL WritePrivateProfileLong (char far * lpAppName, char far * lpKeyName, long val, char far * lpProFile)
{
    char buf[20];

    ltoa (val, buf, 10);
    return (WritePrivateProfileString (lpAppName, lpKeyName, buf, lpProFile));
}

long GetPrivateProfileLong (char far * lpAppName, char far * lpKeyName, long val, char far * lpProFile)
{
    char buf[20];

    GetPrivateProfileString (lpAppName, lpKeyName, "", buf, 20, lpProFile);

    if (*buf)
        return (atol (buf));
    else
        return (val);
}
/* ------------------------------------------------------------------------
 *    Refresh Window functions
 *      Called after a font/color selection has changed to affect all
 *      windows of a certain type (group/article/status)
 *      (JSC 1/9/94)
 */
void
RefreshGroupWnds (BOOL bReSort)
{
    register int i;
    for (i = 0; i < MAXGROUPWNDS; i++)
        if (GroupDocs[i].InUse && GroupDocs[i].hDocWnd) 
        {
/* Ray Dunn:99/10: Resort */
			if (bReSort)
				ResortGroup(GroupDocs[i].hWndFrame);
/* Ray Dunn:99/10:End */
            SetHandleBkBrush (GroupDocs[i].hDocWnd, hListBackgroundBrush);
            SendMessage (GroupDocs[i].hWndFrame, WM_SIZE, 0, 0L);
            InvalidateRect (GroupDocs[i].hWndFrame, NULL, TRUE);
            InvalidateRect (GroupDocs[i].hDocWnd, NULL, TRUE);
            UpdateWindow (GroupDocs[i].hWndFrame);
        }
}

void
RefreshArticleWnds ()
{
    register int i;

    for (i = 0; i < MAXARTICLEWNDS; i++)
        if (ArticleDocs[i].InUse && ArticleDocs[i].hDocWnd) 
        {
            SetHandleBkBrush (ArticleDocs[i].hDocWnd, hArticleBackgroundBrush);
            SendMessage (ArticleDocs[i].hWndFrame, WM_SIZE, 0, 0L);
            InvalidateRect (ArticleDocs[i].hWndFrame, NULL, TRUE);
            InvalidateRect (ArticleDocs[i].hDocWnd, NULL, TRUE);
            UpdateWindow (ArticleDocs[i].hWndFrame);
        }
}

void RefreshComposeWnds ()
{
    register int i;

    for (i = 0; i < MAXPOSTWNDS; i++)
        if (WndPosts[i].hWnd) 
        {
            ResizeComposeControls (&WndPosts[i], 0, 0);
        }

    for (i = 0; i < MAXMAILWNDS; i++)
        if (WndMails[i].hWnd) 
        {
            ResizeComposeControls (&WndMails[i], 0, 0);
        }
}

void RefreshStatusWnds ()
{
    register int i;

    if (DebugText != NULL) 
    {
        SetHandleBkBrush (DebugText->hTextWnd, hStatusBackgroundBrush);
        SendMessage (DebugText->hTextWnd, WM_SIZE, 0, 0L);
        InvalidateRect (DebugText->hTextWnd, NULL, TRUE);
    }

    for (i = 0; i < NumStatusTexts; i++)
        if (CodingStatusText[i]->hTextWnd) 
        {
          SetHandleBkBrush (CodingStatusText[i]->hTextWnd, hStatusBackgroundBrush);
          SendMessage (CodingStatusText[i]->hTextWnd, WM_SIZE, 0, 0L);
          InvalidateRect (CodingStatusText[i]->hTextWnd, NULL, TRUE);
        }
    if (hCodedBlockWnd) 
    {
        RefreshCodedBlockWnd ();
    }

}
/* ------------------------------------------------------------------------
 *    Close Window functions
 *      Batch operation, close all windows of a certain type
 *      (group/article/status)
 *      (JSC 1/18/94)
 */
void CloseWindows ()
{
    CloseArticleWnds ();
    CloseGroupWnds ();
    CloseComposeWnds ();
    CloseStatusWnds ();
}

void CloseGroupWnds ()
{
    register int i;
    for (i = 0; i < MAXGROUPWNDS; i++)
        if (GroupDocs[i].InUse && GroupDocs[i].hDocWnd && (!CommBusy || &GroupDocs[i] != CommDoc))
            SendMessage (GroupDocs[i].hWndFrame, WM_CLOSE, 12345, 0L);
}

void CloseArticleWnds ()
{
    register int i;

    for (i = 0; i < MAXARTICLEWNDS; i++)
        if (ArticleDocs[i].InUse && ArticleDocs[i].hDocWnd && (!CommBusy || &ArticleDocs[i] != CommDoc))
            SendMessage (ArticleDocs[i].hWndFrame, WM_CLOSE, 0, 0L);
}

void CloseComposeWnds ()
{
    register int i;

    for (i = 0; i < MAXPOSTWNDS; i++)
        if (WndPosts[i].hWnd && !WndPosts[i].busy)
            SendMessage (WndPosts[i].hWnd, WM_CLOSE, 0, 0L);

    for (i = 0; i < MAXMAILWNDS; i++)
        if (WndMails[i].hWnd && !WndMails[i].busy)
            SendMessage (WndMails[i].hWnd, WM_CLOSE, 0, 0L);
}

void CloseStatusWnds ()
{
    /* destroying a coding status text is like popping from a stack */
    /* so we just loop while the top of the stack still exists */
    int numSkipped = 0;
    while (numSkipped < NumStatusTexts && CodingStatusText[numSkipped]->hTextWnd)
        if (!CodingStatusText[numSkipped]->IsBusy)
            SendMessage (CodingStatusText[numSkipped]->hTextWnd, WM_CLOSE, 0, 0L);
        else
            numSkipped++;
#if 0
    if (CodingState) 
    {
        MessageBox (CodingStatusText[0]->hTextWnd,
            "Please wait until en/decoding is complete",
            "Cannot close status window", MB_OK | MB_ICONSTOP);
        break;
    }
    else
        SendMessage (CodingStatusText[0]->hTextWnd, WM_CLOSE, 0, 0L);
#endif
}

/* ------------------------------------------------------------------------
 *    CascadeWindows (and helper CascadeWnd)
 *      cascade em
 *      jsc 9/18/94
 */
HWND CascadeWnd (HWND hWnd, HWND prevWnd, int nthWnd, int width, int height, int maxX, int maxY)
{
    short x, y;

    //  if (IsMaximized(hWnd))
    //     ShowWindow(hWnd, SW_SHOWNORMAL);

    x = (nthWnd * 12) % maxX;
    y = (nthWnd * (CaptionHeight + 2)) % maxY;
    SetWindowPos (hWnd, prevWnd, x, y, width, height, SWP_DRAWFRAME);

    return hWnd;
}

void WinVNCascadeWindows ()
{
    register int i;
    int nthWnd, width, height, maxX, maxY;
    HWND prevWnd;

    width = (int) (xScreen >> 1);
    height = (int) (yScreen >> 1);
    maxX = 3 * (width >> 1);      /* 3/4 screen width  */
    maxY = 3 * (height >> 1);     /* 3/4 screen height */

    prevWnd = CascadeWnd (NetDoc.hWndFrame, (HWND) NULL, 1, width, height, maxX, maxY);
    nthWnd = 2;
    for (i = 0; i < MAXGROUPWNDS; i++)
        if (GroupDocs[i].InUse && GroupDocs[i].hWndFrame && !IsMinimized (GroupDocs[i].hWndFrame)) 
        {
            prevWnd = CascadeWnd (GroupDocs[i].hWndFrame, prevWnd, nthWnd, width, height, maxX, maxY);
            nthWnd++;
        }

    for (i = 0; i < MAXARTICLEWNDS; i++)
        if (ArticleDocs[i].InUse && ArticleDocs[i].hWndFrame && !IsMinimized (ArticleDocs[i].hWndFrame)) 
        {
            prevWnd = CascadeWnd (ArticleDocs[i].hWndFrame, prevWnd, nthWnd, width, height, maxX, maxY);
            nthWnd++;
        }

    for (i = 0; i < MAXPOSTWNDS; i++)
        if (WndPosts[i].hWnd && !IsMinimized (WndPosts[i].hWnd)) 
        {
            prevWnd = CascadeWnd (WndPosts[i].hWnd, prevWnd, nthWnd, width, height, maxX, maxY);
            nthWnd++;
        }

    for (i = 0; i < MAXMAILWNDS; i++)
        if (WndMails[i].hWnd && !IsMinimized (WndMails[i].hWnd)) 
        {
            prevWnd = CascadeWnd (WndMails[i].hWnd, prevWnd, nthWnd, width, height, maxX, maxY);
            nthWnd++;
        }

    for (i = 0; i < NumStatusTexts; i++)
        if (CodingStatusText[i]->hTextWnd && !IsMinimized (CodingStatusText[i]->hTextWnd)) 
        {
            prevWnd = CascadeWnd (CodingStatusText[i]->hTextWnd, prevWnd, nthWnd, width, height, maxX, maxY);
            nthWnd++;
        }

    /* move coded block status window to top center */
    if (hCodedBlockWnd && !IsMinimized (hCodedBlockWnd)) 
    {
        SetWindowPos (hCodedBlockWnd, (HWND) NULL, (xScreen - STATUSWIDTH) >> 1, 1, STATUSWIDTH, STATUSHEIGHT, SWP_DRAWFRAME);
    }
}

/* ------------------------------------------------------------------------
 *    MinimizeWindows
 *      jsc 10/18/94
 */
void MinimizeWindows ()
{
    ShowWindow (NetDoc.hWndFrame, SW_MINIMIZE);
    MinimizeGroupWnds ();
    MinimizeArticleWnds ();
    MinimizeComposeWnds ();
    MinimizeStatusWnds ();
}

void MinimizeGroupWnds ()
{
    register int i;

    for (i = 0; i < MAXGROUPWNDS; i++)
        if (GroupDocs[i].InUse && GroupDocs[i].hDocWnd && !IsMinimized (GroupDocs[i].hWndFrame)) 
        {
            ShowWindow (GroupDocs[i].hWndFrame, SW_MINIMIZE);
        }
}

void MinimizeArticleWnds ()
{
    register int i;

    for (i = 0; i < MAXARTICLEWNDS; i++)
        if (ArticleDocs[i].InUse && ArticleDocs[i].hDocWnd && !IsMinimized (ArticleDocs[i].hWndFrame)) 
        {
            ShowWindow (ArticleDocs[i].hWndFrame, SW_MINIMIZE);
        }
}

void MinimizeComposeWnds ()
{
    register int i;

    for (i = 0; i < MAXPOSTWNDS; i++)
        if (WndPosts[i].hWnd && !IsMinimized (WndPosts[i].hWnd)) 
        {
            ShowWindow (WndPosts[i].hWnd, SW_MINIMIZE);
        }

    for (i = 0; i < MAXMAILWNDS; i++)
        if (WndMails[i].hWnd && !IsMinimized (WndMails[i].hWnd)) 
        {
            ShowWindow (WndMails[i].hWnd, SW_MINIMIZE);
        }
}

void MinimizeStatusWnds ()
{
    register int i;

    for (i = 0; i < NumStatusTexts; i++)
        if (CodingStatusText[i]->hTextWnd && !IsMinimized (CodingStatusText[i]->hTextWnd)) 
        {
            ShowWindow (CodingStatusText[i]->hTextWnd, SW_MINIMIZE);
        }

    if (hCodedBlockWnd) 
    {
        ShowWindow (hCodedBlockWnd, SW_MINIMIZE);
    }
}

/* ------------------------------------------------------------------------
 *    RestoreWindows from minimized state
 *      jsc 1/18/95
 */

void RestoreWindows ()
{
    register int i;

    if (IsMinimized(NetDoc.hWndFrame)) 
    {
        ShowWindow (NetDoc.hWndFrame, SW_RESTORE);
    }

    for (i = 0; i < MAXGROUPWNDS; i++)
        if (GroupDocs[i].InUse && GroupDocs[i].hDocWnd && IsMinimized (GroupDocs[i].hWndFrame)) 
        {
            ShowWindow (GroupDocs[i].hWndFrame, SW_RESTORE);
        }
    for (i = 0; i < MAXARTICLEWNDS; i++)
        if (ArticleDocs[i].InUse && ArticleDocs[i].hDocWnd && IsMinimized (ArticleDocs[i].hWndFrame)) {
            ShowWindow (ArticleDocs[i].hWndFrame, SW_RESTORE);
        }
    for (i = 0; i < MAXPOSTWNDS; i++)
        if (WndPosts[i].hWnd && IsMinimized (WndPosts[i].hWnd)) 
        {
            ShowWindow (WndPosts[i].hWnd, SW_RESTORE);
        }

    for (i = 0; i < MAXMAILWNDS; i++)
        if (WndMails[i].hWnd && IsMinimized (WndMails[i].hWnd)) 
        {
            ShowWindow (WndMails[i].hWnd, SW_RESTORE);
        }
    for (i = 0; i < NumStatusTexts; i++)
        if (CodingStatusText[i]->hTextWnd && IsMinimized (CodingStatusText[i]->hTextWnd)) 
        {
            ShowWindow (CodingStatusText[i]->hTextWnd, SW_RESTORE);
        }

    if (hCodedBlockWnd) 
    {
        ShowWindow (hCodedBlockWnd, SW_RESTORE);
    }
}

/* ------------------------------------------------------------------------
 *    BatchSend
 *      type is DOCTYPE_MAIL or _POST
 *      Increments nextBatchIndex and initiates mail/post
 *      Note: on entry nextBatchIndex is the index we will use for this send
 *      on exit, nextBatchIndex is either 0 (no more to send) or the index
 *      of the next mail/post to send
 *      (JSC 1/18/94)
 */
void BatchSend (int type)
{
  int thisSend;
  int nextSend; // S.Yoshida
  int maxWnds;
  WndEdit *WndEdits;

#if 1
  if (CommBusy) 
  {
    if (type == DOCTYPE_MAIL) 
    {
      PostMessage (NetDoc.hWndFrame, WM_COMMAND, IDM_SEND_NEXT_MAIL, 0);
    } 
    else 
    {
      PostMessage (NetDoc.hWndFrame, WM_COMMAND, IDM_SEND_NEXT_POST, 0);
    }
    return;
  }
#endif

  if (type == DOCTYPE_POSTING) 
  {
    WndEdits = WndPosts;
    maxWnds = MAXPOSTWNDS;
  }
  else 
  {
    WndEdits = WndMails;
    maxWnds = MAXMAILWNDS;
  }

//  Variable being used without initialization JD 11/10/95?
//  thisSend = WndEdits[thisSend].nextBatchIndex;
  thisSend = WndEdits[0].nextBatchIndex;
  if (thisSend == 0 || WndEdits[thisSend].hWnd == NULL) {// S.Yoshida  if (thisSend == 0) {          /* find first in batch (if any) */
    thisSend = 0; //S.Yoshida
    while (thisSend < maxWnds)
      if (WndEdits[thisSend].hWnd)
        break;
      else
        thisSend++;

    if (thisSend == maxWnds)
      return;                   /* no open windows.  cancel */

    WndEdits[thisSend].nextBatchIndex = thisSend;
  }
  nextSend = thisSend;
  /* find next in batch (if any) */
//  while (++WndEdits[thisSend].nextBatchIndex < maxWnds)
  while (++nextSend < maxWnds)
    if (WndEdits[nextSend].hWnd)
      break;

//  if (WndEdits[thisSend].nextBatchIndex == maxWnds)
//    WndEdits[thisSend].nextBatchIndex = 0;          /* no more */
  if (nextSend == maxWnds)
    nextSend = 0;

  WndEdits[thisSend].nextBatchIndex = nextSend;
  if (thisSend)
    WndEdits[thisSend].nextBatchIndex = nextSend;

 // ------ the last chance to confirm send composition ----
   
 // -------------------------------------------------------

  SendComposition (&WndEdits[thisSend]);
}

BOOL InlineReconnect(HWND hParentWnd)
{
	static inReconnect = FALSE;
	BOOL bReturn = FALSE;
	MSG msg;

	if (inReconnect || Initializing != INIT_NOT_CONNECTED)
		return FALSE;

	inReconnect = TRUE;
	Reconnecting = 1;

	Initializing = INIT_ESTAB_CONN;
	MRRCloseComm();
	if (MRRInitComm () != 0) 
	{    /* Set up communications */
		Initializing = INIT_NOT_CONNECTED;
		MRRCloseComm ();
		inReconnect = FALSE;
		Reconnecting = 0;
		return FALSE;
	}
	CommDoc = &NetDoc;
	idTimer = SetTimer (NetDoc.hDocWnd, ID_TIMER, COMM_TIMER_INTERVAL, (TIMERPROC) NULL);

	while (1) //wait for comm timer
	{	
		WaitMessage();
		while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) 
		{
	
			TranslateMessage(&msg);
			DispatchMessage(&msg);
			if (Initializing == INIT_NOT_CONNECTED || (CommState == ST_IN_GROUP || CommState == ST_RECONNECTED) )
				break;
		}

		if (Initializing == INIT_NOT_CONNECTED || (CommState == ST_IN_GROUP || CommState == ST_RECONNECTED))
				break;

	}
	inReconnect = FALSE;
	bReturn = CommState == ST_IN_GROUP || CommState == ST_RECONNECTED;
	if (CommState == ST_RECONNECTED)
		CommState = ST_NONE;
	return bReturn;
	
}


BOOL InlineGroup( char * szGroup)
{
	static inGroup = FALSE;
	BOOL bReturn = FALSE;

	char buf[MAXGROUPNAME + 20];
	int oldCommState = CommState;
	int oldInitalizing = Initializing;
	BOOL oldCommBusy = CommBusy;

	if (inGroup || CommBusy) 
		return FALSE;

	inGroup = TRUE;

    CommLinePtr = CommLineIn;
    CommBusy = TRUE;

	CommState = ST_INLINEGROUP_RESP;
	_sprintf(buf,"GROUP %s\r\n",szGroup);

    hSaveCursor = SetCursor (hHourGlass);
    ShowCursor (TRUE);
    PutCommLine (buf);


	//wait for comm timer
	do
	{	
		Sleep(50);	//good enough for now
		DoCommInput ();

	} while ( Initializing != INIT_NOT_CONNECTED && CommState == ST_INLINEGROUP_RESP );
	 
	//have we been disconntected?
	if (Initializing == INIT_NOT_CONNECTED)
	{
		SetCursor (hSaveCursor);
		return FALSE;
	}


	CommState = ST_GROUP_RETURN;
	_sprintf(buf,"GROUP %s\r\n",CurrentGroup);

    PutCommLine (buf);

	do
	{	
		Sleep(50); //good enough for now
		DoCommInput ();

	} while ( Initializing != INIT_NOT_CONNECTED && CommState == ST_GROUP_RETURN );

	if (Initializing != INIT_NOT_CONNECTED)
	{
		CommBusy = oldCommBusy;
		Initializing = oldInitalizing;
		CommState = oldCommState;
	}
	
	inGroup = FALSE;
	SetCursor (hSaveCursor);
	return TRUE;
	
}

/* ------------------------------------------------------------------------
 *    Test busy functions
 *      Called to test if a comm or decoding is busy
 *      Returns true if busy, false if not busy
 *      (JSC 1/9/94)
 */


BOOL TestCommBusy (HWND hParentWnd, UINT IDmsg)
{

	if ( Initializing == INIT_NOT_CONNECTED && Autoconnect && Reconnecting)
	{
		if (InlineReconnect(hParentWnd))
			return FALSE;
	}

/* Ray Dunn: 99/11 - warn if not connected (specifically useful when posting and
   connection has timed out) - used to hang winvn */
  if (Initializing != INIT_DONE)
  {
    WVMessageBox (hParentWnd, IDS_WVUTIL_NOTCONNECTED, IDmsg, MB_OK | MB_ICONASTERISK);

    return(TRUE);
  }
  else if (CommBusy)
  {
    WVMessageBox (hParentWnd, IDS_WVUTIL_TESTCOMMBUSY, IDmsg, MB_OK | MB_ICONASTERISK);
    return (TRUE);
  }
  else
    return (FALSE);
}

BOOL
TestCodingBusy (HWND hParentWnd, UINT IDmsg)
{
  if (CodingState != INACTIVE) {
    WVMessageBox (hParentWnd, IDS_WVUTIL_TESTCODING, IDmsg, MB_OK | MB_ICONASTERISK);
    return (TRUE);
  }
  else
    return (FALSE);
}

/* ------------------------------------------------------------------------
 *        Update the mail menus -- called on mail transport change
 *      jsc 9/9/94
 */
void
UpdateAllMailMenus ()
{
  register int i;

  SetMainMailMenu (&NetDoc);
  for (i = 0; i < MAXARTICLEWNDS; i++)
    if (ArticleDocs[i].hDocWnd &&
        (!CommBusy || CommDoc != &ArticleDocs[i]))
      SetArticleMailMenu (&ArticleDocs[i]);

  for (i = 0; i < MAXGROUPWNDS; i++)
    if (GroupDocs[i].hDocWnd)
      SetGroupMailMenu (&GroupDocs[i]);
}

/* ------------------------------------------------------------------------
 *        GetFreeDiskSpace in bytes, given a drive name (i.e. "c:\\")
 *      jsc 9/29/94
 */
unsigned long 
GetFreeDiskSpace (char *drive)
{
#ifdef _WIN32
  DWORD sectorsPerCluster, bytesPerSector, freeClusters, totalClusters;
  if (GetDiskFreeSpace (drive, &sectorsPerCluster, &bytesPerSector,
                        &freeClusters, &totalClusters) == FALSE) {
    return 0L;
  }
  return (unsigned long) (sectorsPerCluster * bytesPerSector * freeClusters);
#else
  struct _diskfree_t d;
  if (_dos_getdiskfree (tolower (*drive) - 'a', &d) != 0) {
    return 0L;
  }
  return (unsigned long) d.avail_clusters *
    (unsigned long) d.sectors_per_cluster *
    (unsigned long) d.bytes_per_sector;
#endif
}

/* ------------------------------------------------------------------------
 *        GetFileLength in bytes  given a file name
 *      jsc 9/29/94
 */
unsigned long 
GetFileLength (char *fileName)
{
  int fh;
  long len;

  if ((fh = _open (fileName, _O_RDONLY)) < 0) {
    return 0L;
  }
  if ((len = _filelength (fh)) < 0) {
    len = 0L;
  }
  _close (fh);
  return (unsigned long) len;
}

void
DrawDragingCursor(HWND hWnd, long oldPosition, long newPosition, BYTE flag)
{
  HDC hDc;
  HPEN hOrigPen;
  int nOrigDrawMode;
  RECT clientRect;

  hDc = GetDC(hWnd);
  hOrigPen = (HPEN)SelectObject(hDc, CreatePen(PS_SOLID, 1, ListBackgroundColor));
  nOrigDrawMode = GetROP2(hDc);
  SetROP2(hDc, R2_XORPEN);
  GetClientRect(hWnd, &clientRect);
  if ((flag == 1) || (flag == 2)) {
    MoveToEx(hDc, (int) oldPosition, 0, NULL);
    LineTo(hDc, (int) oldPosition, clientRect.bottom);
  }
  if (flag == 2) {
    MoveToEx(hDc, (int) newPosition, 0, NULL);
    LineTo(hDc, (int) newPosition, clientRect.bottom);
  }
  SetROP2(hDc, nOrigDrawMode);
  DeleteObject(SelectObject(hDc,hOrigPen));
  ReleaseDC(hWnd, hDc);
}

//Messagebox replacement that takes string IDs instead of strings
//resources save on DGROUP space for Win16

HHOOK hhk; //global hook to center our message boxes


//CBT hook to center our messageboxes on our application
LRESULT CALLBACK CBTProc(INT nCode, WPARAM wParam, LPARAM lParam)
{
	HWND  hParentWnd, hChildWnd;    // msgbox is "child"
	RECT  rParent, rChild, rDesktop;
	POINT pCenter, pStart;
	INT   nWidth, nHeight;

	// notification that a window is about to be activated
	// window handle is wParam
	if (nCode == HCBT_ACTIVATE)
	{
		
		// set window handles
		//hParentWnd = GetForegroundWindow();
		hChildWnd  = (HWND)wParam;
		hParentWnd = GetParent(hChildWnd);
		WINDOWPLACEMENT wndpl;
		ZeroMemory(&wndpl, sizeof(WINDOWPLACEMENT));
		wndpl.length = sizeof(WINDOWPLACEMENT);

		if((hParentWnd != 0) &&
			(hChildWnd != 0) &&
			(GetWindowRect(GetDesktopWindow(), &rDesktop) != 0) &&
			GetWindowPlacement(hParentWnd, &wndpl) != 0 &&
			//(GetWindowRect(hParentWnd, &rParent) != 0) &&
			(GetWindowRect(hChildWnd, &rChild) != 0))
		{
			
			

			if (wndpl.showCmd == SW_MINIMIZE)
			{
				CopyRect(&rParent, &wndpl.rcNormalPosition);	
			}
			else 
			{
				GetWindowRect(hParentWnd, &rParent);
			}


			
			
			// calculate message box dimensions
			nWidth  = (rChild.right - rChild.left);
			nHeight = (rChild.bottom - rChild.top);

			// calculate parent window center point
			pCenter.x = rParent.left+((rParent.right- rParent.left)/2);
			pCenter.y = rParent.top+((rParent.bottom- rParent.top)/2);

			// calculate message box starting point
			pStart.x = (pCenter.x - (nWidth/2));
			pStart.y = (pCenter.y - (nHeight/2));

			// adjust if message box is off desktop
			if(pStart.x < 0) 
				pStart.x = 0;
			if(pStart.y < 0) 
				pStart.y = 0;
			if(pStart.x + nWidth > rDesktop.right)
				pStart.x = rDesktop.right - nWidth;
			if(pStart.y + nHeight > rDesktop.bottom)
				pStart.y = rDesktop.bottom - nHeight;

			// move message box
			MoveWindow(hChildWnd, pStart.x, pStart.y,nWidth, nHeight,FALSE);
		}
		// exit CBT hook
		UnhookWindowsHookEx(hhk);
	}
	// otherwise, continue with any possible chained hooks
	else CallNextHookEx(hhk, nCode, wParam, lParam);
		return 0;
}

#define	TIMERID 1219
static HWND g_TimerHwnd;





void CALLBACK TimerProc(HWND hWnd, UINT uMsg, UINT idEvent, DWORD  dwTime)
{
	switch (uMsg)
	{
		case WM_TIMER:
		{
			if (idEvent == TIMERID)
			{

				TCHAR buf[1024];
				int nLen = GetDlgItemText(hWnd, 0xFFFF, buf, 1023);
				int nEnd, nStart;
				long lval = -1;
				nEnd = nStart = nLen;
				while (nStart)
				{
					
					if (isdigit(buf[nStart]))
					{

						nEnd = nStart;
						while (isdigit(buf[--nStart]))	{};
						lval = atol(buf + (++nStart));
						break;


					}
					nStart--;
				}
				
				if (lval > -1)
				{
					if (lval==0)
					{
						//end dialog
						HWND hButton = GetDlgItem(hWnd, 0x0001);
						RECT rect;
						GetWindowRect(hButton, &rect);
						POINT pt;
						pt.x = rect.left + (rect.right-rect.left)/2;
						pt.y = rect.top + (rect.bottom-rect.top)/2;

						ScreenToClient(hWnd, &pt);
						//PostMessage(hWnd, WM_LBUTTONDOWN, MK_LBUTTON, (LPARAM) POINTTOPOINTS(pt));
						//PostMessage(hWnd, WM_LBUTTONUP, MK_LBUTTON, (LPARAM) POINTTOPOINTS(pt));
						EndDialog(hWnd, 0x001);
						
					}
					else
					{
						--lval;
					}
					{
						char cval[10];
						int clen = 0;
						
						_ltoa(lval, cval, 10);
						clen = strlen(cval);
						memcpy(buf+nStart, cval, clen);

						
						if (clen < nEnd - nStart +1)
						{
							memcpy(buf +nStart+clen, buf+nEnd+1, nLen-nEnd); 
						}

					}
				}

				SetDlgItemText(hWnd, 0xFFFF, buf);
				
			}
		}

	}
	return;
}

LRESULT CALLBACK CBTTimerProc(INT nCode, WPARAM wParam, LPARAM lParam)
{
	// notification that a window is about to be activated
	// window handle is wParam
	if (nCode == HCBT_ACTIVATE)
	{
		HWND hWnd  = (HWND)wParam;
		HWND hParent = GetParent(hWnd);
		RECT rect;
		WINDOWPLACEMENT wndpl;
		ZeroMemory(&wndpl, sizeof(WINDOWPLACEMENT));
		wndpl.length = sizeof(WINDOWPLACEMENT);

		g_TimerHwnd = hWnd;

		if((hParent != 0) &&
			(hWnd != 0) &&
			(GetWindowRect(GetDesktopWindow(), &rect) != 0) &&
			GetWindowPlacement(hParent, &wndpl) != 0 &&
			wndpl.showCmd != SW_MINIMIZE)
		{
			CenterWindow(hWnd, hParent);
		}


		SetTimer( hWnd, TIMERID, 2000, &TimerProc);
		
		SetDlgItemText(hWnd, 0x0001, "Retry");
		
		
		// exit CBT hook
		UnhookWindowsHookEx(hhk);
	}
	// otherwise, continue with any possible chained hooks
	else CallNextHookEx(hhk, nCode, wParam, lParam);
		return 0;

}



#define MESSAGESIZE 1024
int WVMessageBox(HWND hwnd, LPSTR lpText, LPSTR lpCaption,UINT uType)
{
	hhk = SetWindowsHookEx(WH_CBT, &CBTProc, 0,GetCurrentThreadId());
	return MessageBox(hwnd, lpText, lpCaption, uType);
}

int WVTimedMessageBox(HWND hwnd, LPSTR lpText, LPSTR lpCaption,UINT uType)
{

	hhk = SetWindowsHookEx(WH_CBT, &CBTTimerProc, 0,GetCurrentThreadId());
	UINT uReturn = MessageBox(hwnd, lpText, lpCaption, uType);
	KillTimer(g_TimerHwnd,	 TIMERID);
	return uReturn;
}

int WVMessageBox(HWND hwndParent, UINT idText, UINT idTitle, UINT fuStyle)
{
	HGLOBAL hBuf;
	char far *pText = NULL;
	char far *pTitle = NULL;
	int nReturn = 0;

	hBuf = ::GlobalAlloc(GHND, (MESSAGESIZE) * 2 );
	if (hBuf) 
	{
		pText = (char far *) ::GlobalLock(hBuf);
		if (pText) 
		{

			pTitle = pText + MESSAGESIZE;
			LoadString(hInst,idText, pText, MESSAGESIZE);
			LoadString(hInst,idTitle, pTitle, MESSAGESIZE);

			hhk = SetWindowsHookEx(WH_CBT, &CBTProc, 0, GetCurrentThreadId());

			nReturn = MessageBox(hwndParent, pText, pTitle, fuStyle);

			::GlobalUnlock(hBuf);
			pText = pTitle = NULL;
		}
		::GlobalFree(hBuf);
		hBuf = NULL;             
	}
	return nReturn;
}

int WVArgMessageBox(HWND hwndParent, UINT idFormat, UINT idTitle, UINT fuStyle, const char far *psubst)
{

	HGLOBAL hBuf;
	char far *pFormat = NULL;
	char far *pTitle = NULL;
	char far *pMesg = NULL;
	int nReturn = 0;

	hBuf = ::GlobalAlloc(GHND, (MESSAGESIZE) * 3 );
	if (hBuf) 
	{
		pFormat = (char far *) ::GlobalLock(hBuf);
		if (pFormat) 
		{

			pTitle = pFormat + MESSAGESIZE;
			pMesg = pTitle + MESSAGESIZE;
			LoadString(hInst,idTitle, pTitle, MESSAGESIZE);
			if (LoadString(hInst,idFormat, pFormat, MESSAGESIZE)) 
			{

				_snprintf(pMesg, MESSAGESIZE, pFormat, psubst);
			}

			hhk = SetWindowsHookEx(WH_CBT, &CBTProc, 0, GetCurrentThreadId());
			nReturn = MessageBox(hwndParent, pMesg, pTitle, fuStyle);

			::GlobalUnlock(hBuf);
			pFormat = pMesg = pTitle = NULL;
		}
		::GlobalFree(hBuf);
		hBuf = NULL;             
	}
	return nReturn;
}

#if 0
int WVArg2MessageBox(HWND hwndParent, UINT idFormat, UINT idTitle, UINT fuStyle, const char far *psubst1, const char far *psubst2)
{

	HGLOBAL hBuf;
	char far *pFormat = NULL;
	char far *pTitle = NULL;
	char far *pMesg = NULL;
	int nReturn = 0;

	hBuf = ::GlobalAlloc(GHND, (MESSAGESIZE) * 3 );
	if (hBuf) 
	{
		pFormat = (char far *) ::GlobalLock(hBuf);
		if (pFormat) 
		{

			pTitle = pFormat + MESSAGESIZE;
			pMesg = pTitle + MESSAGESIZE;
			LoadString(hInst,idTitle, pTitle, MESSAGESIZE);
			if (LoadString(hInst,idFormat, pFormat, MESSAGESIZE)) 
			{

				_snprintf(pMesg, MESSAGESIZE, pFormat, psubst1, psubst2);
			}
			
			hhk = SetWindowsHookEx(WH_CBT, &CBTProc, 0, GetCurrentThreadId());
			nReturn = MessageBox(hwndParent, pMesg, pTitle, fuStyle);

			::GlobalUnlock(hBuf);
			pFormat = pMesg = pTitle = NULL;
		}
		::GlobalFree(hBuf);
		hBuf = NULL;             
	}
	return nReturn;
}
#endif

int WVArgXMessageBox(HWND hwndParent, UINT idFormat, UINT idTitle, UINT fuStyle, ...)
{
	va_list vargs;
	char far *pFormat = NULL;
	char far *pMesg = NULL;
	char far *pTitle = NULL;
	BOOL bReturn = FALSE;
	int  nStrLen = 0;
	int nReturn = 0;

	pMesg = (char far *)GlobalAllocPtr (GMEM_MOVEABLE, MESSAGESIZE * sizeof (char));
	pFormat = (char far *)GlobalAllocPtr (GMEM_MOVEABLE, MESSAGESIZE * sizeof (char) * 2);

	if (pFormat && pMesg) 
	{
		*pMesg = '\0';
		pTitle = pFormat + MESSAGESIZE;
		nStrLen = LoadString(hInst, idFormat, pFormat, MESSAGESIZE);
		nStrLen = LoadString(hInst, idTitle, pTitle, MESSAGESIZE);

		if (nStrLen && nStrLen < MESSAGESIZE) 
		{
		  va_start( vargs, fuStyle );
		  wvsprintf( pMesg, pFormat, vargs );
		  va_end( vargs );        
		}
		hhk = SetWindowsHookEx(WH_CBT, &CBTProc, 0, GetCurrentThreadId());
		nReturn = MessageBox(hwndParent, pMesg, pTitle, fuStyle);
          
	}
	if (pFormat)
		GlobalFreePtr(pFormat);

	if (pMesg)
		GlobalFreePtr(pMesg);

	return nReturn;
}


BOOL WVSetWindowText(HWND hwndParent, UINT idText)
{
	HGLOBAL hBuf;
	char far *pText = NULL;
	BOOL bReturn = FALSE;

	hBuf = ::GlobalAlloc(GHND, (MESSAGESIZE));
	if (hBuf) 
	{
		pText = (char far *) ::GlobalLock(hBuf);
		if (pText) 
		{
			LoadString(hInst, idText, pText, MESSAGESIZE);

			bReturn = SetWindowText(hwndParent, pText);
    
			::GlobalUnlock(hBuf);
			pText= NULL;
		}
		::GlobalFree(hBuf);
		hBuf = NULL;             
	}
  return bReturn;
}

BOOL WVArgSetWindowText(HWND hwndParent, UINT idFormat, const char far *psubst1)
{
	HGLOBAL hBuf;
	char far *pFormat = NULL;
	char far *pMesg = NULL;
	BOOL bReturn = FALSE;

	hBuf = ::GlobalAlloc(GHND, (MESSAGESIZE) * 2 );
	if (hBuf) 
	{
		pFormat = (char far *) ::GlobalLock(hBuf);
		if (pFormat) 
		{
			pMesg = pFormat + MESSAGESIZE;
			if (LoadString(hInst, idFormat, pFormat, MESSAGESIZE)) 
			{
				_snprintf(pMesg, MESSAGESIZE, pFormat, psubst1);
			}

			bReturn = SetWindowText(hwndParent, pMesg);

			::GlobalUnlock(hBuf);
			pFormat = pMesg = NULL;
		}
		::GlobalFree(hBuf);
		hBuf = NULL;             
	}
	return bReturn;
}


BOOL WVArg2SetWindowText(HWND hwndParent, UINT idFormat, const char far *psubst1, const char far *psubst2)
{
  HGLOBAL hBuf;
  char far *pFormat = NULL;
  char far *pMesg = NULL;
  BOOL bReturn = FALSE;
    
  hBuf = ::GlobalAlloc(GHND, (MESSAGESIZE) * 2 );
  if (hBuf) 
  {
    pFormat = (char far *) ::GlobalLock(hBuf);
    if (pFormat) {
      pMesg = pFormat + MESSAGESIZE;
      if (LoadString(hInst, idFormat, pFormat, MESSAGESIZE)) 
	  {
        _snprintf(pMesg, MESSAGESIZE, pFormat, psubst1, psubst2);
      }
      
#ifdef _WIN32
      bReturn = SetWindowText(hwndParent, pMesg);
#else
      SetWindowText(hwndParent, pMesg);
      bReturn = TRUE; /* assume no error because win16 hasn't a return value */
#endif

      ::GlobalUnlock(hBuf);
      pFormat = pMesg = NULL;
    }
    ::GlobalFree(hBuf);
    hBuf = NULL;             
  }
  return bReturn;
}


BOOL WVArg3SetWindowText(HWND hwndParent, UINT idFormat, const char far *psubst1, const char far *psubst2, const char far *psubst3)
{
  HGLOBAL hBuf;
  char far *pFormat = NULL;
  char far *pMesg = NULL;
  BOOL bReturn = FALSE;
    
  hBuf = ::GlobalAlloc(GHND, (MESSAGESIZE) * 2 );
  if (hBuf) {
    pFormat = (char far *) ::GlobalLock(hBuf);
    if (pFormat) {
      pMesg = pFormat + MESSAGESIZE;
      if (LoadString(hInst, idFormat, pFormat, MESSAGESIZE)) {
        _snprintf(pMesg, MESSAGESIZE, pFormat, psubst1, psubst2, psubst3);
      }
      
#ifdef _WIN32
      bReturn = SetWindowText(hwndParent, pMesg);
#else
      SetWindowText(hwndParent, pMesg);
      bReturn = TRUE; /* assume no error because win16 hasn't a return value */
#endif
            
      ::GlobalUnlock(hBuf);
      pFormat = pMesg = NULL;
    }
    ::GlobalFree(hBuf);
    hBuf = NULL;             
  }
  return bReturn;
}

BOOL WVArgXSetWindowText(HWND hwndParent, UINT idFormat, ...)
{
  va_list vargs;
  char far *pFormat = NULL;
  char far *pMesg = NULL;
  BOOL bReturn = FALSE;
  int  nStrLen = 0;

  pMesg = (char far *)GlobalAllocPtr (GMEM_MOVEABLE, MESSAGESIZE * sizeof (char));
  pFormat = (char far *)GlobalAllocPtr (GMEM_MOVEABLE, MESSAGESIZE * sizeof (char));

  if (pFormat && pMesg) 
  {
    *pMesg = '\0';

    nStrLen = LoadString(hInst, idFormat, pFormat, MESSAGESIZE);

    if (nStrLen && nStrLen < MESSAGESIZE) {
      va_start( vargs, idFormat );
      wvsprintf( pMesg, pFormat, vargs );
      va_end( vargs );        
    }

#ifdef _WIN32
  bReturn = SetWindowText(hwndParent, pMesg);
#else
  SetWindowText(hwndParent, pMesg);
  bReturn = TRUE; /* assume no error because win16 hasn't a return value */
#endif                 
  }
  if (pFormat)
    GlobalFreePtr(pFormat);

  if (pMesg)
    GlobalFreePtr(pMesg);

  return bReturn;
}



BOOL
TestArgCommBusy (HWND hParentWnd, UINT IDmsg, const char far *psubst)
{
  HGLOBAL hBuf;
  char far *pFormat = NULL;
  char far *pMesg = NULL;
  char far *pTitle = NULL;
  BOOL bReturn = FALSE;
    
  hBuf = ::GlobalAlloc(GHND, (MESSAGESIZE) * 3 );
  if (hBuf) {
    pFormat = (char far *) ::GlobalLock(hBuf);
    if (pFormat) {
      pMesg = pFormat + MESSAGESIZE;
      pTitle = pMesg +  MESSAGESIZE;
      LoadString(hInst, IDS_WVUTIL_TESTCOMMBUSY, pMesg, MESSAGESIZE);
      if (LoadString(hInst, IDmsg, pFormat, MESSAGESIZE)) {
        _snprintf(pTitle, MESSAGESIZE, pFormat, psubst);
      }

      if (CommBusy) {
      MessageBox (hParentWnd, pMesg, pTitle, MB_OK | MB_ICONASTERISK);
      bReturn = TRUE;
      } else
      bReturn = FALSE;

      ::GlobalUnlock(hBuf);
      pFormat = pMesg = pTitle = NULL;
    }
    ::GlobalFree(hBuf);
    hBuf = NULL;             
  }
  return bReturn;
}

BOOL RetrieveNextGroup(
  TypBlock far *BlockPtr,
  TypLine far *LinePtr,
  int nTitle,
  BOOL bConfirm )
{
  FlagJumpGroup = 0;
  while(1) {
    if( !NextActiveLine( &BlockPtr, &LinePtr ) ) {
      return FALSE;
    }
    if( !(GetGroup(LinePtr))->Determined ) { // Ignore NNTP
      break;
    }
    if( !(GetGroup(LinePtr))->ServerEstNum
     || !(GetGroup(LinePtr))->NumUnread ) {
      continue;
    }
    break;
  }
  HWND hWnd = CommDoc ? CommDoc->hDocWnd : NetDoc.hDocWnd;
  if ( bConfirm || WVArgMessageBox(hWnd, IDS_WVUTIL_NEXTGROUP, nTitle, MB_YESNO, GetGroupName(LinePtr)) == IDYES) {
    SetSelections( &NetDoc, LinePtr );
    AdjustTopSc( BlockPtr, LinePtr );
    if( GroupListMultiSelect ) {
      /* set the active and anchor line IDs = LnkFind.LineID */
      NetDoc.AnchorLineID = LinePtr->LineID;
    }
    NetDoc.ActiveLineID = LinePtr->LineID;
    //if( ConfigJumpGroup & GJUMP_OPENGROUP ) {
      //ConfigJumpGroup |= GJUMP_OPN_GROUP;
      FlagJumpGroup = JUMPFLAG_GROUP;
      PostMessage( NetDoc.hDocWnd, WM_CHAR, '\r', 0 );
    //} else {
     //SetForegroundWindow( NetDoc.hWndFrame );
     //InvalidateRect(NetDoc.hWndFrame,NULL,FALSE);
     //InvalidateRect(NetDoc.hDocWnd,NULL,FALSE);
    //}
  }
  return TRUE;
}

// Search HKCL_Root for the specified key, return the default shell command
// MDS 20000410
BOOL GetRegDefaultCommand(LPCSTR szKey, LPSTR szCommandBuf, int nLen)
{
  
  HKEY hKey = 0;
  long lSize = nLen;
  BOOL bFound = FALSE;
  char *endcmd = NULL;

  if (!szCommandBuf || !szKey || *szKey == '\0')
    return FALSE;

  *szCommandBuf = '\0';
  if (RegOpenKey(HKEY_CLASSES_ROOT, szKey, &hKey) == ERROR_SUCCESS)
  {
    if (RegQueryValue( hKey, "shell\\open\\command", 
                    szCommandBuf, &lSize) == ERROR_SUCCESS)
    {

      bFound = TRUE;

      //remove quote wrapped path
      if (*szCommandBuf == '"')
      {

        memmove(szCommandBuf, szCommandBuf+1, strlen(szCommandBuf+1));
        endcmd = strchr(szCommandBuf, '"');
        if (endcmd)
        {
          //toss args
          *endcmd = '\0';
        }
        
#ifdef _WIN32
        if (strchr(szCommandBuf, ' ') )
        {

          char szBuf[_MAX_PATH];
          DWORD dwReturn;
          dwReturn = GetShortPathName(szCommandBuf, szBuf, _MAX_PATH);
          if (dwReturn != 0 && dwReturn < _MAX_PATH)
          {
            strncpy(szCommandBuf, szBuf, nLen);
          } 
        } 
#endif 
      }
      else
      {
        endcmd = strchr(szCommandBuf, ' ');
        if (endcmd)
        {
          //toss args
          *endcmd = '\0';;
        }
      }

    }
    
    RegCloseKey(hKey);

  } 

  return bFound;

}

/////////////////////////////////////////////////////////////////////////////
// DWORD GetFileVersion(LPTSTR lpszFilename)
//
// Retrieve the windows file version of the specified file
//
//  Input: a string containing the name of a path & file to search
//
//  Returns a DWORD containing the file's FileVersion. A file's version is 
//  stored as two DWORD values. This routine uses the low byte of each
//  part of the version number to create the dword. Thus the maximum
//  version number of each of the 4 parts is 255.
//  
//  Returns 0 if version information not found.
//
DWORD GetFileVersion(LPTSTR lpszFilename)
{
    DWORD   dwVersion = 0; 
    DWORD   dwSize;         // Size of mem needed to hold version info
    DWORD   dwDummy;        // Pointer to variable to receive zero
    
    // Get Version info size  
    dwSize = GetFileVersionInfoSize(lpszFilename, &dwDummy);
    if (dwSize != 0)
    {
        // Allocate space to hold the info
        LPBYTE pInfoBuf = (LPBYTE) GlobalAllocPtr (GHND, dwSize);
        if (pInfoBuf != NULL)
        {
            // Get the info 
            if (GetFileVersionInfo(lpszFilename, 0, dwSize, pInfoBuf))
            {
                // Got version info, get pointer to VS_FIXEDFILEINFO & size
                // (non language dependant part)
                VS_FIXEDFILEINFO FAR *lpBuffer;     // Pointer to version results 
                unsigned int nBuffSize;             // Size of version results

                VerQueryValue(pInfoBuf, _T("\\"), (void far**)&lpBuffer, &nBuffSize);
                if (nBuffSize != 0)
                {
                    // File version stored as 64bit word (4 16 bit values)
                    // Convert this to 32bit word (4 8 bit values)
                    dwVersion =
                     ( (DWORD)(LOBYTE(HIWORD(lpBuffer->dwFileVersionMS))) << 24 
                    |  (DWORD)(LOBYTE(LOWORD(lpBuffer->dwFileVersionMS))) << 16
                    |  (DWORD)(LOBYTE(HIWORD(lpBuffer->dwFileVersionLS))) << 8
                    |  (DWORD)(LOBYTE(LOWORD(lpBuffer->dwFileVersionLS))) );
                } // if we got VS_FIXEDFILEINFO data
            } // If we found version info

            // Let go of the memory
            GlobalFreePtr(pInfoBuf);

        } // If we allocated memory
    } // If file has version info
    
    return dwVersion;
}  

DWORD GetDllVersion(LPCTSTR lpszDllName)
{

        
#ifdef _WIN32

    TCHAR szPath[_MAX_PATH+1];
    LPTSTR pFile;
    if (SearchPath( NULL, lpszDllName, NULL, _MAX_PATH, szPath, &pFile ))
    {
        return GetFileVersion(szPath);
    }

#else

    HMODULE  hModule;                 // handle to a module
    MODULEENTRY myModuleEntry;        // module entry structure
    
    myModuleEntry.dwSize = sizeof(myModuleEntry);
    
    //The display driver in memory
    hModule = ModuleFindName( &myModuleEntry, lpszDllName); 
    if (hModule != NULL)
    {
        //Path to loaded display driver
        return GetFileVersion(myModuleEntry.szExePath);
    }    

#endif
    return 0L;    

}

void MakeXFaceWnd(const char * szFace)
{
#ifdef _WIN32
    XFaceOb Face;
    HGLOBAL hDib = NULL;
    
    if (!CommDoc || !szFace || !*szFace || !ShowXFace)
		return;
	
	if (Face.Xface2BMP( szFace, &hDib) == 1 && hDib)
    {
        
        if (hDib && CommDoc->hWndXFace)
		{
			SendMessage(CommDoc->hWndXFace, WM_UPDATEFACE, (WPARAM) hDib, 0);
			GlobalFree(hDib);
			return;
		}
		
		
		RECT rect;
		WINDOWPLACEMENT wndpl;
		int nShow = SW_SHOWNOACTIVATE;

		ZeroMemory(&wndpl, sizeof(WINDOWPLACEMENT));
		wndpl.length = sizeof(WINDOWPLACEMENT);
		
		GetWindowRect(CommDoc->hWndFrame, &rect);

		if (GetWindowPlacement(CommDoc->hWndFrame, &wndpl) )
		{
			if (wndpl.showCmd == SW_MINIMIZE || IsIconic(CommDoc->hWndFrame))
			{
				CopyRect(&rect, &wndpl.rcNormalPosition);
				nShow = SW_HIDE;	
			}
			
		}

		rect.top += ptXFaceCoord.y;
		rect.left += ptXFaceCoord.x;
		rect.right = rect.left + 47 + 4* GetSystemMetrics(SM_CXEDGE);
		rect.bottom = rect.top + 47 + 4* GetSystemMetrics(SM_CYEDGE);
    
		MakeRectVisible(rect);
		if (CommDoc->hWndXFace)
		{
			DestroyWindow(CommDoc->hWndXFace);
		}
    
		CommDoc->hWndXFace = CreateWindow ("WinVnArtFace",
			"",
			 WS_DLGFRAME | WS_POPUP ,
			rect.left,
			rect.top,
			rect.right-rect.left,
			rect.bottom-rect.top,
			CommDoc->hWndFrame,
			NULL,
			hInst,
			(void *)hDib);
		if (CommDoc->hWndXFace)
		{
			//dont want xface to take focus away
			ShowWindow(CommDoc->hWndXFace, nShow);

		}


    }

    if (hDib)
    {
        GlobalFree(hDib);
        hDib = NULL;
    }

#endif
}

void MakeFaceWnd(const char * szFace)
{
#ifdef _WIN32

	HGLOBAL hDib = NULL;
    if (!CommDoc || !szFace || !*szFace || !ShowXFace)
		return;

	{
		
		if (!g_GDIPlusImage.Init())
		{
			return;
		}
		

		 hDib = g_GDIPlusImage.DIBFromFace( szFace );
	}    

	
	if (hDib)
    {
        
        if (CommDoc->hWndXFace)
		{
			LPBYTE pData = (LPBYTE) GlobalLock(hDib);
			SendMessage(CommDoc->hWndXFace, WM_UPDATEFACE, (WPARAM) pData, 0);
			GlobalUnlock(hDib);
			GlobalFree(hDib);
			return;
		}
		
		
		RECT rect;
		WINDOWPLACEMENT wndpl;
		int nShow = SW_SHOWNOACTIVATE;

		ZeroMemory(&wndpl, sizeof(WINDOWPLACEMENT));
		wndpl.length = sizeof(WINDOWPLACEMENT);
		
		GetWindowRect(CommDoc->hWndFrame, &rect);

		if (GetWindowPlacement(CommDoc->hWndFrame, &wndpl) )
		{
			if (wndpl.showCmd == SW_MINIMIZE || IsIconic(CommDoc->hWndFrame))
			{
				CopyRect(&rect, &wndpl.rcNormalPosition);
				nShow = SW_HIDE;	
			}
			
		}

		rect.top += ptXFaceCoord.y;
		rect.left += ptXFaceCoord.x;
		rect.right = rect.left + 47 + 4* GetSystemMetrics(SM_CXEDGE);
		rect.bottom = rect.top + 47 + 4* GetSystemMetrics(SM_CYEDGE);
    
		MakeRectVisible(rect);
		if (CommDoc->hWndXFace)
		{
			DestroyWindow(CommDoc->hWndXFace);
		}
    
		CommDoc->hWndXFace = CreateWindow ("WinVnArtFace",
			"",
			 WS_DLGFRAME | WS_POPUP ,
			rect.left,
			rect.top,
			rect.right-rect.left,
			rect.bottom-rect.top,
			CommDoc->hWndFrame,
			NULL,
			hInst,
			(void *)hDib);
		if (CommDoc->hWndXFace)
		{
			//dont want xface to take focus away
			ShowWindow(CommDoc->hWndXFace, nShow);

		}



        GlobalFree(hDib);
        hDib = NULL;
    }

#endif
}



//returns the number of bytes copied
int figurePathFromPidl( LPITEMIDLIST pidl, TCHAR *szRetPath, int nPathLen)
{
	TCHAR szPath[_MAX_PATH] = {'\0'};

	if (! SHGetPathFromIDList(pidl, szPath))
		*szPath = '\0'; // just 2 b safe
	// Allocate a pointer to an IMalloc interface
	LPMALLOC pMalloc;

	// Get the address of our task allocator's IMalloc interface
	if (SUCCEEDED( SHGetMalloc(&pMalloc)))
	{	// Free the item ID list allocated by SHGetSpecialFolderLocation
		pMalloc->Free(pidl);

		// Free our task allocator
		pMalloc->Release();
	}
	if (szRetPath)
	{
		_tcsncpy(szRetPath, szPath, nPathLen);
	}
	return min((int) _tcslen(szPath), nPathLen);
}

int getSpecialFolder( UINT uCSIDL, TCHAR *szRetPath, int nPathLen)
{
	LPITEMIDLIST pidl;
	HRESULT hr = SHGetSpecialFolderLocation(NULL, uCSIDL, &pidl);

	if (SUCCEEDED( hr))
		return figurePathFromPidl( pidl, szRetPath, nPathLen);
	else
		return 0;

}

BOOL WVGetFileWriteTime(LPCTSTR szFile, FILETIME& writeTime)
{

	BOOL bOK = FALSE;
	HANDLE hFile = ::CreateFile(szFile, GENERIC_READ , FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, 
					 OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	
	writeTime.dwLowDateTime = 0;
	writeTime.dwHighDateTime = 0;

	if (hFile != INVALID_HANDLE_VALUE)
	{
		bOK =  GetFileTime(hFile,NULL,NULL, &writeTime);
		CloseHandle(hFile);
	}

	return bOK;

}


//
// Local variables:
// tab-width: 2
// end:
//

